
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

extern void	Error();

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : TIFF ",path); Error(msg);
}

#define XSIZE	depth->xSize
#define YSIZE	depth->ySize

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	char	*ptr,*desc,
		*fnf="file not found.",
		*idnf="image description not found.",
		*bit="bad image type.",
		*idnm="image description not match.",
		*isnm="image size not match.";
	int	code,z,x,y,biasR,biasG,biasB,bgR,bgG,bgB,R,G,B,gray=1;
	Image	*depth,*label,*shade;
	CM	*cm;
	double	den,gammaR,gammaG,gammaB,mulR,mulG,mulB,indR,indG,indB,ratio;

	if (argc!=14)
	    Error(
"usage : bevmCS depth label shade gamma[R,G,B] bias[R,G,B] bg[R,G,B] color"
	    );

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Abort(argv[1],fnf);

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(argv[1],idnf);

	if ((desc=(char *)malloc(strlen(ptr)+1))==NULL)
	    Abort(argv[1],"image description memory allocation error.");

	(void)strcpy(desc,ptr);

	if ((code=ReadTIFF(tiff,&depth,&cm))>0)
	    Abort(argv[1],ReadTIFFError(code));

	TIFFClose(tiff);

	if (cm==NULL) Abort(argv[1],bit);

	FreeCM(cm);

	if ((tiff=TIFFOpen(argv[2],"r"))==NULL) Abort(argv[2],fnf);

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(argv[2],idnf);

	if (strcmp(desc,ptr)) Abort(argv[2],idnm);

	if ((code=ReadTIFF(tiff,&label,&cm))>0)
	    Abort(argv[2],ReadTIFFError(code));

	TIFFClose(tiff);

	if (label->xSize!=XSIZE || label->ySize!=YSIZE) Abort(argv[2],isnm);

        if (cm!=NULL) {
	    for (z=0; z<cm->pixels; z++) {
		cm->r[z]>>=8;
		cm->g[z]>>=8;
		cm->b[z]>>=8;
	    }
	    for (x=0; x<XSIZE; x++)
	    for (y=0; y<YSIZE; y++) {
		z=GETPIXEL(label,x,y); label->r[x][y]=(byte)cm->r[z];
				       label->g[x][y]=(byte)cm->g[z];
				       label->b[x][y]=(byte)cm->b[z];
	    }
	    FreeCM(cm);
	}

	if ((tiff=TIFFOpen(argv[3],"r"))==NULL) Abort(argv[3],fnf);

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(argv[3],idnf);

	if (strcmp(desc,ptr)) Abort(argv[3],idnm);

	if ((code=ReadTIFF(tiff,&shade,&cm))>0)
	    Abort(argv[3],ReadTIFFError(code));

	TIFFClose(tiff);

	if (shade->xSize!=XSIZE || shade->ySize!=YSIZE) Abort(argv[3],isnm);

	if (cm==NULL) Abort(argv[3],bit);

	den=(double)(cm->pixels-1); FreeCM(cm);

	if ((gammaR=atof(argv[4]))<=0.0 ||
	    (gammaG=atof(argv[5]))<=0.0 ||
	    (gammaB=atof(argv[6]))<=0.0) Error("bad gamma factor.");

	if ((biasR=atoi(argv[7]))<0 || biasR>255 ||
	    (biasG=atoi(argv[8]))<0 || biasG>255 ||
	    (biasB=atoi(argv[9]))<0 || biasB>255) Error("bad bias value.");

	if ((bgR=atoi(argv[10]))<0 || bgR>255 ||
	    (bgG=atoi(argv[11]))<0 || bgG>255 ||
	    (bgB=atoi(argv[12]))<0 || bgB>255) Error("bad background value.");

	mulR=(double)(255-biasR)/255.0; indR=1.0/gammaR;
	mulG=(double)(255-biasG)/255.0; indG=1.0/gammaG;
	mulB=(double)(255-biasB)/255.0; indB=1.0/gammaB;
	for (x=0; x<XSIZE; x++)
	for (y=0; y<YSIZE; y++) {
	    if (GETPIXEL(depth,x,y)) {
		ratio=(double)GETPIXEL(shade,x,y)/den;
		R=biasR+(int)(mulR*pow(ratio,indR)*(double)label->r[x][y]+0.5);
		G=biasG+(int)(mulG*pow(ratio,indG)*(double)label->g[x][y]+0.5);
		B=biasB+(int)(mulB*pow(ratio,indB)*(double)label->b[x][y]+0.5);
	    }
	    else {
		R=bgR;
		G=bgG;
		B=bgB;
	    }
	    shade->r[x][y]=(byte)R;
	    shade->g[x][y]=(byte)G;
	    shade->b[x][y]=(byte)B; if (gray) gray=(R==G && R==B);
	}
	if (gray) {
	    for (x=0; x<XSIZE; x++)
	    for (y=0; y<YSIZE; y++) {
		z=(int)shade->r[x][y]; PUTPIXEL(shade,x,y,z);
	    }

	    if ((cm=AllocCM(256))==NULL)
		Error("colormap memory allocation error.");

	    for (z=0; z<256; z++) cm->r[z]=cm->g[z]=cm->b[z]=(word)((z<<8)|z);

	}
	else
	    cm=NULL;

	if ((tiff=TIFFOpen(argv[13],"w"))==NULL)
	    Abort(argv[13],"file not stored.");

	if (!TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,desc))
	    Abort(argv[13],"image description not stored.");

	if ((code=WriteTIFF(tiff,shade,cm))>0)
	    Abort(argv[13],WriteTIFFError(code));

	TIFFClose(tiff);

	return 0;
}
