
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

extern void	Error();

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : TIFF ",path); Error(msg);
}

#define XSIZE	depth->xSize
#define YSIZE	depth->ySize

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	char	*ptr,*desc,
		*fnf="file not found.",
		*idnf="image description not found.",
		*bit="bad image type.";
	int	code,limit,bias,bg,x,y,z;
	Image	*depth,*shade;
	CM	*cm;
	double	gamma,mul,den,ind;

	if (argc!=7) Error("usage : bevmGS depth shade gamma bias bg gray");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Abort(argv[1],fnf);

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(argv[1],idnf);

	if ((desc=(char *)malloc(strlen(ptr)+1))==NULL)
	    Abort(argv[1],"image description memory allocation error.");

	(void)strcpy(desc,ptr);

	if ((code=ReadTIFF(tiff,&depth,&cm))>0)
	    Abort(argv[1],ReadTIFFError(code));

	TIFFClose(tiff);

	if (cm==NULL) Abort(argv[1],bit);

	FreeCM(cm);

	if ((tiff=TIFFOpen(argv[2],"r"))==NULL) Abort(argv[2],fnf);

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(argv[2],idnf);

	if (strcmp(desc,ptr)) Abort(argv[2],"image description not match.");

	if ((code=ReadTIFF(tiff,&shade,&cm))>0)
	    Abort(argv[2],ReadTIFFError(code));

	TIFFClose(tiff);

	if (shade->xSize!=XSIZE ||
	    shade->ySize!=YSIZE) Abort(argv[2],"image size not match.");

	if (cm==NULL) Abort(argv[2],bit);

	limit=cm->pixels-1;

	if ((gamma=atof(argv[3]))<=0.0) Error("bad gamma factor.");

	if ((bias=atoi(argv[4]))<0 || bias>limit) Error("bad bias value.");

	if ((bg=atoi(argv[5]))<0 || bg>limit) Error("bad background value.");

	mul=(double)(limit-bias); den=(double)limit; ind=1.0/gamma;
	for (x=0; x<XSIZE; x++)
	for (y=0; y<YSIZE; y++) {
	    z=(GETPIXEL(depth,x,y))?
	      bias+(int)(mul*pow((double)GETPIXEL(shade,x,y)/den,ind)+0.5):bg;
	    PUTPIXEL(shade,x,y,z);
	}

	if ((tiff=TIFFOpen(argv[6],"w"))==NULL)
	    Abort(argv[6],"file not stored.");

	if (!TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,desc))
	    Abort(argv[6],"image description not stored.");

	if ((code=WriteTIFF(tiff,shade,cm))>0)
	    Abort(argv[6],WriteTIFFError(code));

	TIFFClose(tiff);

	return 0;
}
