
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "slice.h"
#include "bev.h"

#define EXTENSION_DEPTH	"d.tif"
#define EXTENSION_LABEL	"l.tif"
#define EXTENSION_SHADE	"s.tif"
#define EXTENSION_EDGE	"e.tif"

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

typedef unsigned short	Word;
typedef unsigned char	Byte;

static Word	**AllocateWordMemory(wid,hig)
int		wid,hig;
{
	Word	**word;
	int	x,y;
	char	*msg="word memory allocation error.";

	if ((word=(Word **)MA(hig,word))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((word[y]=(Word *)MA(wid,*word))==NULL) Error(msg);

	    for (x=0; x<wid; x++) word[y][x]=0;
	}
	return word;
}

#define AllocateCellMemory	AllocateSliceMemory

static Byte	**AllocateByteMemory(wid,hig)
int		wid,hig;
{
	Byte	**byte;
	int	x,y;
	char	*msg="byte memory allocation error.";

	if ((byte=(Byte **)MA(hig,byte))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((byte[y]=(Byte *)MA(wid,*byte))==NULL) Error(msg);

	    for (x=0; x<wid; x++) byte[y][x]=0;
	}
	return byte;
}

static void	ScanEdge(h1,v1,h2,v2,hMin,hMax)
int		h1,v1,h2,v2,*hMin,*hMax;
{
	int	wh,wv,sh,sv,m,p,e,h,v;

	if (h2>=h1) { wh=h2-h1; sh=1; } else { wh=h1-h2; sh=(-1); }
	if (v2>=v1) { wv=v2-v1; sv=1; } else { wv=v1-v2; sv=(-1); }
	if (wh>=wv) {
	    h2+=sh; m=2*wv; e=m-wh; p=e-wh;
	    for (v=v1, h=h1; h!=h2; h+=sh) {
		if (h<hMin[v]) hMin[v]=h;
		if (h>hMax[v]) hMax[v]=h;
		if (e>=0) { v+=sv; e+=p; } else e+=m;
	    }
	}
	else {
	    v2+=sv; m=2*wh; e=m-wv; p=e-wv;
	    for (h=h1, v=v1; v!=v2; v+=sv) {
		if (h<hMin[v]) hMin[v]=h;
		if (h>hMax[v]) hMax[v]=h;
		if (e>=0) { h+=sh; e+=p; } else e+=m;
	    }
	}
}

static void	PaintParallelogram(a,b,c,h0,v0,dh1,dv1,dh2,dv2,hMin,hMax,depth)
double		a,b,c,h0,v0,dh1,dv1,dh2,dv2;
int		*hMin,*hMax;
Word		**depth;
{
	int	hmin,hmax,vMin,vMax,h,v,d,
		h1=(int)floor(h0        )+1,v1=(int)floor(v0        )+1,
		h2=(int)floor(h0+dh1    )+1,v2=(int)floor(v0+dv1    )+1,
		h3=(int)floor(h0+dh1+dh2)+1,v3=(int)floor(v0+dv1+dv2)+1,
		h4=(int)floor(h0    +dh2)+1,v4=(int)floor(v0    +dv2)+1,
		h12=h2-h1,h23=h3-h2,h34=h4-h3,h41=h1-h4,
		v12=v2-v1,v23=v3-v2,v34=v4-v3,v41=v1-v4;
	double	dh,dv;

	hmin=hmax=h1; if (h2<hmin) hmin=h2; if (h2>hmax) hmax=h2;
		      if (h3<hmin) hmin=h3; if (h3>hmax) hmax=h3;
		      if (h4<hmin) hmin=h4; if (h4>hmax) hmax=h4;
	vMin=vMax=v1; if (v2<vMin) vMin=v2; if (v2>vMax) vMax=v2;
		      if (v3<vMin) vMin=v3; if (v3>vMax) vMax=v3;
		      if (v4<vMin) vMin=v4; if (v4>vMax) vMax=v4;

	for (v=vMin; v<=vMax; v++) { hMin[v]=hmax; hMax[v]=hmin; }

	ScanEdge(h1,v1,h2,v2,hMin,hMax);
	ScanEdge(h2,v2,h3,v3,hMin,hMax);
	ScanEdge(h3,v3,h4,v4,hMin,hMax);
	ScanEdge(h4,v4,h1,v1,hMin,hMax);

	for (dv=c+b*(double)((v=vMin)-1),
	     v1=h12*(v-v1)+v12*h1,
	     v2=h23*(v-v2)+v23*h2,
	     v3=h34*(v-v3)+v34*h3,
	     v4=h41*(v-v4)+v41*h4; v<=vMax; v1+=h12,
					    v2+=h23,
					    v3+=h34,
					    v4+=h41, dv+=b, v++)
	for (dh=dv+a*(double)((h=hMin[v])-1),
	     h1=v1-v12*h,
	     h2=v2-v23*h,
	     h3=v3-v34*h,
	     h4=v4-v41*h; h<=hMax[v]; h1-=v12,
				      h2-=v23,
				      h3-=v34,
				      h4-=v41, dh+=a, h++)
	    if ((h1>=0 && h2>=0 && h3>=0 && h4>=0) ||
		(h1<=0 && h2<=0 && h3<=0 && h4<=0))
		depth[v][h]=((d=(int)floor(dh))>1)?(Word)d:1;
}

#define NH	bev.Nh
#define NV	bev.Nv
#define ND	bev.Nd
#define HX	bev.hx
#define HY	bev.hy
#define HZ	bev.hz
#define VX	bev.vx
#define VY	bev.vy
#define VZ	bev.vz
#define DX	bev.dx
#define DY	bev.dy
#define DZ	bev.dz
#define H0	bev.h0
#define V0	bev.v0
#define D0	bev.d0

typedef struct {
		int	Nh,Nv;
		Word	**depth;
		Cell	**label;
		Byte	**shade,**edge;
	} Map;

#define DEPTH	(map.depth)
#define LABEL	(map.label)
#define SHADE	(map.shade)
#define EDGE	(map.edge)

#define PP(c,h0,v0,depth)	\
	PaintParallelogram(a,b,c,h0,v0,dh1,dv1,dh2,dv2,hMin,hMax,depth)

static Map	BirdEyeViewMapper(Nx,Ny,Nz,cell,Sx,Sy,Sz,vLo,vLa,lLo,lLa,
				  x0,y0,z0,lo,la)
int		Nx,Ny,Nz;
Cell		***cell;
double		Sx,Sy,Sz,vLo,vLa,lLo,lLa,x0,y0,z0,lo,la;
{
	BEV	bev;
	Map	map;
	int	*hMin,*hMax,x,y,z,h,v;
	double	hx,hy,hz,vx,vy,vz,dx,dy,dz,
		xx,yy,zz,g,a,b,h0,v0,dh1,dv1,dh2,dv2,h1,v1,h2,v2,c1,c2,
		nx,ny,nz,nh,nv,nd,dh,dv,
		hh,vv,dd,xh,yh,zh,xv,yv,zv,xd,yd,zd,lx,ly,lz,ip;
	Word	**depth,d1,d2,d,dE,dW,dN,dS;

	bev=SetUpBEV(Sx*(-0.5),
		     Sy*(-0.5),
		     Sz*(-0.5),Sx*((double)Nx-0.5),
			       Sy*((double)Ny-0.5),
			       Sz*((double)Nz-0.5),vLo,vLa);

	if (ND>(1<<(8*sizeof(Word)))) Error("too large image depth.");

	map.Nh=NH;
	map.Nv=NV; DEPTH=AllocateWordMemory(NH,NV);
		   LABEL=AllocateCellMemory(NH,NV);
		   SHADE=AllocateByteMemory(NH,NV);
		   EDGE =AllocateByteMemory(NH,NV);

	depth=AllocateWordMemory(NH,NV);

	if ((hMin=(int *)MA(NV,hMin))==NULL ||
	    (hMax=(int *)MA(NV,hMax))==NULL)
	    Error("raster scan memory allocation error.");

	hx=Sx*HX; vx=Sx*VX; dx=Sx*DX;
	hy=Sy*HY; vy=Sy*VY; dy=Sy*DY;
	hz=Sz*HZ; vz=Sz*VZ; dz=Sz*DZ;

	if (DX!=0.0 && (g=HY*VZ-HZ*VY)!=0.0) {
	    a=(VZ*DY-VY*DZ)/g; b=(DZ*HY-DY*HZ)/g;
	    dh1=(double)Ny*hy; dv1=(double)Ny*vy;
	    dh2=(double)Nz*hz; dv2=(double)Nz*vz;
	    h0=(-0.5)*(hy+hz); v0=(-0.5)*(vy+vz);
	    c1=(-0.5)*dx-D0-a*(h1=(-0.5)*hx-H0)-b*(v1=(-0.5)*vx-V0)+1.0;
	    xx=(double)Nx-0.5; c2=xx*dx-D0-a*(h2=xx*hx-H0)-b*(v2=xx*vx-V0)+1.0;
	    if (DX<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	if (DY!=0.0 && (g=HZ*VX-HX*VZ)!=0.0) {
	    a=(VX*DZ-VZ*DX)/g; b=(DX*HZ-DZ*HX)/g;
	    dh1=(double)Nz*hz; dv1=(double)Nz*vz;
	    dh2=(double)Nx*hx; dv2=(double)Nx*vx;
	    h0=(-0.5)*(hz+hx); v0=(-0.5)*(vz+vx);
	    c1=(-0.5)*dy-D0-a*(h1=(-0.5)*hy-H0)-b*(v1=(-0.5)*vy-V0)+1.0;
	    yy=(double)Ny-0.5; c2=yy*dy-D0-a*(h2=yy*hy-H0)-b*(v2=yy*vy-V0)+1.0;
	    if (DY<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	if (DZ!=0.0 && (g=HX*VY-HY*VX)!=0.0) {
	    a=(VY*DX-VX*DY)/g; b=(DY*HX-DX*HY)/g;
	    dh1=(double)Nx*hx; dv1=(double)Nx*vx;
	    dh2=(double)Ny*hy; dv2=(double)Ny*vy;
	    h0=(-0.5)*(hx+hy); v0=(-0.5)*(vx+vy);
	    c1=(-0.5)*dz-D0-a*(h1=(-0.5)*hz-H0)-b*(v1=(-0.5)*vz-V0)+1.0;
	    zz=(double)Nz-0.5; c2=zz*dz-D0-a*(h2=zz*hz-H0)-b*(v2=zz*vz-V0)+1.0;
	    if (DZ<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	free(hMin); free(hMax);

	for (v=NV-2; v>0; v--)
	for (h=NH-2; h>0; h--)
	    if ((d1=DEPTH[v][h])>(d2=depth[v][h]) && d2==0)
		depth[v][h]=d1;
	    else if (d1<d2)
		DEPTH[v][h]=d2;

	LightDirection(lo,la,&nx,&ny,&nz); nh=HX*nx+HY*ny+HZ*nz;
					   nv=VX*nx+VY*ny+VZ*nz;
					   nd=DX*nx+DY*ny+DZ*nz;

	x0*=Sx;
	y0*=Sy;
	z0*=Sz; dh=(double)(NH-2)-0.5+H0-HX*x0-HY*y0-HZ*z0;
		dv=(double)(NV-2)-0.5+V0-VX*x0-VY*y0-VZ*z0;

	if (nd==0.0) {
	    for (vv=nh*dh+nv*dv, v=NV-2; v>0; v--, vv-=nv)
	    for (hh=vv,          h=NH-2; h>0; h--, hh-=nh)
		if (hh>0.0) DEPTH[v][h]=0;
	}
	else {
	    a=nh/nd;
	    b=nv/nd; vv=DX*x0+DY*y0+DZ*z0-D0+1.0-a*dh-b*dv;
	    if (nd>0.0) {
		for (       v=NV-2; v>0; v--, vv+=b)
		for (hh=vv, h=NH-2; h>0; h--, hh+=a)
		    if (DEPTH[v][h]>0)
			if (hh<(double)depth[v][h])
			    DEPTH[v][h]=0;
			else if (hh<(double)(DEPTH[v][h]+1)) {
			    DEPTH[v][h]=(Word)floor(hh); EDGE[v][h]=2;
			}
	    }
	    else
		for (       v=NV-2; v>0; v--, vv+=b)
		for (hh=vv, h=NH-2; h>0; h--, hh+=a)
		    if (DEPTH[v][h]>0)
			if (hh>=(double)(DEPTH[v][h]+1))
			    DEPTH[v][h]=0;
			else if (hh>=(double)depth[v][h]) {
			    depth[v][h]=(Word)floor(hh); EDGE[v][h]=2;
			}

	    for (v=NV-2; v>0; v--)
	    for (h=NH-2; h>0; h--)
		if (EDGE[v][h]==2 && (EDGE[v][h-1]==0 ||
				      EDGE[v][h+1]==0 ||
				      EDGE[v-1][h]==0 ||
				      EDGE[v+1][h]==0)) EDGE[v][h]|=1;
	}

	hx=HX/Sx; hy=HY/Sy; hz=HZ/Sz; hh=H0+(double)(NH-2)-0.5;
	vx=VX/Sx; vy=VY/Sy; vz=VZ/Sz; vv=V0+(double)(NV-2)-0.5;
	dx=DX/Sx; dy=DY/Sy; dz=DZ/Sz; dd=D0               -1.0;

	for (xv=hx*hh+vx*vv+dx*dd+0.5,
	     yv=hy*hh+vy*vv+dy*dd+0.5,
	     zv=hz*hh+vz*vv+dz*dd+0.5,
	     v=NV-2; v>0; v--, xv-=vx, yv-=vy, zv-=vz) {
	    for (xh=xv, yh=yv, zh=zv, h=NH-2; h>0; h--, xh-=hx, yh-=hy, zh-=hz)
		if ((d1=DEPTH[v][h])>0) {
		    DEPTH[v][h]=0;
		    for (xd=xh+dx*(double)d1,
			 yd=yh+dy*(double)d1,
			 zd=zh+dz*(double)d1,
			 d2=depth[v][h],
			 d=d1; d>=d2; d--, xd-=dx, yd-=dy, zd-=dz)
			if ((x=(int)floor(xd))>=0 && x<Nx &&
			    (y=(int)floor(yd))>=0 && y<Ny &&
			    (z=(int)floor(zd))>=0 && z<Nz && cell[z][y][x]) {
			    DEPTH[v][h]=d; LABEL[v][h]=cell[z][y][x]; break;
			}

		    if (nd>0.0 && d< d1) EDGE[v][h]&=1;
		    if (nd<0.0 && d>=d2) EDGE[v][h] =0;
		}

	    free(depth[v]);
	}
	free(depth);

	LightDirection(lLo,lLa,&lx,&ly,&lz);

	for (v=NV-2; v>0; v--)
	for (h=NH-2; h>0; h--)
	    if (DEPTH[v][h]) {
		dh=((double)(dE=DEPTH[v][h+1])-(double)(dW=DEPTH[v][h-1]))/2.0;
		dv=((double)(dN=DEPTH[v+1][h])-(double)(dS=DEPTH[v-1][h]))/2.0;

		if ((ip=(nx=DX-HX*dh-VX*dv)*lx+
			(ny=DY-HY*dh-VY*dv)*ly+
			(nz=DZ-HZ*dh-VZ*dv)*lz)>0.0)
		    SHADE[v][h]=(Byte)(255.0*ip/sqrt(nx*nx+ny*ny+nz*nz));

		if (nd<=0.0) {
		    if (dE==0 || dW==0 || dN==0 || dS==0) EDGE[v][h]=1;
		}
		else
		    if (dE==0 || dW==0 || dN==0 || dS==0 ||
		       (EDGE[v][h]==2 && (EDGE[v][h+1]<2 ||
					  EDGE[v][h-1]<2 ||
					  EDGE[v+1][h]<2 ||
					  EDGE[v-1][h]<2))) EDGE[v][h]|=1;
	    }

	return map;
}

#define PI	((bps==1)?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK)
#define SPP	1
#define PC	PLANARCONFIG_CONTIG
#define TSF	TIFFSetField

static TIFF	*OpenTIFF(path,wid,hig,bps,desc,tsl)
char		*path,*desc;
int		wid,hig,bps;
uint8		**tsl;
{
	TIFF	*tiff=TIFFOpen(path,"w");
	int	tss;

	if (tiff==NULL ||
	    !TSF(tiff,TIFFTAG_IMAGEWIDTH,     (uint32)wid) ||
	    !TSF(tiff,TIFFTAG_IMAGELENGTH,    (uint32)hig) ||
	    !TSF(tiff,TIFFTAG_PHOTOMETRIC,    (uint16)PI ) ||
	    !TSF(tiff,TIFFTAG_BITSPERSAMPLE,  (uint16)bps) ||
	    !TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)SPP) ||
	    !TSF(tiff,TIFFTAG_PLANARCONFIG,   (uint16)PC ) ||
	    (*tsl=(uint8 *)malloc(tss=TIFFScanlineSize(tiff)))==NULL) {
	    fprintf(stderr,"%s : ",path); Error("image file not stored.");
	}
	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(uint32)((tss>8192)?1:8192/tss));
#ifdef	LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	TSF(tiff,TIFFTAG_IMAGEDESCRIPTION,desc);

	return tiff;
}

#define WriteTIFF(tiff,tsl,v,path)	\
	if (TIFFWriteScanline(tiff,tsl,(uint32)v,(uint16)0)<0) { \
	    fprintf(stderr,"%s : ",path); Error("write data error."); \
	}

static void	StoreWordImage(path,word,Nh,Nv,bps,desc)
char		*path,*desc;
Word		**word;
int		Nh,Nv,bps;
{
	int	v,h;
	Word	*data,mask,msb=(Word)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=word[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(word[v]);
	}
	free(word); free(tsl); TIFFClose(tiff);
}

static void	StoreCellImage(path,cell,Nh,Nv,bps,desc)
char		*path,*desc;
Cell		**cell;
int		Nh,Nv,bps;
{
	int	v,h;
	Cell	*data,mask,msb=(Cell)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=cell[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(cell[v]);
	}
	free(cell); free(tsl); TIFFClose(tiff);
}

static void	StoreByteImage(path,byte,Nh,Nv,bps,desc)
char		*path,*desc;
Byte		**byte;
int		Nh,Nv,bps;
{
	int	v,h;
	Byte	*data,mask,msb=(Byte)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=byte[v]; ptr=tsl;
	    if (bps==8)
		for (h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else
		for (bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(byte[v]);
	}
	free(byte); free(tsl); TIFFClose(tiff);
}

#define NAME_LEN	512
#define DESC_LEN	256
#define PATH_LEN	512

#define RADIAN	(M_PI/180.0)

#define BPS(limit)	\
	(limit<2)?1:(limit<4)?2:(limit<16)?4:(limit<256)?8:(limit<65536)?16:32

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,bps,h,v,limit;
	char	**names,name[NAME_LEN],desc[DESC_LEN],path[PATH_LEN];
	double	Sx,Sy,Sz,vLo,vLa,lLo,lLa,x0,y0,z0,lo,la;
	Cell	***cell;
	Map	map;

	if (argc!=6)
	    Error("usage : bevmHSC directory nameFile scaleX scaleY scaleZ");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,NULL);

	if ((Sx=atof(argv[3]))<=0.0 ||
	    (Sy=atof(argv[4]))<=0.0 ||
	    (Sz=atof(argv[5]))<=0.0) Error("bad scaling factor.");

	cell=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	while (scanf("%lf %lf %lf %lf %lf %lf %lf %lf %lf %s",
		     &vLo,&vLa,&lLo,&lLa,&x0,&y0,&z0,&lo,&la,name)==10) {
	    map=BirdEyeViewMapper(Nx,Ny,Nz,cell,Sx,Sy,Sz,
				  vLo*RADIAN,vLa*RADIAN,lLo*RADIAN,lLa*RADIAN,
				  x0,y0,z0,lo*RADIAN,la*RADIAN);

	    sprintf(desc,
		    "%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\n",
		    -0.5,-0.5,-0.5,(double)Nx-0.5,(double)Ny-0.5,(double)Nz-0.5,
		    Sx,Sy,Sz,vLo,vLa);

	    limit=0;
	    for (v=0; v<map.Nv; v++)
	    for (h=0; h<map.Nh; h++)
		if ((int)DEPTH[v][h]>limit) limit=(int)DEPTH[v][h];

	    sprintf(path,"%s.%s",name,EXTENSION_DEPTH);
	    StoreWordImage(path,DEPTH,map.Nh,map.Nv,BPS(limit),desc);

	    sprintf(path,"%s.%s",name,EXTENSION_LABEL);
	    StoreCellImage(path,LABEL,map.Nh,map.Nv,bps,       desc);

	    sprintf(path,"%s.%s",name,EXTENSION_SHADE);
	    StoreByteImage(path,SHADE,map.Nh,map.Nv,8,         desc);

	    limit=0;
	    for (v=0; v<map.Nv; v++)
	    for (h=0; h<map.Nh; h++)
		if ((int)EDGE[v][h]>limit) limit=(int)EDGE[v][h];

	    sprintf(path,"%s.%s",name,EXTENSION_EDGE );
	    StoreByteImage(path,EDGE, map.Nh,map.Nv,BPS(limit),desc);
	}
	return 0;
}
