
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

extern void	Error();

static void	Abort(path,msg)
char		*path,*msg;
{
	fprintf(stderr,"%s : TIFF ",path); Error(msg);
}

static void	ScanTIFF(path,desc,image,cm)
char		*path,**desc;
Image		**image;
CM		**cm;
{
	TIFF	*tiff=TIFFOpen(path,"r");
	char	*ptr;
	int	code;

	if (tiff==NULL) Abort(path,"file not found.");

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    Abort(path,"image description not found.");

	if ((*desc=(char *)malloc(strlen(ptr)+1))==NULL)
	    Abort(path,"image description memory allocation error.");

	(void)strcpy(*desc,ptr);

	if ((code=ReadTIFF(tiff,image,cm))>0) Abort(path,ReadTIFFError(code));

	TIFFClose(tiff);
}

static void	ScanAndCheckTIFF(path,image,cm,desc,xSize,ySize)
char		*path,*desc;
Image		**image;
CM		**cm;
int		xSize,ySize;
{
	char	*ptr;

	ScanTIFF(path,&ptr,image,cm);

	if (strcmp(desc,ptr)) Abort(path,"image description not match.");

	free(ptr);

	if ((*image)->xSize!=xSize ||
	    (*image)->ySize!=ySize) Abort(path,"image size not match.");
}

static void	CM2RGB(image,cm)
Image		*image;
CM		*cm;
{
	int	z,x,y;

	for (z=0; z<cm->pixels; z++) {
	    cm->r[z]>>=8;
	    cm->g[z]>>=8;
	    cm->b[z]>>=8;
	}
	for (x=0; x<image->xSize; x++)
	for (y=0; y<image->ySize; y++) {
	    z=GETPIXEL(image,x,y); image->r[x][y]=(byte)cm->r[z];
				   image->g[x][y]=(byte)cm->g[z];
				   image->b[x][y]=(byte)cm->b[z];
	}
	FreeCM(cm);
}

static void	StoreTIFF(path,desc,image,cm)
char		*path,*desc;
Image		*image;
CM		*cm;
{
	TIFF	*tiff=TIFFOpen(path,"w");
	int	code;

	if (tiff==NULL) Abort(path,"file not stored.");

	if (!TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,desc))
	    Abort(path,"image description not stored.");

	if ((code=WriteTIFF(tiff,image,cm))>0) Abort(path,WriteTIFFError(code));

	TIFFClose(tiff);
}

#define XSIZE	image0->xSize
#define YSIZE	image0->ySize

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	*desc,*bit="bad image type.";
	Image	*image0,*image1,*image2;
	CM	*cm0,*cm1,*cm2;
	int	x,y,z,gray;

	if (argc<4 || argc%3!=1)
	    Error("usage : bevmIS depth[3] {other[3] ...}");

	ScanTIFF(*(++argv),&desc,&image0,&cm0);
	if (cm0==NULL) Abort(*argv,bit);

	ScanAndCheckTIFF(*(++argv),&image1,&cm1,desc,XSIZE,YSIZE);
	if (cm1==NULL) Abort(*argv,bit);

	for (x=0; x<XSIZE; x++)
	for (y=0; y<YSIZE; y++)
	    if ((z=GETPIXEL(image0,x,y))>=GETPIXEL(image1,x,y)) {
		PUTPIXEL(image1,x,y,z); PUTPIXEL(image0,x,y,1);
	    }
	    else
		PUTPIXEL(image0,x,y,0);

	StoreTIFF(*(++argv),desc,image1,(cm0->pixels>cm1->pixels)?cm0:cm1);
	FreeCM(cm1); FreeImage(image1); FreeCM(cm0);

	while ((argc-=3)>=4) {
	    ScanAndCheckTIFF(*(++argv),&image1,&cm1,desc,XSIZE,YSIZE);
	    ScanAndCheckTIFF(*(++argv),&image2,&cm2,desc,XSIZE,YSIZE);

	    if (cm1==NULL || cm2==NULL) {
		if (cm1!=NULL) CM2RGB(image1,cm1); 
		if (cm2!=NULL) CM2RGB(image2,cm2); cm0=NULL;
	    }
/*
	    else
		if (cm1->pixels!=cm2->pixels) {
		    CM2RGB(image1,cm1);
		    CM2RGB(image2,cm2); cm0=NULL;
		}
		else {
		    for (z=0; z<cm1->pixels; z++)
			if (cm1->r[z]!=cm2->r[z] ||
			    cm1->g[z]!=cm2->g[z] ||
			    cm1->b[z]!=cm2->b[z]) break;

		    if (z<cm1->pixels) {
			CM2RGB(image1,cm1);
			CM2RGB(image2,cm2); cm0=NULL;
		    }
		    else {
			FreeCM(cm2); cm0=cm1;
		    }
		}
*/
	    else
		if (cm1->pixels>cm2->pixels) {
		    FreeCM(cm2); cm0=cm1;
		}
		else {
		    FreeCM(cm1); cm0=cm2;
		}

	    if (cm0==NULL) {
		gray=1;
		for (x=0; x<XSIZE; x++)
		for (y=0; y<YSIZE; y++) {
		    if (GETPIXEL(image0,x,y)) {
			image2->r[x][y]=image1->r[x][y];
			image2->g[x][y]=image1->g[x][y];
			image2->b[x][y]=image1->b[x][y];
		    }
		    if (gray) gray=(image2->r[x][y]==image2->g[x][y] &&
				    image2->r[x][y]==image2->b[x][y]);
		}
		if (!gray)
		    StoreTIFF(*(++argv),desc,image2,NULL);
		else {
		    for (x=0; x<XSIZE; x++)
		    for (y=0; y<YSIZE; y++) {
			z=(int)image2->r[x][y]; PUTPIXEL(image2,x,y,z);
		    }

		    if ((cm0=AllocCM(256))==NULL)
			Error("colormap memory allocation error.");

		    for (z=0; z<256; z++)
			cm0->r[z]=cm0->g[z]=cm0->b[z]=(word)((z<<8)|z);

		    StoreTIFF(*(++argv),desc,image2,cm0); FreeCM(cm0);
	        }
	    }
	    else {
		for (x=0; x<XSIZE; x++)
		for (y=0; y<YSIZE; y++)
		    if (GETPIXEL(image0,x,y)) {
			z=GETPIXEL(image1,x,y); PUTPIXEL(image2,x,y,z);
		    }

		StoreTIFF(*(++argv),desc,image2,cm0); FreeCM(cm0);
	    }
	    FreeImage(image2); FreeImage(image1);
	}
	FreeImage(image0); return 0;
}
