
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"
#include "bev.h"

extern void	Error();

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

static byte	**AllocateMaskMemory(wid,hig)
int		wid,hig;
{
	byte	**mask;
	int	x,y;
	char	*msg="mask memory allocation error.";

	if ((mask=(byte **)MA(wid,mask))==NULL) Error(msg);

	for (x=0; x<wid; x++) {
	    if ((mask[x]=(byte *)MA(hig,*mask))==NULL) Error(msg);

	    for (y=0; y<hig; y++) mask[x][y]=0;
	}
	return mask;
}

#define PlotPoint(h,v,d)	\
	if (h>=0 && h<depth->xSize && \
	    v>=0 && v<depth->ySize && (d)>=GETPIXEL(depth,h,v)) mask[h][v]=1

static void	DrawLine(depth,bev,Sx,Sy,Sz,mask)
Image		*depth;
BEV		bev;
double		Sx,Sy,Sz;
byte		**mask;
{
	int	cnt,idx,h2,v2,d2,wh,sh,wv,sv,m,p,e,h,v,h1=0,v1=0,d1=0;
	double	x,y,z,dd,d;

	while (scanf("%d",&cnt)==1 && cnt!=0) {
	    if (cnt<0) cnt=(-cnt);

	    for (idx=0; idx<cnt; idx++, h1=h2, v1=v2, d1=d2) {
		(void)scanf("%lf %lf %lf",&x,&y,&z);

		x*=Sx; y*=Sy; z*=Sz;

		h2=(int)floor(x*bev.hx+y*bev.hy+z*bev.hz-bev.h0)+1;
		v2=(int)floor(x*bev.vx+y*bev.vy+z*bev.vz-bev.v0)+1;
		d2=(int)floor(x*bev.dx+y*bev.dy+z*bev.dz-bev.d0)+1;

		if (idx>0)
		    if (h1==h2 && v1==v2) {
			PlotPoint(h1,v1,(d1>d2)?d1:d2);
		    }
		    else {
			if (h2>h1) { wh=h2-h1; sh=  1 ; } else
				   { wh=h1-h2; sh=(-1); }
			if (v2>v1) { wv=v2-v1; sv=  1 ; } else
				   { wv=v1-v2; sv=(-1); }
	    		dd=(double)(d2-d1); d=(double)d1;
			if (wh>=wv) {
			    m=2*wv; e=m-wh; p=e-wh; dd/=(double)wh;
			    for (v=v1, h=h1; h!=h2; h+=sh, d+=dd) {
				PlotPoint(h,v,(int)ceil(d));
				if (e>=0) { v+=sv; e+=p; } else e+=m;
			    }
			}
			else {
			    m=2*wh; e=m-wv; p=e-wv; dd/=(double)wv;
			    for (h=h1, v=v1; v!=v2; v+=sv, d+=dd) {
				PlotPoint(h,v,(int)ceil(d));
				if (e>=0) { h+=sh; e+=p; } else e+=m;
			    }
			}
		    }
	    }
	    PlotPoint(h1,v1,d1);
	}
}

static void	CopyMaskOnDepth(depth,mask)
Image		*depth;
byte		**mask;
{
	int	x,y;

	for (x=0; x<depth->xSize; x++)
	for (y=0; y<depth->ySize; y++) PUTPIXEL(depth,x,y,(int)mask[x][y]);
}

#define DESC_LEN	256

#define BAD_FILE(arg,msg)	\
	{ fprintf(stderr,"%s : TIFF ",argv[arg]); Error(msg); }

#define RAD	(M_PI/180.0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	char	*ptr,desc[DESC_LEN];
	int	idx,code;
	double	xMin,yMin,zMin,xMax,yMax,zMax,Sx,Sy,Sz,vLo,vLa;
	BEV	bev;
	Image	*depth;
	CM	*cm;
	byte	**mask;

	if (argc!=3) Error("usage : bevmLD depthTIFF maskTIFF");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) BAD_FILE(1,"file not found.");

	if (!TIFFGetField(tiff,TIFFTAG_IMAGEDESCRIPTION,&ptr))
	    BAD_FILE(1,"image description not found.");

	for (idx=0; (desc[idx]=(*ptr))!='\0'; idx++) ++ptr;

	if (sscanf(desc,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
			&xMin,&yMin,&zMin,&xMax,&yMax,&zMax,&Sx,&Sy,&Sz,
			&vLo,&vLa)!=11)
	    BAD_FILE(1,"bad BEVM parameter numbers.");

	if ((code=ReadTIFF(tiff,&depth,&cm))>0) BAD_FILE(1,ReadTIFFError(code));

	TIFFClose(tiff);

	if (cm==NULL) BAD_FILE(1,"image type not match.");

	bev=SetUpBEV(Sx*xMin,Sy*yMin,Sz*zMin,
		     Sx*xMax,Sy*yMax,Sz*zMax,vLo*RAD,vLa*RAD);

	if (bev.Nh!=depth->xSize ||
	    bev.Nv!=depth->ySize) BAD_FILE(1,"image size not match.");

	mask=AllocateMaskMemory(depth->xSize,depth->ySize);

	DrawLine(depth,bev,Sx,Sy,Sz,mask);

	CopyMaskOnDepth(depth,mask);

	cm->pixels=2; cm->r[0]=cm->g[0]=cm->b[0]=0xffff;
		      cm->r[1]=cm->g[1]=cm->b[1]=0;

	if ((tiff=TIFFOpen(argv[2],"w"))==NULL) BAD_FILE(2,"file not created.");

	if (!TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,desc))
	    BAD_FILE(2,"image description not stored.");

	if ((code=WriteTIFF(tiff,depth,cm))>0) BAD_FILE(2,WriteTIFFError(code));

	TIFFClose(tiff);

	return 0;
}
