
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "slice.h"
#include "bev.h"

#define EXTENSION_DEPTH	"od.tif"
#define EXTENSION_LABEL	"ol.tif"
#define EXTENSION_SHADE	"os.tif"
#define EXTENSION_EDGE	"oe.tif"

static double	OvoidExtentXYZ(A,B,C)
double		A,B,C;
{
	return sqrt(A*A+B*B+C*C);
}

typedef struct Ovoid {
		Cell		label;
		double		x0,y0,z0,ax,ay,az,A,
					 bx,by,bz,B,
					 cx,cy,cz,C,x1,y1,z1,x2,y2,z2;
		struct Ovoid	*next;
	} Ovoid;

#define X0	ovoid->x0
#define Y0	ovoid->y0
#define Z0	ovoid->z0
#define AX	ovoid->ax
#define AY	ovoid->ay
#define AZ	ovoid->az
#define BX	ovoid->bx
#define BY	ovoid->by
#define BZ	ovoid->bz
#define CX	ovoid->cx
#define CY	ovoid->cy
#define CZ	ovoid->cz
#define OA	ovoid->A
#define OB	ovoid->B
#define OC	ovoid->C
#define X1	ovoid->x1
#define Y1	ovoid->y1
#define Z1	ovoid->z1
#define X2	ovoid->x2
#define Y2	ovoid->y2
#define Z2	ovoid->z2

static Ovoid	*SetUpOvoid(label,x0,y0,z0,lon,lat,rot,A,B,C)
int		label;
double		x0,y0,z0,lon,lat,rot,A,B,C;
{
	Ovoid	*ovoid=(Ovoid *)malloc(sizeof(Ovoid));
	double	d,clon=cos(lon),clat=cos(lat),crot=cos(rot),
		  slon=sin(lon),slat=sin(lat),srot=sin(rot);

	if (ovoid==NULL) Error("ovoid memory allocation error.");

	X0=x0; AX=  slon*crot+clon*slat*srot ;
	       AY=(-clon*crot+slon*slat*srot);
	       AZ=(               -clat*srot); OA=A;
	Y0=y0; BX=(-slon*srot+clon*slat*crot);
	       BY=  clon*srot+slon*slat*crot ;
	       BZ=(               -clat*crot); OB=B;
	Z0=z0; CX=            clon*clat      ;
	       CY=            slon*clat      ;
	       CZ=                 slat      ; OC=C;

	d=OvoidExtentXYZ(A*AX,B*BX,C*CX); X1=x0-d; X2=x0+d;
	d=OvoidExtentXYZ(A*AY,B*BY,C*CY); Y1=y0-d; Y2=y0+d;
	d=OvoidExtentXYZ(A*AZ,B*BZ,C*CZ); Z1=z0-d; Z2=z0+d;

	ovoid->label=(Cell)label; ovoid->next=NULL; return ovoid;
}

#define RADIAN	(M_PI/180.0)

static Ovoid	*ReadOvoid(x1,y1,z1,x2,y2,z2)
double		*x1,*y1,*z1,*x2,*y2,*z2;
{
	int	label;
	double	x0,y0,z0,lon,lat,rot,A,B,C;
	Ovoid	*ovoid,*top=NULL,**ptr=(&top);

	while (scanf("%d %lf %lf %lf %lf %lf %lf %lf %lf %lf",
		     &label,&x0,&y0,&z0,&lon,&lat,&rot,&A,&B,&C)==10) {
	    if (label<=0 || label>=1<<(8*sizeof(Cell))) {
		fputs("bad ovoid label.\n",stderr); continue;
	    }
	    if (A<=0.0 || B<=0.0 || C<=0.0) {
		fputs("bad ovoid radii.\n",stderr); continue;
	    }
	    ovoid=SetUpOvoid(label,
			     x0,y0,z0,lon*RADIAN,lat*RADIAN,rot*RADIAN,A,B,C);

	    if (top==NULL) {
		*x1=X1; *x2=X2; *y1=Y1; *y2=Y2; *z1=Z1; *z2=Z2;
	    }
	    else {
		if (*x1>X1) *x1=X1; if (*x2<X2) *x2=X2;
		if (*y1>Y1) *y1=Y1; if (*y2<Y2) *y2=Y2;
		if (*z1>Z1) *z1=Z1; if (*z2<Z2) *z2=Z2;
	    }
	    *ptr=ovoid; ptr=(&(ovoid->next));
	}
	return top;
}

typedef unsigned short	Word;
typedef unsigned char	Byte;

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

static Word	**AllocateWordMemory(wid,hig)
int		wid,hig;
{
	Word	**word;
	int	x,y;
	char	*msg="word memory allocation error.";

	if ((word=(Word **)MA(hig,word))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((word[y]=(Word *)MA(wid,*word))==NULL) Error(msg);

	    for (x=0; x<wid; x++) word[y][x]=0;
	}
	return word;
}

static Cell	**AllocateCellMemory(wid,hig)
int		wid,hig;
{
	Cell	**cell;
	int	x,y;
	char	*msg="cell memory allocation error.";

	if ((cell=(Cell **)MA(hig,cell))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((cell[y]=(Cell *)MA(wid,*cell))==NULL) Error(msg);

	    for (x=0; x<wid; x++) cell[y][x]=0;
	}
	return cell;
}

static Byte	**AllocateByteMemory(wid,hig)
int		wid,hig;
{
	Byte	**byte;
	int	x,y;
	char	*msg="byte memory allocation error.";

	if ((byte=(Byte **)MA(hig,byte))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((byte[y]=(Byte *)MA(wid,*byte))==NULL) Error(msg);

	    for (x=0; x<wid; x++) byte[y][x]=0;
	}
	return byte;
}

#define NH	bev.Nh
#define NV	bev.Nv
#define ND	bev.Nd
#define HX	bev.hx
#define HY	bev.hy
#define HZ	bev.hz
#define VX	bev.vx
#define VY	bev.vy
#define VZ	bev.vz
#define DX	bev.dx
#define DY	bev.dy
#define DZ	bev.dz
#define H0	bev.h0
#define V0	bev.v0
#define D0	bev.d0

#define MINMAX(exp)	hv=exp; if (hv<min) min=hv; if (hv>max) max=hv

static void	FillOvoid(ovoid,scale,bev,lx,ly,lz,depth,label,shade)
Ovoid		*ovoid;
double		scale,lx,ly,lz;
BEV		bev;
Word		**depth;
Cell		**label;
Byte		**shade;
{
	double	hv,min,max,
		h,v,xhv,yhv,zhv,ahv1,bhv1,chv1,ahv2,bhv2,chv2,Q,R,d,a,b,c,ip,
		h0=X0*HX+Y0*HY+Z0*HZ,
		v0=X0*VX+Y0*VY+Z0*VZ,
		d0=X0*DX+Y0*DY+Z0*DZ,
		ad1=DX*AX+DY*AY+DZ*AZ,ad2=ad1/OA,
		bd1=DX*BX+DY*BY+DZ*BZ,bd2=bd1/OB,
		cd1=DX*CX+DY*CY+DZ*CZ,cd2=cd1/OC,P=ad2*ad2+bd2*bd2+cd2*cd2,
		A2=OA*OA,la=lx*AX+ly*AY+lz*AZ,
		B2=OB*OB,lb=lx*BX+ly*BY+lz*BZ,
		C2=OC*OC,lc=lx*CX+ly*CY+lz*CZ;
	int	i1,i2,j1,j2,i,j,k;

	min=max=X1*HX+Y1*HY+Z1*HZ;
	MINMAX(X2*HX+Y1*HY+Z1*HZ);
	MINMAX(X1*HX+Y2*HY+Z1*HZ);
	MINMAX(X2*HX+Y2*HY+Z1*HZ);
	MINMAX(X1*HX+Y1*HY+Z2*HZ);
	MINMAX(X2*HX+Y1*HY+Z2*HZ);
	MINMAX(X1*HX+Y2*HY+Z2*HZ);
	MINMAX(X2*HX+Y2*HY+Z2*HZ);
	if ((i1=(int)floor(scale*min-H0)+1)<1   ) i1=1;
	if ((i2=(int)ceil( scale*max-H0)+1)>NH-2) i2=NH-2;

	min=max=X1*VX+Y1*VY+Z1*VZ;
	MINMAX(X2*VX+Y1*VY+Z1*VZ);
	MINMAX(X1*VX+Y2*VY+Z1*VZ);
	MINMAX(X2*VX+Y2*VY+Z1*VZ);
	MINMAX(X1*VX+Y1*VY+Z2*VZ);
	MINMAX(X2*VX+Y1*VY+Z2*VZ);
	MINMAX(X1*VX+Y2*VY+Z2*VZ);
	MINMAX(X2*VX+Y2*VY+Z2*VZ);
	if ((j1=(int)floor(scale*min-V0)+1)<1   ) j1=1;
	if ((j2=(int)ceil( scale*max-V0)+1)>NV-2) j2=NV-2;

	for (i=i1; i<=i2; i++) { h=((double)(i-1)+H0)/scale-h0;
	for (j=j1; j<=j2; j++) { v=((double)(j-1)+V0)/scale-v0;
	    xhv=h*HX+v*VX;
	    yhv=h*HY+v*VY;
	    zhv=h*HZ+v*VZ; ahv1=xhv*AX+yhv*AY+zhv*AZ; ahv2=ahv1/OA;
			   bhv1=xhv*BX+yhv*BY+zhv*BZ; bhv2=bhv1/OB;
			   chv1=xhv*CX+yhv*CY+zhv*CZ; chv2=chv1/OC;
	    Q=ad2*ahv2+bd2*bhv2+cd2*chv2;
	    if ((R=Q*Q-P*(ahv2*ahv2+bhv2*bhv2+chv2*chv2-1.0))>=0.0) {
		d=(sqrt(R)-Q)/P;
		if ((k=(int)floor(scale*(d+d0)-D0)+1)>(int)depth[j][i]) {
		    a=(ahv1+ad1*d)/A2;
		    b=(bhv1+bd1*d)/B2;
		    c=(chv1+cd1*d)/C2; ip=(la*a+lb*b+lc*c)/sqrt(a*a+b*b+c*c);
		    shade[j][i]=(ip>0.0)?(Byte)(255.0*ip+0.5):0;
		    label[j][i]=ovoid->label;
		    depth[j][i]=(Word)k;

		}
	    }
	}}
}

typedef struct {
		int	Nh,Nv;
		Word	**depth;
		Cell	**label;
		Byte	**shade,**edge;
	} Map;

#define DEPTH	(map.depth)
#define LABEL	(map.label)
#define SHADE	(map.shade)
#define EDGE	(map.edge)

static Map	OvoidBEVM(ovoid,x1,y1,z1,x2,y2,z2,scale,vLo,vLa,lLo,lLa)
Ovoid		*ovoid;
double		x1,y1,z1,x2,y2,z2,scale,vLo,vLa,lLo,lLa;
{
	BEV	bev;
	Map	map;
	double	lx,ly,lz;
	int	h,v;
	Word	d1,d2;
	Cell	l1,l2;

	bev=SetUpBEV(scale*x1,scale*y1,scale*z1,
		     scale*x2,scale*y2,scale*z2,vLo,vLa);

	if (ND>(1<<(8*sizeof(Cell)))) Error("too large image depth.");

	map.Nh=NH;
	map.Nv=NV; DEPTH=AllocateWordMemory(NH,NV);
		   LABEL=AllocateCellMemory(NH,NV);
		   SHADE=AllocateByteMemory(NH,NV);
		   EDGE =AllocateByteMemory(NH,NV);

	LightDirection(lLo,lLa,&lx,&ly,&lz);

	do
	    FillOvoid(ovoid,scale,bev,lx,ly,lz,DEPTH,LABEL,SHADE);
	while ((ovoid=ovoid->next)!=NULL);

	for (v=0; v<NV; v++) for (d1=0, l1=0, h=0; h<NH; h++, d1=d2, l1=l2) {
	    d2=DEPTH[v][h]; l2=LABEL[v][h];
	    if (l1!=l2)
		if (l1 && l2)
		    if (d1>=d2) EDGE[v][h-1]=1; else EDGE[v][h]=1;
		else
		    if (l1) EDGE[v][h-1]=1; else if (l2) EDGE[v][h]=1;
	}
	for (h=0; h<NH; h++) for (d1=0, l1=0, v=0; v<NV; v++, d1=d2, l1=l2) {
	    d2=DEPTH[v][h]; l2=LABEL[v][h];
	    if (l1!=l2)
		if (l1 && l2)
		    if (d1>=d2) EDGE[v-1][h]=1; else EDGE[v][h]=1;
		else
		    if (l1) EDGE[v-1][h]=1; else if (l2) EDGE[v][h]=1;
	}
	return map;
}

#define PI	((bps==1)?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK)
#define SPP	1
#define PC	PLANARCONFIG_CONTIG
#define TSF	TIFFSetField

static TIFF	*OpenTIFF(path,wid,hig,bps,desc,tsl)
char		*path,*desc;
int		wid,hig,bps;
uint8		**tsl;
{
	TIFF	*tiff=TIFFOpen(path,"w");
	int	tss;

	if (tiff==NULL ||
	    !TSF(tiff,TIFFTAG_IMAGEWIDTH,     (uint32)wid) ||
	    !TSF(tiff,TIFFTAG_IMAGELENGTH,    (uint32)hig) ||
	    !TSF(tiff,TIFFTAG_PHOTOMETRIC,    (uint16)PI ) ||
	    !TSF(tiff,TIFFTAG_BITSPERSAMPLE,  (uint16)bps) ||
	    !TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)SPP) ||
	    !TSF(tiff,TIFFTAG_PLANARCONFIG,   (uint16)PC ) ||
	    (*tsl=(uint8 *)malloc(tss=TIFFScanlineSize(tiff)))==NULL) {
	    fprintf(stderr,"%s : ",path); Error("image file not stored.");
	}
	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(uint32)((tss>8192)?1:8192/tss));
#ifdef	LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	TSF(tiff,TIFFTAG_IMAGEDESCRIPTION,desc);

	return tiff;
}

#define WriteTIFF(tiff,tsl,v,path)	\
	if (TIFFWriteScanline(tiff,tsl,(uint32)v,(uint16)0)<0) { \
	    fprintf(stderr,"%s : ",path); Error("write data error."); \
	}

static void	StoreWordImage(path,word,Nh,Nv,bps,desc)
char		*path,*desc;
Word		**word;
int		Nh,Nv,bps;
{
	int	v,h;
	Word	*data,mask,msb=(Word)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=word[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path);
	}
	free(tsl); TIFFClose(tiff);
}

static void	StoreCellImage(path,cell,Nh,Nv,bps,desc)
char		*path,*desc;
Cell		**cell;
int		Nh,Nv,bps;
{
	int	v,h;
	Cell	*data,mask,msb=(Cell)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=cell[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path);
	}
	free(tsl); TIFFClose(tiff);
}

static void	StoreByteImage(path,byte,Nh,Nv,bps,desc)
char		*path,*desc;
Byte		**byte;
int		Nh,Nv,bps;
{
	int	v,h;
	Byte	*data,mask,msb=(Byte)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=byte[v]; ptr=tsl;
	    if (bps==8)
		for (h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else
		for (bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path);
	}
	free(tsl); TIFFClose(tiff);
}

#define DESC_LEN	256
#define PATH_LEN	512

#define BPS(limit)	\
	(limit<2)?1:(limit<4)?2:(limit<16)?4:(limit<256)?8:(limit<65536)?16:32

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,scale=1.0;
	Ovoid	*ovoid;
	Map	map;
	int	v,h,depth=0,label=0;
	char	desc[DESC_LEN],path[PATH_LEN];

	if (argc!=6 && argc!=7 && argc!=12 && argc!=13)
Error("usage : bevmO {x1 y1 z1 x2 y2 z2} {scale} vLon vLat lLon lLat name");

	if (argc==12 || argc==13)
	    if ((x1=atof(argv[1]))>=(x2=atof(argv[4])) ||
		(y1=atof(argv[2]))>=(y2=atof(argv[5])) ||
		(z1=atof(argv[3]))>=(z2=atof(argv[6]))) Error("bad area.");

	if (argc==7 || argc==13)
	    if ((scale=atof(argv[argc-6]))<=0.0) Error("bad scaling factor.");

	if ((ovoid=ReadOvoid(&x3,&y3,&z3,
			     &x4,&y4,&z4))==NULL) Error("no ovoid.");

	printf("%lf\t%lf\t%lf\n%lf\t%lf\t%lf\n",x3,y3,z3,x4,y4,z4);

	if (argc==6 || argc==7) {
	    x1=x3; y1=y3; z1=z3; x2=x4; y2=y4; z2=z4;
	}
/*
	else
	    if (x1>x3 || x2<x4 ||
		y1>y3 || y2<y4 ||
		z1>z3 || z2<z4) fputs("too small area (warning).\n",stderr);
*/
	map=OvoidBEVM(ovoid,x1,y1,z1,x2,y2,z2,scale,
		      atof(argv[argc-5])*RADIAN,atof(argv[argc-4])*RADIAN,
		      atof(argv[argc-3])*RADIAN,atof(argv[argc-2])*RADIAN);

	sprintf(desc,
		"%lf\t%lf\t%lf\n%lf\t%lf\t%lf\n%lg\t%lg\t%lg\n%s\t%s\n",
		x1,y1,z1,x2,y2,z2,scale,scale,scale,argv[argc-5],argv[argc-4]);

	for (v=0; v<map.Nv; v++)
	for (h=0; h<map.Nh; h++) {
	    if ((int)DEPTH[v][h]>depth) depth=(int)DEPTH[v][h];
	    if ((int)LABEL[v][h]>label) label=(int)LABEL[v][h];
	}

	sprintf(path,"%s.%s",argv[argc-1],EXTENSION_DEPTH);
	StoreWordImage(path,DEPTH,map.Nh,map.Nv,BPS(depth),desc);

	sprintf(path,"%s.%s",argv[argc-1],EXTENSION_LABEL);
	StoreCellImage(path,LABEL,map.Nh,map.Nv,BPS(label),desc);

	sprintf(path,"%s.%s",argv[argc-1],EXTENSION_SHADE);
	StoreByteImage(path,SHADE,map.Nh,map.Nv,8,         desc);

	sprintf(path,"%s.%s",argv[argc-1],EXTENSION_EDGE );
	StoreByteImage(path,EDGE, map.Nh,map.Nv,1,         desc); return 0;
}
