
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "slice.h"

#define NAME_LEN	256
#define BITS		(8*sizeof(Cell))
#define MM(hvd,min,max)	if (hvd<min) min=hvd; if (hvd>max) max=hvd
#define RESTORE_CELL(d)	cell=depo[(d)/BITS]; mask=(Cell)(1<<((d)%BITS)); \
			for (v=v1; v<=v2; v++) \
			for (h=h1; h<=h2; h++) succ[v][h]=cell[v][h]&mask

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,lo,up,Td,Te,Nh,Nv,Nd,
		d,v,h,h1,v1,d1,h2,v2,d2,z,y,x,t,h3,v3,d3,h4,v4,d4;
	char	**names,**descs,name[NAME_LEN],*ptr[1],format[NAME_LEN],
		*nop="no occupied pixel.";
	Cell	***depo,**curr,**prev,**succ,**cell,mask;

	if (argc!=8)
	    Error("usage : de_asm org name lower upper dilation erosion new");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,&descs);

	if ((lo=atoi(argv[3]))<0 ||
	    (up=atoi(argv[4]))<lo) Error("bad pixel value range.");

	if ((Td=atoi(argv[5]))<0 ||
	    (Te=atoi(argv[6]))<0) Error("bad dilation / erosion thickness.");

	Nh=Nx+Td*2; Nv=Ny+Td*2; Nd=Nz+Td*2;

	depo=AllocateImageMemory(Nh,Nv,(Nd+BITS-1)/BITS);
	curr=AllocateSliceMemory(Nh,Nv);
	prev=AllocateSliceMemory(Nh,Nv);
	succ=AllocateSliceMemory(Nh,Nv);

	printf("%d\t%d\t%d\n",Nx,Ny,Nz);

	h1=Nh; v1=Nv; d1=Nd; h2=v2=d2=(-1);
	for (d=Td, z=0; z<Nz; z++, d++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&curr);

	    cell=depo[d/BITS]; mask=(Cell)(1<<(d%BITS));
	    for (v=Td, y=0; y<Ny; y++, v++)
	    for (h=Td, x=0; x<Nx; x++, h++)
		if ((int)curr[y][x]>=lo && (int)curr[y][x]<=up) {
		    cell[v][h]|=mask; MM(h,h1,h2); MM(v,v1,v2); MM(d,d1,d2);
		}
	}
	if (h1>h2 || v1>v2 || d1>d2) Error(nop);

	printf("%d\t%d\t%d\t%d\t%d\t%d\n",h1-Td,v1-Td,d1-Td,h2-Td,v2-Td,d2-Td);

	for (t=0; t<Td; t++) {
	    --h1; --v1; --d1; ++h2; ++v2; ++d2;

	    for (v=v1; v<=v2; v++)
	    for (h=h1; h<=h2; h++) succ[v][h]=0;

	    h3=Nh; v3=Nv; d3=Nd; h4=v4=d4=(-1);
	    for (d=d1; d<=d2; d++) {
		cell=prev; prev=curr; curr=succ; succ=cell;
		if (d!=d2) {
		    RESTORE_CELL(d+1);
		}
		cell=depo[d/BITS]; mask=(Cell)(1<<(d%BITS));
		for (v=v1; v<=v2; v++)
		for (h=h1; h<=h2; h++)
		    if (curr[v][h]) {
			MM(h,h3,h4); MM(v,v3,v4); MM(d,d3,d4);
		    }
		    else
			if ((d!=d1 && prev[v  ][h  ]) ||
			    (d!=d2 && succ[v  ][h  ]) ||
			    (v!=v1 && curr[v-1][h  ]) ||
			    (v!=v2 && curr[v+1][h  ]) ||
			    (h!=h1 && curr[v  ][h-1]) ||
			    (h!=h2 && curr[v  ][h+1])) {
			    cell[v][h]|=mask;
			    MM(h,h3,h4); MM(v,v3,v4); MM(d,d3,d4);
			}
	    }
	    h1=h3; v1=v3; d1=d3; h2=h4; v2=v4; d2=d4;
	}
	printf("%d\t%d\t%d\t%d\t%d\t%d\n",h1-Td,v1-Td,d1-Td,h2-Td,v2-Td,d2-Td);

	for (t=0; t<Te; t++) {
	    RESTORE_CELL(d1);

	    h3=Nh; v3=Nv; d3=Nd; h4=v4=d4=(-1);
	    for (d=d1; d<=d2; d++) {
		cell=prev; prev=curr; curr=succ; succ=cell;
		if (d!=d2) {
		    RESTORE_CELL(d+1);
		}
		cell=depo[d/BITS]; mask=(Cell)(~(1<<(d%BITS)));
		for (v=v1; v<=v2; v++)
		for (h=h1; h<=h2; h++)
		    if (curr[v][h])
			if (d==d1 || !prev[v  ][h  ] ||
			    d==d2 || !succ[v  ][h  ] ||
			    v==v1 || !curr[v-1][h  ] ||
			    v==v2 || !curr[v+1][h  ] ||
			    h==h1 || !curr[v  ][h-1] ||
			    h==h2 || !curr[v  ][h+1])
			    cell[v][h]&=mask;
			else {
			    MM(h,h3,h4); MM(v,v3,v4); MM(d,d3,d4);
			}
	    }
	    if (h3>h4 || v3>v4 || d3>d4) Error(nop);

	    h1=h3; v1=v3; d1=d3; h2=h4; v2=v4; d2=d4;
	}
	printf("%d\t%d\t%d\t%d\t%d\t%d\n",h1-Td,v1-Td,d1-Td,h2-Td,v2-Td,d2-Td);

	if (argv[7][0]=='-')
	    for (d=Td, z=0; z<Nz; z++, d++) {
		cell=depo[d/BITS]; mask=(Cell)(1<<(d%BITS));
		for (v=Td, y=0; y<Ny; y++, v++)
		for (h=Td, x=0; x<Nx; x++, h++)
		    curr[y][x]=(cell[v][h]&mask)?1:0;

		StoreImageFile(argv[7]+1,names+z,&curr,0,0,0,Nx-1,Ny-1,0,1,
			       descs+z,NULL);
	    }
	else {
	    printf("%d\t%d\t%d\n",h2-h1+1,v2-v1+1,d2-d1+1);

	    ptr[0]=name;
	    sprintf(name,"%d",d2-d1);
	    sprintf(format,"%%0%dd.tif",strlen(name));
	    for (z=0, d=d1; d<=d2; d++, z++) {
		cell=depo[d/BITS]; mask=(Cell)(1<<(d%BITS));
		for (v=v1; v<=v2; v++)
		for (h=h1; h<=h2; h++) curr[v][h]=(cell[v][h]&mask)?1:0;

		sprintf(name,format,z);
		StoreImageFile(argv[7],ptr,&curr,h1,v1,0,h2,v2,0,1,NULL,NULL);
	    }
	}
	return 0;
}
