
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

static void	Error(char *head,char *tail)
{
	(void)fprintf(stderr,"%s : %s\n",head,tail); exit(1);
}

static char	*Dir;

static char	**Aalloc(int noe)
{
	char	**mem=(char **)malloc(sizeof(char *)*noe);

	if (mem==NULL) Error(Dir,"name array memory allocation error.");

	return mem;
}

static char	*Salloc(char *str)
{
	char	*mem=strdup(str);

	if (mem==NULL) Error(Dir,"name string memory allocation error.");

	return mem;
}

static int	Normal(char **str1,char **str2)
{
	return strcmp(*str1,*str2);
}

static int	Reverse(char **str1,char **str2)
{
	return Normal(str2,str1);
}

static int	Numeric(char *str,double *val)
{
	for (; *str!='\0'; str++)
	    if (*str=='-' || *str=='+' ||
	       (*str>='0' && *str<='9')) return sscanf(str,"%lf",val)==1;

	return 0;
}

static int	Ascending(char **str1,char **str2)
{
	double	val1,val2;
	int	num1=Numeric(*str1,&val1),
		num2=Numeric(*str2,&val2);

	return (num1 && num2)?(val1>val2)? 1:
			      (val1<val2)?-1:0:(num1)? 1:
					       (num2)?-1:Normal(str1,str2);
}

static int	Descending(char **str1,char **str2)
{
	return Ascending(str2,str1);
}

#define LEN	1024

static char	**FetchName(char *nf,int *N)
{
	FILE		*file;
	char		str[LEN],**name,*top,
			*nifn="no image file name.",
			*bnifn="bad number of image file name.";
	int		n;
	DIR		*dir;
	struct dirent	*ent;

	if (nf[0]!='-') {
	    if ((file=fopen(nf,"r"))==NULL) Error(nf,"name file not found.");

	    for (*N=0; fscanf(file,"%s",str)==1; (*N)++) ;

	    if (*N==0) Error(nf,nifn);

	    name=Aalloc(*N);

	    rewind(file);
	    for (n=0; fscanf(file,"%s",str)==1; n++) {
		if (*N==n) Error(nf,bnifn);

		name[n]=Salloc(str);
	    }
	    if (*N!=n) Error(nf,bnifn);

	    (void)fclose(file);
	}
	else if (nf[1]!='\0' && nf[1]!='-') {
	    top=nf+1;
	    for (*N=0; sscanf(top,"%s",str)==1; (*N)++)
		top=strstr(top,str)+strlen(str);

	    if (*N==0) Error(nf+1,nifn);

	    name=Aalloc(*N);

	    top=nf+1;
	    for (n=0; sscanf(top,"%s",str)==1; n++) {
		if (*N==n) Error(nf,bnifn);

		name[n]=Salloc(str);

		top=strstr(top,str)+strlen(str);
	    }
	    if (*N!=n) Error(nf+1,bnifn);
	}
	else {
	    if ((dir=opendir(Dir))==NULL) Error(Dir,"directory not found.");

	    for (*N=0; (ent=readdir(dir))!=NULL; )
		if (strcmp(ent->d_name,".") && strcmp(ent->d_name,"..")) ++(*N);

	    if (*N==0) Error(Dir,nifn);

	    name=Aalloc(*N);

	    rewinddir(dir);
	    for (n=0; (ent=readdir(dir))!=NULL; )
		if (strcmp(ent->d_name,".") && strcmp(ent->d_name,"..")) {
		    if (*N==n) Error(Dir,bnifn);

		    name[n++]=Salloc(ent->d_name);
		}

	    if (*N!=n) Error(Dir,bnifn);

	    (void)closedir(dir);

	    if (nf[1]=='\0')
		qsort(name,*N,sizeof(char *),(int (*)())Normal);
	    else if (nf[2]=='r' && nf[3]=='\0')
		qsort(name,*N,sizeof(char *),(int (*)())Reverse);
	    else if (nf[2]=='n' && nf[3]=='\0')
		qsort(name,*N,sizeof(char *),(int (*)())Ascending);
	    else if ((nf[2]=='r' && nf[3]=='n') ||
		     (nf[2]=='n' && nf[3]=='r'))
		qsort(name,*N,sizeof(char *),(int (*)())Descending);
	    else
		Error(nf+1,"unknown sorting option.");
	}
	return name;
}

static char	Path[LEN];

static int	GetC(FILE *file)
{
	int	code=fgetc(file);

	if (code==EOF) Error(Path,"unexpected end of file.");

	return code;
}

static void	Skip(FILE *file,int bytes)
{
	while (--bytes>=0) (void)GetC(file);
}

#define BYTE	unsigned char

static double	Ratio;

static void	Read(FILE *file,int size,BYTE *cm)
{
	if (fread(cm,size,1,file)!=1) Error(Path,"colormap read error.");

	while (--size>=0) cm[size]=(BYTE)(Ratio*(double)cm[size]+0.5);
}

static void	Open(char *name,FILE **file,int *bps,BYTE *cm)
{
	char	s_v[6];
	int	flag,size;
	double	ratio;

	(void)sprintf(Path,"%s/%s",Dir,name);

	if ((*file=fopen(Path,"rb"))==NULL) Error(Path,"GIF file not found.");

	if (fread(s_v,6,1,*file)!=1 ||
	   (strncmp(s_v,"GIF87a",6) &&
	    strncmp(s_v,"GIF89a",6))) Error(Path,"not a GIF file.");

	Skip(*file,4);	/* logical screen width and height */
	flag=GetC(*file);
	Skip(*file,2);	/* background color index and pixel aspect ratio */

	Ratio=255.0/(double)((1<<((flag>>4)&7)+1)-1);

	*bps=(flag&7)+1; size=(1<<*bps)*3;
	if (flag&128)
	    if (cm!=NULL) Read(*file,size,cm); else Skip(*file,size);
	else
	    if (cm!=NULL)
		for (ratio=255.0/(double)(size-=3); size>=0; size-=3)
		cm[size+2]=cm[size+1]=cm[size]=(BYTE)(ratio*(double)size+0.5);
}

static int	GetW(FILE *file)
{
	int	code=GetC(file);

	return code|GetC(file)<<8;
}

static char	*mcm="missing color map.",
		*uei="unknown extension introducer.";

static void	ScanGIF(char *name,int *Nx,int *Ny,int *bps,int *same,BYTE *cm)
{
	FILE	*file;
	int	bpsG,ei,nxy,flag,bpsL,size,bytes;
	BYTE	cmG[256*3],cmL[256*3];

	Open(name,&file,&bpsG,(*same)?cmG:NULL);

	while ((ei=GetC(file))!=';')
	    if (ei==',') {
		Skip(file,4);	/* image left and top position */

		if ((nxy=GetW(file))>*Nx) *Nx=nxy;
		if ((nxy=GetW(file))>*Ny) *Ny=nxy;

		flag=GetC(file);

		bpsL=(flag&7)+1; size=(1<<bpsL)*3;
		if (flag&128)
		    if (*same) {
			Read(file,size,cmL);

			if (*bps==0)
			    (void)memcpy(cm,cmL,size);
			else
			    *same=(*bps)==bpsL && !memcmp(cm,cmL,size);
		    }
		    else
			Skip(file,size);
		else {
		    if (bpsL>bpsG) Error(Path,mcm);

		    if (*same)
			if (*bps==0)
			    (void)memcpy(cm,cmG,size);
			else
			    *same=(*bps)==bpsL && !memcmp(cm,cmG,size);
		}
		if (bpsL>*bps) *bps=bpsL;

		for (bytes=1; bytes!=0; bytes=GetC(file)) Skip(file,bytes);
	    }
	    else if (ei=='!')
		for (bytes=1; bytes!=0; bytes=GetC(file)) Skip(file,bytes);
	    else
		Error(Path,uei);

	(void)fclose(file);
}

static FILE	*File;

#define PUTC(code)	(void)fputc(code,File)

static void	PutW(int word)
{
	PUTC(word&255); PUTC(word>>8);
}

static void	CopyGIF(char *name,int Nx,int Ny,int bps,int same,unsigned dt)
{
	FILE	*file;
	int	bpsG,ei,nx,ny,flag,bpsL,size,bytes;
	BYTE	cmG[256*3],cmL[256*3];

	Open(name,&file,&bpsG,(same)?NULL:cmG);

	while ((ei=GetC(file))!=';')
	    if (ei==',') {
		Skip(file,4);	/* image left and top position */

		if ((nx=GetW(file))>Nx ||
		    (ny=GetW(file))>Ny) Error(Path,"bad image size.");

		PUTC(',');		/* image separator */
		PutW((Nx-nx)/2);	/* image left position */
		PutW((Ny-ny)/2);	/* image top position */
		PutW(nx);		/* image width */
		PutW(ny);		/* image height */

		flag=GetC(file);

		if ((bpsL=(flag&7)+1)>bps) Error(Path,"bad bits per sample.");

		size=(1<<bpsL)*3;
		if (same) {
		    if (flag&128) Skip(file,size);

		    PUTC(flag&95);	/* no local colormap */
		}
		else
		    if (flag&128) {
			Read(file,size,cmL);

			PUTC(flag); (void)fwrite(cmL,size,1,File);
		    }
		    else {
			if (bpsL>bpsG) Error(Path,mcm);

			PUTC((flag&95)|128); (void)fwrite(cmG,size,1,File);
		    }

		PUTC(GetC(file));	/* LZW code size */

		while ((bytes=GetC(file))!=0) {
		    PUTC(bytes); while (--bytes>=0) PUTC(GetC(file));
		}
		PUTC(0);		/* 0 = end of block */

		if (dt!=0) {
		    PUTC('!');	/* extension introducer */
		    PUTC(249);	/* graphic control label */
		    PUTC(4);	/* 4 = block size in bytes */
		    PUTC(0);	/* transparent color flag */
		    PutW(dt);	/* delay time in 1/100 sec. */
		    PUTC(0);	/* transparent color index */
		    PUTC(0);	/* 0 = end of block */
		}
	    }
	    else if (ei=='!')
		for (bytes=1; bytes!=0; bytes=GetC(file)) Skip(file,bytes);
	    else
		Error(Path,uei);

	(void)fclose(file);
}

int	main(int argc,char **argv)
{
	char		**name;
	int		N,n,Nx=0,Ny=0,bps=0,same=1;
	unsigned	dt,rc;
	BYTE		cm[256*3];

	if (argc!=6)
	Error("usage","gif_movie directory nameFile delayTime repeatCount GIF");

	Dir=argv[1];

	name=FetchName(argv[2],&N);

if (sscanf(argv[3],"%u",&dt)!=1 || dt>65535) Error(argv[3],"bad delay time.");
if (sscanf(argv[4],"%u",&rc)!=1 || rc>65536) Error(argv[4],"bad repeat count.");

	for (n=0; n<N; n++) ScanGIF(name[n],&Nx,&Ny,&bps,&same,cm);

	if (argv[5][0]=='-')
	    File=stdout;
	else
	    if ((File=fopen(argv[5],"wb"))==NULL)
		Error(argv[5],"file not open.");

	(void)fwrite((dt==0 && rc==0)?"GIF87a":"GIF89a",6,1,File);

	PutW(Nx);	/* logical screen width */
	PutW(Ny);	/* logical screen height */

	PUTC(((same)?128:0)|112|(bps-1));

	PUTC(0);	/* background color index */
	PUTC(0);	/* pixel aspect ratio */

	if (same) (void)fwrite(cm,(1<<bps)*3,1,File);

	if (rc!=0) {
	    PUTC('!');		/* extension introducer */
	    PUTC(255);		/* application extension label */
	    PUTC(11);		/* 11 = block size in bytes */
	    (void)fwrite("NETSCAPE2.0",11,1,File);
	    PUTC(3);		/* 3 = block size in bytes */
	    PUTC(1);		/* loop control parameter */
	    PutW(rc&65535);	/* repeat count */
	    PUTC(0);		/* 0 = end of block */
	}
	for (n=0; n<N; n++) CopyGIF(name[n],Nx,Ny,bps,same,dt);

	PUTC(';');	/* trailer */

	if (File!=stdout) (void)fclose(File);

	for (n=0; n<N; n++) free(name[n]);

	free(name);

	return 0;
}
