
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

#define MA(num,ptr)	malloc((num)*sizeof(*(ptr)))

static void	PareObjectSkin(xSize,ySize,zSize,cell,thres,thick)
int		xSize,ySize,zSize,thres,thick;
Cell		***cell;
{
	char	***flag,*msg="flag memory allocation error.";
	int	x,y,z,t,h=xSize-1,v=ySize-1,d=zSize-1;

	if ((flag=(char ***)MA(zSize,flag))==NULL) Error(msg);

	for (z=0; z<zSize; z++) {
	    if ((flag[z]=(char **)MA(ySize,*flag))==NULL) Error(msg);

	    for (y=0; y<ySize; y++) {
		if ((flag[z][y]=(char *)MA(xSize,**flag))==NULL) Error(msg);

		for (x=0; x<xSize; x++)
		    flag[z][y][x]=((int)cell[z][y][x]>=thres)?1:0;
	    }
	}

	for (t=0; t<thick; t++) {
	    for (z=0; z<zSize; z++)
	    for (y=0; y<ySize; y++)
	    for (x=0; x<xSize; x++)
		if (flag[z][y][x])
		    if (x==0 || !flag[z][y][x-1] ||
			x==h || !flag[z][y][x+1] ||
			y==0 || !flag[z][y-1][x] ||
			y==v || !flag[z][y+1][x] ||
		       (z!=0 && !flag[z-1][y][x]) ||
		       (z!=d && !flag[z+1][y][x])) flag[z][y][x]=2;

	    for (z=0; z<zSize; z++)
	    for (y=0; y<ySize; y++)
	    for (x=0; x<xSize; x++) if (flag[z][y][x]==2) flag[z][y][x]=0;
	}

	for (z=0; z<zSize; z++) {
	    for (y=0; y<ySize; y++) {
		for (x=0; x<xSize; x++) if (!flag[z][y][x]) cell[z][y][x]=0;

		free(flag[z][y]);
	    }
	    free(flag[z]);
	}
	free(flag);
}

#define IPV	(int)cell[y][x]
#define DPV	(double)cell[y][x]

static void	CheckInPlanePixelValueRange(x1,y1,x2,y2,cell,lo,up,
					    no,min,max,mean,SD)
int		x1,y1,x2,y2,lo,up,*no,*min,*max;
Cell		**cell;
double		*mean,*SD;
{
	int	x,y;

	*no=0; *min=up; *max=lo; *mean=(*SD)=0.0;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (IPV>=lo && IPV<=up) {
		++(*no); if (IPV<*min) *min=IPV;
			 if (IPV>*max) *max=IPV; *mean+=DPV; *SD+=(DPV*DPV);
	    }

	if (*no) {
	    *mean/=(double)*no; *SD=sqrt(*SD/(double)*no-*mean*(*mean));
	}
}

#define LIMIT	((1<<(8*sizeof(Cell)))-1)

#define DESC_LEN	512

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	xSize,ySize,zSize,thres,thick,x1,y1,x2,y2,lo,up,z,no,min,max;
	char	**names,desc[DESC_LEN];
	Cell	***cell;
	double	mean,SD;

	if (argc<5 || argc>12)
	Error(
"usage : iPRC+OSP orgD. nameF. thres. thick. {x1 y1 x2 y2} {lo. up.} {newD.}"
	);

	ScanNameFile(argv[1],argv[2],&zSize,&names);
	CheckImageFile(argv[1],zSize,names,&xSize,&ySize,NULL,NULL);

	if ((thres=atoi(argv[3]))<=0) Error("bad threshold.");

	thick=atoi(argv[4]);

	switch(argc) {
	    case 5 : case 6 :
		x1=y1=0; x2=xSize-1; y2=ySize-1;
		lo=0; up=LIMIT;
		break;
	    case 7 : case 8 :
		x1=y1=0; x2=xSize-1; y2=ySize-1;
		lo=atoi(argv[5]); up=atoi(argv[6]);
		break;
	    case 9 : case 10 :
		x1=atoi(argv[5]); y1=atoi(argv[6]);
		x2=atoi(argv[7]); y2=atoi(argv[8]);
		lo=0; up=LIMIT;
		break;
	    case 11 : case 12 :
		x1=atoi(argv[5]); y1=atoi(argv[ 6]);
		x2=atoi(argv[7]); y2=atoi(argv[ 8]);
		lo=atoi(argv[9]); up=atoi(argv[10]);
		break;
	}
	if (x1<0 || x1>x2 || x2>=xSize ||
	    y1<0 || y1>y2 || y2>=ySize) Error("bad cell area.");

	if (lo<0 || lo>up || up>LIMIT) Error("bad pixel value range.");

	cell=AllocateImageMemory(xSize,ySize,zSize);
	ReadImageFile(argv[1],names,xSize,ySize,zSize,cell);

	if (thick>0)
	    PareObjectSkin(xSize,ySize,zSize,cell,thres,thick);
	else if (thick==0)
{
	int	x,y;

	for (z=0; z<zSize; z++) for (y=y1; y<=y2; y++) for (x=x1; x<=x2; x++)
	    if ((int)cell[z][y][x]<thres) cell[z][y][x]=0;
}
	printf("%d\n",zSize);

	for (z=0; z<zSize; z++) {
	    CheckInPlanePixelValueRange(x1,y1,x2,y2,cell[z],lo,up,
					&no,&min,&max,&mean,&SD);
	    printf("%d",no);
	    if (no)
		printf("\t%d\t%d\t%lg\t%lg\n",min,max,mean,SD);
	    else
		puts("\t0\t0\t0\t0");
	}

	if (argc==6 || argc==8 || argc==10 || argc==12) {
	    sprintf(desc,"%d\t%d\t%d\t%d",x1,y1,x2,y2);
	    StoreImageFile(argv[argc-1],names,cell,x1,y1,0,x2,y2,zSize-1,
			   GetImageBPS(cell,x1,y1,0,x2,y2,zSize-1),NULL,desc);
	}
	return 0;
}
