
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"

#ifndef TABLE_RATIO
#define TABLE_RATIO	((27.0/256.0)/(1.0/4.0))
#endif

#define TABLE_SIZE	((~(1<<(8*sizeof(int)-1)))/sizeof(Table))
#define PV_STORAGE	(Cell)(~0)

typedef struct {
		Cell	pv;
		int	lpc;	/* label or pixels or cluster */
	} Table;

static int	Search(table,label)
Table		*table;
int		label;
{
	while (table[label].lpc<0) label=(-table[label].lpc); return label;
}

typedef struct {
		Cell	pv;
		int	pixels;
		union {
			int	label;
			struct {
				short	x1,y1,z1,x2,y2,z2;
			} area;
		} u;
#ifdef	LEV
		int	lev;
#endif
	} Data;

static int	ComparePV(ptr1,ptr2)
Data		*ptr1,*ptr2;
{
	return (ptr1->pv    ==ptr2->pv    )?
	       (ptr1->pixels==ptr2->pixels)?ptr1->u.label-ptr2->u.label:
					    ptr2->pixels -ptr1->pixels:
					   (ptr1->pv     >ptr2->pv)?1:-1;
}

#ifndef	LEV
static int	ComparePixels(ptr1,ptr2)
Data		*ptr1,*ptr2;
{
	return (ptr1->pixels==ptr2->pixels)?
	       (ptr1->pv    ==ptr2->pv    )?ptr1->u.label-ptr2->u.label:
					   (ptr1->pv     >ptr2->pv)?1:-1:
					    ptr2->pixels -ptr1->pixels;
}
#endif

#define MA(noe,top)	malloc((size_t)(noe)*sizeof(*(top)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nz,Nx,Ny,y,x,size,labels,z,
			pixels,new,old,clusters,label,cluster,bps;
	char		**names,**descs,*amae="array memory allocation error.";
	Cell		**cell,pv;
	Table		**last,*table,*tmp;
	double		table_size;
	Data		*data,*ptr;
	unsigned long	pvs;
#ifdef	LEV
	int		lev;
#endif

	if (argc!=3 && argc!=4)
	    Error("usage : mcl_mpv orgDir nameFile {newDir}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,&descs);

	cell=AllocateSliceMemory(Nx,Ny);

	if ((last=(Table **)MA(Ny+1,last))==NULL) Error(amae);

	++last;
	for (y=(-1); y<Ny; y++) {
	    if ((last[y]=(Table *)MA(Nx+1,last[y]))==NULL) Error(amae);

	    ++last[y];
	    for (x=(-1); x<Nx; x++) last[y][x].pv=0;
	}

	table_size=TABLE_RATIO*(double)Nx*(double)Ny*(double)Nz+1.0;
	size=(table_size>(double)TABLE_SIZE)?TABLE_SIZE:(int)table_size;

	for (; size>0 && (table=(Table *)MA(size,table))==NULL; size/=2)
	fprintf(stderr,"no allocatable memory for table of size %d.\n",size);

	if (size<=0) Error("table memory allocation error.");

table[0].pv=0; table[0].lpc=0; labels=1;
for (z=0; z<Nz; z++) {
    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

    for (y=0; y<Ny; y++)
    for (x=0; x<Nx; x++) {
	pixels=1;
	if ((pv=cell[y][x])==0) {
	    pixels+=table[new=0].lpc; last[y][x].pv=0;
	}
	else
	    if (pv==last[y][x].pv) {
		pixels+=table[new=Search(table,last[y][x].lpc)].lpc;
		if (pv==last[y-1][x].pv &&
		   (old=Search(table,last[y-1][x].lpc))!=new) {
		    pixels+=table[old].lpc; table[old].lpc=(-new);
		}
		if (pv==last[y][x-1].pv &&
		   (old=Search(table,last[y][x-1].lpc))!=new) {
		    pixels+=table[old].lpc; table[old].lpc=(-new);
		}
	    }
	    else {
		if (pv==last[y-1][x].pv) {
		    pixels+=table[new=Search(table,last[y-1][x].lpc)].lpc;
		    if (pv==last[y][x-1].pv && (old=last[y][x-1].lpc)!=new) {
			pixels+=table[old].lpc; table[old].lpc=(-new);
		    }
		}
		else
		    if (pv==last[y][x-1].pv)
			pixels+=table[new=last[y][x-1].lpc].lpc;
		    else {
			if (labels==size) Error("too many labels.");

			table[new=labels++].pv=pv;
		    }

		last[y][x].pv=pv;
	    }

	if ((table[last[y][x].lpc=new].lpc=pixels)<0) Error("too many pixels.");
    }
}

	if ((tmp=(Table *)realloc(table,labels*sizeof(Table)))!=NULL) table=tmp;

	clusters=1;
	for (label=1; label<labels; label++) if (table[label].lpc>0) ++clusters;

	if ((data=(Data *)MA(clusters,data))==NULL)
	    Error("data memory allocation error.");

	ptr=data;
	for (label=0; label<labels; label++)
	    if (table[label].lpc>=0) {
		ptr->pv=table[label].pv;
		ptr->pixels=table[label].lpc;
		ptr->u.label=label;
		++ptr;
	    }
	    else
		table[label].lpc=(-Search(table,label));

	if (clusters>1) qsort(data+1,clusters-1,sizeof(*data),ComparePV);

	pvs=1L; pv=0;
	for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++)
	    if (pv!=ptr->pv) {
		++pvs; pv=ptr->pv;
	    }

#ifndef	LEV
	if (clusters>1) qsort(data+1,clusters-1,sizeof(*data),ComparePixels);
#endif

	for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++) {
	    table[ptr->u.label].lpc=cluster;
	    ptr->u.area.x1=(short)(Nx-1); ptr->u.area.x2=0;
	    ptr->u.area.y1=(short)(Ny-1); ptr->u.area.y2=0;
	    ptr->u.area.z1=(short)(Nz-1); ptr->u.area.z2=0;
#ifdef	LEV
	    if (ptr->pv==0) {
		pv=0; ptr->lev=0;
	    }
	    else {
		if (pv!=ptr->pv) {
		    pv=ptr->pv; lev=1;
		}
		ptr->lev=lev++;
	    }
#endif
	}
	for (label=1; label<labels; label++)
	    if (table[label].lpc<0)
		table[label].lpc=table[-table[label].lpc].lpc;

	if (argc==4) { 
#ifdef	LEV
	    lev=0;
	    for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++)
		if (ptr->lev>lev) lev=ptr->lev;

	    pv=(lev>PV_STORAGE)?PV_STORAGE:(Cell)lev;
#else
	    pv=(clusters-1>PV_STORAGE)?PV_STORAGE:(Cell)(clusters-1);
#endif
	    bps=(pv<2)?1:(pv<4)?2:(pv<16)?4:(pv<256)?8:(pv<65536)?16:32;
	}

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) last[y][x].pv=0;

	label=1;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		if ((pv=cell[y][x])==0) {
		    cluster=0; last[y][x].pv=0;
		}
		else if (pv==last[y][x].pv)
		    cluster=last[y][x].lpc;
		else {
		    if (pv==last[y-1][x].pv)
			cluster=last[y-1][x].lpc;
		    else if (pv==last[y][x-1].pv)
			cluster=last[y][x-1].lpc;
		    else
			cluster=table[label++].lpc;

		    last[y][x].pv=pv;
		}
		ptr=data+(last[y][x].lpc=cluster);
		if (x<(int)ptr->u.area.x1) ptr->u.area.x1=(short)x;
		if (x>(int)ptr->u.area.x2) ptr->u.area.x2=(short)x;
		if (y<(int)ptr->u.area.y1) ptr->u.area.y1=(short)y;
		if (y>(int)ptr->u.area.y2) ptr->u.area.y2=(short)y;
		if (z<(int)ptr->u.area.z1) ptr->u.area.z1=(short)z;
		if (z>(int)ptr->u.area.z2) ptr->u.area.z2=(short)z;

		if (argc==4)
#ifdef	LEV
		    cell[y][x]=(ptr->lev>PV_STORAGE)?PV_STORAGE:(Cell)ptr->lev;
#else
		    cell[y][x]=(cluster>PV_STORAGE)?PV_STORAGE:(Cell)cluster;
#endif
	    }
	    if (argc==4)
		StoreImageFile(argv[3],names+z,&cell,
			       0,0,0,Nx-1,Ny-1,0,bps,descs+z,NULL);
	}

	printf("%lu\t%d\n",pvs,clusters);

#ifdef	LEV
	pv=0;
#endif
	for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++)
#ifdef	LEV
{
	    if (ptr->pv!=pv) {
		pv=ptr->pv;
		for (lev=1; cluster+lev<clusters && pv==ptr[lev].pv; lev++) ;

		printf("%u\t%d\n",pv,lev);
	    }
	    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
		   ptr->pixels,(int)ptr->u.area.x1,
			       (int)ptr->u.area.y1,
			       (int)ptr->u.area.z1,
			       (int)ptr->u.area.x2,
			       (int)ptr->u.area.y2,
			       (int)ptr->u.area.z2);
}
#else
	    printf("%u\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
		   ptr->pv,ptr->pixels,(int)ptr->u.area.x1,
				       (int)ptr->u.area.y1,
				       (int)ptr->u.area.z1,
				       (int)ptr->u.area.x2,
				       (int)ptr->u.area.y2,
				       (int)ptr->u.area.z2);
#endif

	return 0;
}
