
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

#ifndef	SEGMENT
#define SEGMENT		180
#endif

#ifndef	ITERATION
#define ITERATION	1024
#endif

static double	xx,yy,zz,xy,yz,zx,L0;

#define VAL	vec[0]
#define LON	vec[1]
#define LAT	vec[2]

#define LON1	(-M_PI)
#define LON2	M_PI
#define LAT1	0.0
#define LAT2	(M_PI/2.0)

static void	EvaluateAxis(vec)
double		*vec;
{
	double	co,so,co2,so2,cso,ca,sa,ca2,sa2,csa;

	if (LON>LON1 && LON<=LON2 && LAT>=LAT1 && LAT<=LAT2) {
	    co=cos(LON); so=sin(LON); co2=co*co; so2=so*so; cso=co*so; 
	    ca=cos(LAT); sa=sin(LAT); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
#ifdef	OBLATE
	    VAL=((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#else
	    VAL=L0-
		((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#endif
	}
	else
	    VAL=L0;
}

typedef	struct {
		double	M,x0,y0,z0,xx,yy,zz,xy,yz,zx;
	} Moment;

typedef struct {
		double	a,b,c,lon,lat,rot;
	} Axis;

static double	can0[3],can1[3],can2[3],*can[3]={can0,can1,can2};

static Axis	FitOvoid(Sx,Sy,Sz,moment)
double		Sx,Sy,Sz;
Moment		moment;
{
	Axis	axis;
	double	val,lon,lat,vec[3],step=M_PI/(double)SEGMENT,
		co,so,co2,so2,cso,
		ca,sa,ca2,sa2,csa,H2,V2,HV,HmV,rot2,dHV,HV0,a,b;

	xx=Sx*Sx*moment.xx;
	yy=Sy*Sy*moment.yy;
	zz=Sz*Sz*moment.zz;
	xy=Sx*Sy*moment.xy;
	yz=Sy*Sz*moment.yz;
	zx=Sz*Sx*moment.zx;

	val=L0=xx+yy+zz; lon=LON2; lat=LAT2;
	for (LON=LON2; LON> LON1; LON-=step)
	for (LAT=LAT2; LAT>=LAT1; LAT-=step) {
	    EvaluateAxis(vec);
	    if (VAL<val) {
		val=VAL; lon=LON; lat=LAT;
	    }
	}
	can0[1]=can2[1]=lon; can1[1]=lon+step;
	can0[2]=can1[2]=lat; can2[2]=lat-step;

	Simplex(2,can,EvaluateAxis,ITERATION,vec);

	axis.lon=LON;
	axis.lat=LAT;
#ifdef	OBLATE
	axis.c=sqrt(5.0*VAL/moment.M);
#else
	axis.c=sqrt(5.0*(L0-VAL)/moment.M);
#endif
	co=cos(axis.lon); so=sin(axis.lon); co2=co*co; so2=so*so; cso=co*so;
	ca=cos(axis.lat); sa=sin(axis.lat); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
	H2=xx*so2+yy*co2-2.0*xy*cso;
	V2=(xx*co2+yy*so2+2.0*xy*cso)*sa2-2.0*(zx*co+yz*so)*csa+zz*ca2;
	HV=((xx-yy)*cso-xy*(co2-so2))*sa-(zx*so-yz*co)*ca;
	HmV=(H2-V2)/2.0;
	rot2=atan2(HV,HmV);
	dHV=HmV*cos(rot2)+HV*sin(rot2);
	HV0=(H2+V2)/2.0;

#ifdef	OBLATE
	if (dHV>=0.0) {
#else
	if (dHV<=0.0) {
#endif
	    b=HV0-dHV;
	    a=HV0+dHV; if (rot2<0.0) rot2+=(M_PI*2.0);
	}
	else {
	    b=HV0+dHV;
	    a=HV0-dHV; rot2+=M_PI;
	}
	axis.rot=rot2/2.0; axis.b=sqrt(5.0*b/moment.M);
			   axis.a=sqrt(5.0*a/moment.M);

	return axis;
}

#define DEGREE	(180.0/M_PI)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,z,x,y;
	char	**names;
	Moment	moment;
	double	Sx,Sy,Sz,base,step,m;
	Cell	**cell;
	Axis	axis;

	if (argc!=8 && argc!=11)
	    Error("usage : of_md dir. name {x0 y0 z0} sX sY sZ base step");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==8)
	    moment.x0=
	    moment.y0=
	    moment.z0=0.0;
	else {
	    moment.x0=atof(argv[3]);
	    moment.y0=atof(argv[4]);
	    moment.z0=atof(argv[5]);
	}
	moment.M=moment.xx=moment.yy=moment.zz=
		 moment.xy=moment.yz=moment.zx=0.0;

	if ((Sx=atof(argv[argc-5]))<=0.0 ||
	    (Sy=atof(argv[argc-4]))<=0.0 ||
	    (Sz=atof(argv[argc-3]))<=0.0) Error("bad scaling factor.");

	base=atof(argv[argc-2]);
	step=atof(argv[argc-1]);

	cell=AllocateSliceMemory(Nx,Ny);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		moment.M+=(m=base+step*(double)cell[y][x]);

		if (argc==8) {
		    moment.x0+=(m*(double)x);
		    moment.y0+=(m*(double)y);
		    moment.z0+=(m*(double)z);
		}
		moment.xx+=(m*((double)x*(double)x+1.0/12.0));
		moment.yy+=(m*((double)y*(double)y+1.0/12.0));
		moment.zz+=(m*((double)z*(double)z+1.0/12.0));
		moment.xy+=(m*(double)x*(double)y);
		moment.yz+=(m*(double)y*(double)z);
		moment.zx+=(m*(double)z*(double)x);
	    }
	}
	if (moment.M<=0.0) Error("non-positive total mass.");

	if (argc==8) {
	    moment.x0/=moment.M;
	    moment.y0/=moment.M;
	    moment.z0/=moment.M;
	}
	moment.xx-=(moment.x0*moment.x0*moment.M);
	moment.yy-=(moment.y0*moment.y0*moment.M);
	moment.zz-=(moment.z0*moment.z0*moment.M);
	moment.xy-=(moment.x0*moment.y0*moment.M);
	moment.yz-=(moment.y0*moment.z0*moment.M);
	moment.zx-=(moment.z0*moment.x0*moment.M);

	axis=FitOvoid(Sx,Sy,Sz,moment);

	printf(
"%d\n%lf\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n",
	       Nx*Ny*Nz,
	       moment.M,
	       moment.x0,moment.y0,moment.z0,
	       Sx*moment.x0,Sy*moment.y0,Sz*moment.z0,
	       axis.lon*DEGREE,axis.lat*DEGREE,axis.rot*DEGREE,
	       axis.a,axis.b,axis.c);

	return 0;
}
