
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STRLEN	256

static int	Skip(pnm)
FILE		*pnm;
{
	int	chr;

	while ((chr=fgetc(pnm))==' ' || chr=='\t' || chr=='\r' || chr=='\n');

	if (chr==EOF) Error("unexpected end of file.");

	if (chr=='#') {
	    while ((chr=fgetc(pnm))!=EOF && chr!='\n');

	    return Skip(pnm);
	}
	return chr;
}

static int	GetNum(pnm)
FILE		*pnm;
{
	int	chr=Skip(pnm),num=0;

	if (chr<'0' || chr>'9') Error("not a numeric character.");

	do num=num*10+chr-'0'; while ((chr=fgetc(pnm))>='0' && chr<='9');

	ungetc(chr,pnm); return num;
}

static int	Get01(pnm)
FILE		*pnm;
{
	int	chr=Skip(pnm);

	if (chr!='0' && chr!='1') Error("not a binary number character.");

	return chr-'0';
}

static FILE	*pnm;
static int	max,(*get)();

static int	GetVal()
{
	int	val=(*get)(pnm);

	if (val>max) Error("too large value.");

	return val;
}

#define TSF	TIFFSetField

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	magic,msb,tss,y,z,x,val,msk;
	uint32	wid,hei,*ptr32;
	uint16	bps,pi,spp,*ptr16;
	TIFF	*tiff;
	uint8	*tsl,*ptr,bit;

	if (argc!=2 && argc!=3) Error("usage : pnm2tiff {PNM} TIFF");

	if (argc==2)
	    pnm=stdin;
	else
	    if ((pnm=fopen(argv[1],"rb"))==NULL) Error("PNM file not found.");

	if (fgetc(pnm)!='P') Error("magic number not found.");

	if ((magic=fgetc(pnm)-'0')<1 || magic>6) Error("bad magic number.");

	if ((wid=(uint32)GetNum(pnm))==0 ||
	    (hei=(uint32)GetNum(pnm))==0) Error("bad image size.");

	if (magic==1 || magic==4)
	    bps=1;
	else {
	    if ((max=GetNum(pnm))>255 && magic!=2) Error("bad maximum value.");

	    for (bps=1; (1<<bps)<=max; bps++);

	    if (bps!=8 && bps!=16 && bps!=32) msb=1<<(bps-1);
	}
	if (magic>=4) (void)fgetc(pnm);

	switch(magic) {
	    case 1: pi=PHOTOMETRIC_MINISWHITE; spp=1; break;
	    case 2: pi=PHOTOMETRIC_MINISBLACK; spp=1; get=GetNum; break;
	    case 3: pi=PHOTOMETRIC_RGB; spp=3; get=GetNum; break;
	    case 4: pi=PHOTOMETRIC_MINISWHITE; spp=1; break;
	    case 5: pi=PHOTOMETRIC_MINISBLACK; spp=1; get=fgetc; break;
	    case 6: pi=PHOTOMETRIC_RGB; spp=3; get=fgetc; break;
	}

	if ((tiff=TIFFOpen(argv[argc-1],"w"))==NULL)
	    Error("TIFF file not created.");

	TSF(tiff,TIFFTAG_IMAGEWIDTH,wid);
	TSF(tiff,TIFFTAG_IMAGELENGTH,hei);
	TSF(tiff,TIFFTAG_PHOTOMETRIC,pi);
	TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,spp);
	TSF(tiff,TIFFTAG_BITSPERSAMPLE,bps);
	TSF(tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);

	if ((tsl=(uint8 *)malloc(tss=TIFFScanlineSize(tiff)))==NULL)
	    Error("TIFF scanline allocation error.");

	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(tss>8192)?1:8192/tss);
#ifdef LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	for (y=0; y<hei; y++) {
	    if (magic==1)
		for (ptr=tsl, bit=0x80, x=0; x<wid; x++) {
		    if (Get01(pnm)) *ptr|=bit; else *ptr&=(~bit);

		    if ((bit>>=1)==0) {
			bit=0x80; ++ptr;
		    }
		}
	    else if (magic==4)
		for (ptr=tsl, x=0; x<wid; x+=8) *(ptr++)=(uint8)fgetc(pnm);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=0; x<wid; x++)
		    *(ptr32++)=(uint32)GetVal();
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=0; x<wid; x++)
		    *(ptr16++)=(uint16)GetVal();
	    else if (bps==8)
		for (ptr=tsl, x=0; x<wid; x++)
		    for (z=0; z<spp; z++) *(ptr++)=(uint8)GetVal();
	    else
		for (ptr=tsl, bit=0x80, x=0; x<wid; x++)
		    for (z=0; z<spp; z++)
			for (val=GetVal(), msk=msb; msk!=0; msk>>=1) {
			    if (val&msk) *ptr|=bit; else *ptr&=(~bit);

			    if ((bit>>=1)==0) {
				bit=0x80; ++ptr;
			    }
			}

	    if (TIFFWriteScanline(tiff,tsl,y,0)<0)
		Error("TIFF scanline data write error.");
	}
	TIFFClose(tiff); if (pnm!=stdin) fclose(pnm); return 0;
}
