
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiffio.h"
#include "slice.h"

#define PIPE_LEN	1024
#define NAME_LEN	512

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

typedef struct List {
			char		*name;
			struct List	*next;
	} List;

void	ScanNameFile(dir,nfn,cnt,names)
char	*dir,*nfn,***names;
int	*cnt;
{
	char	pipe[PIPE_LEN],name[NAME_LEN];
	FILE	*file;
	List	*next,*top=NULL,**ptr=(&top);
	int	idx;

	if (nfn[0]=='-') {
	    sprintf(pipe,"( cd %s ; ls %s )",dir,&(nfn[1]));
	    if ((file=popen(pipe,"r"))==NULL) {
		fprintf(stderr,"%s : ",pipe); Error("pipe not open.");
	    }
	}
	else {
	    *pipe='\0';
	    if ((file=fopen(nfn,"r"))==NULL) {
		fprintf(stderr,"%s : ",nfn); Error("name file not found.");
	    }
	}
	for (*cnt=0; fscanf(file,"%s",name)==1; (*cnt)++) {
	    if ((*ptr=(List *)malloc(sizeof(**ptr)))==NULL ||
	       ((*ptr)->name=(char *)malloc(strlen(name)+1))==NULL)
		Error("name list memory allocation error.");

	    strcpy((*ptr)->name,name); (*ptr)->next=NULL; ptr=(&((*ptr)->next));
	}
	if (*pipe) pclose(file); else fclose(file);

	if (*cnt==0) Error("no name exists.");

	if ((*names=(char **)MA(*cnt,*names))==NULL)
	    Error("name array memory allocation error.");

	for (idx=0; idx<*cnt; idx++) {
	    (*names)[idx]=top->name; next=top->next; free(top); top=next;
	}
}

#define BAD_IMAGE(msg)	{ fprintf(stderr,"%s : ",path); Error(msg); }

#define TGF	TIFFGetField
#define TGFD	TIFFGetFieldDefaulted

static TIFF	*GetTIFFField(path,iw,il,bps,flag)
char		*path;
uint32		*iw,*il;
uint16		*bps;
int		flag;
{
	TIFF	*tiff;
	uint16	pi,spp,pc;

	if ((tiff=TIFFOpen(path,"r"))==NULL) {
	    if (flag) fprintf(stderr,"%s : image file not found.\n",path);

	    return NULL;
	}

	if (!TGF(tiff,TIFFTAG_IMAGEWIDTH, iw) ||
	    !TGF(tiff,TIFFTAG_IMAGELENGTH,il) ||
	    !TGF(tiff,TIFFTAG_PHOTOMETRIC,&pi))
	    BAD_IMAGE("image attribute not found.");

	if (pi!=PHOTOMETRIC_MINISWHITE &&
	    pi!=PHOTOMETRIC_MINISBLACK &&
	    pi!=PHOTOMETRIC_PALETTE)
	    BAD_IMAGE("bad photometric interpretation.");

	if (!TGF(tiff,TIFFTAG_SAMPLESPERPIXEL,&spp))
	    TGFD(tiff,TIFFTAG_SAMPLESPERPIXEL,&spp);

	if (spp!=1) BAD_IMAGE("too many samples per pixel.");

	if (!TGF(tiff,TIFFTAG_BITSPERSAMPLE,bps))
	    TGFD(tiff,TIFFTAG_BITSPERSAMPLE,bps);

	if (*bps>8*sizeof(Cell)) BAD_IMAGE("too many bits per sample.");

	if (!TGF(tiff,TIFFTAG_PLANARCONFIG,&pc))
	    TGFD(tiff,TIFFTAG_PLANARCONFIG,&pc);

	if (pc!=PLANARCONFIG_CONTIG) BAD_IMAGE("bad planar configuration.");

	return tiff;
}

#define PATH_LEN	512

void	CheckImageFile(dir,cnt,names,wid,hig,BPS,descs)
char	*dir,**names,***descs;
int	cnt,*wid,*hig,*BPS;
{
	int	idx,flag=1;
	char	path[PATH_LEN],*desc;
	TIFF	*tiff;
	uint32	iw,il;
	uint16	bps;

	if (descs!=NULL)
	    if ((*descs=(char **)MA(cnt,*descs))==NULL)
		Error("description array memory allocation error.");

	for (idx=0; idx<cnt; idx++) {
	    sprintf(path,"%s/%s",dir,names[idx]);

	    if ((tiff=GetTIFFField(path,&iw,&il,&bps,1))==NULL) {
		if (descs!=NULL) (*descs)[idx]=NULL;

		continue;
	    }

	    if (flag) {
		flag=0;
		if (wid!=NULL) *wid=iw;
		if (hig!=NULL) *hig=il;
		if (BPS!=NULL) *BPS=bps;
	    }
	    else {
		if (wid!=NULL && *wid<iw) *wid=iw;
		if (hig!=NULL && *hig<il) *hig=il;
		if (BPS!=NULL && *BPS<bps) *BPS=bps;
	    }

	    if (descs!=NULL)
		if (TGF(tiff,TIFFTAG_IMAGEDESCRIPTION,&desc)) {
		    if (((*descs)[idx]=(char *)malloc(strlen(desc)+1))==NULL)
			BAD_IMAGE("description memory allocation error.");

		    strcpy((*descs)[idx],desc);
		}
		else
		    (*descs)[idx]=NULL;

	    TIFFClose(tiff);
	}
	if (flag) Error("no image file found.");
}

Cell	**AllocateSliceMemory(wid,hig)
int	wid,hig;
{
	Cell	**cell=(Cell **)MA(hig,cell);
	char	*msg="slice memory allocation error.";
	int	y,x;

	if (cell==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((cell[y]=(Cell *)MA(wid,*cell))==NULL) Error(msg);

	    for (x=0; x<wid; x++) cell[y][x]=0;
	}
	return cell;
}

Cell	***AllocateImageMemory(wid,hig,num)
int	wid,hig,num;
{
	Cell	***cell=(Cell ***)MA(num,cell);
	int	z;

	if (cell==NULL) Error("image memory allocation error.");

	for (z=0; z<num; z++) cell[z]=AllocateSliceMemory(wid,hig);

	return cell;
}

void	ReadSliceFile(path,wid,hig,cell)
char	*path;
int	wid,hig;
Cell	**cell;
{
	TIFF	*tiff;
	uint32	iw,il,*ptr32;
	uint16	bps,*ptr16;
	int	y,x,idx;
	Cell	*line,data;
	uint8	*tsl,*ptr,bit;

	if ((tiff=GetTIFFField(path,&iw,&il,&bps,0))==NULL)
	    for (y=0; y<hig; y++) {
		line=(*(cell++)); for (x=0; x<wid; x++) *(line++)=0;
	    }
	else {
	    if (iw>wid || il>hig) BAD_IMAGE("too large image size.");

	    if ((tsl=(uint8 *)malloc(TIFFScanlineSize(tiff)))==NULL)
		BAD_IMAGE("scanline memory allocation error.");

	    for (y=0; y<il; y++) {
		line=(*(cell++));

		if (TIFFReadScanline(tiff,tsl,(uint32)y,(uint16)0)<0)
		    for (x=0; x<wid; x++) *(line++)=0;
		else {
		    if (bps==8)
			for (ptr=tsl, x=0; x<iw; x++) *(line++)=(Cell)*(ptr++);
		    else if (bps==16)
			for (ptr16=(uint16 *)tsl, x=0; x<iw; x++)
			    *(line++)=(Cell)*(ptr16++);
		    else if (bps==32)
			for (ptr32=(uint32 *)tsl, x=0; x<iw; x++)
			    *(line++)=(Cell)*(ptr32++);
		    else
			for (ptr=tsl, bit=0x80, x=0; x<iw; x++) {
			    for (data=0, idx=0; idx<bps; idx++) {
				data<<=1; if (*ptr&bit) data|=1;
				if ((bit>>=1)==0) { ++ptr; bit=0x80; }
			    }
			    *(line++)=data;
			}

		    for (x=iw; x<wid; x++) *(line++)=0;
		}
	    }
	    for (y=il; y<hig; y++) {
		line=(*(cell++)); for (x=0; x<wid; x++) *(line++)=0;
	    }
	    free(tsl); TIFFClose(tiff);
	}
}

void	ReadImageFile(dir,names,wid,hig,num,cell)
char	*dir,**names;
int	wid,hig,num;
Cell	***cell;
{
	int	z;
	char	path[PATH_LEN];

	for (z=0; z<num; z++) {
	    sprintf(path,"%s/%s",dir,names[z]);
	    ReadSliceFile(path,wid,hig,cell[z]);
	}
}
