
#include <stdio.h>
#include <stdlib.h>
#include "cmimage.h"

typedef struct {
		int	number,lo,up,size,r0,g0,b0;
	} Cell;

static Cell	*cell;
static int	*pxl,*lut,pixels;

static void	Store(lo,up,size,r0,g0,b0)
int		lo,up,size,r0,g0,b0;
{
	Cell	*ptr;
	int	pixel;

	for (ptr=cell; ptr->number!=0; ptr++) ;

	for (pixel=lo; pixel<up; pixel++) ptr->number+=pxl[lut[pixel]];
	if ((ptr->lo=lo)==(ptr->up=up)-1)
	    ptr->size=1;
	else {
	    ptr->size=size; ptr->r0=r0; ptr->g0=g0; ptr->b0=b0;
	}
	++pixels;
}

static int	Divide(lo,up,rgb,pivot)
int		lo,up;
word		*rgb,pivot;
{
	int	top,idx,tmp;

	for (top=lo; top<up; top++) if (rgb[lut[top]]>=pivot) break;
	for (idx=top+1; idx<up; idx++)
	    if (rgb[lut[idx]]<pivot) {
		tmp=lut[top]; lut[top]=lut[idx]; lut[idx]=tmp; ++top;
	    }
	return(top);
}

CM	*ReduceCM(image,org,limit)
Image	*image;
CM	*org;
int	limit;
{
	int	pixel,x,y,number,lo,up,size,r0,g0,b0,rP,gP,bP,rC,gC,bC,i,j;
	Cell	*ptr;
	CM	*new;
	double	r,g,b,cnt;

	if ((lut=(int *)calloc(org->pixels,sizeof(*lut)))==NULL ||
	    (pxl=(int *)calloc(org->pixels,sizeof(*pxl)))==NULL ||
	    (cell=(Cell *)calloc(limit,sizeof(*cell)))==NULL) return(NULL);

	for (pixel=0; pixel<org->pixels; pixel++) {
	    lut[pixel]=pixel; pxl[pixel]=0;
	}
	for (x=0; x<image->xSize; x++)
	for (y=0; y<image->ySize; y++) ++(pxl[GETPIXEL(image,x,y)]);

	ptr=cell; ptr->number=image->xSize*image->ySize;
		  ptr->lo=0; ptr->up=org->pixels;
		  ptr->size=0x10000; ptr->r0=ptr->g0=ptr->b0=0; pixels=1;
	for (++ptr, pixel=1; pixel<limit; pixel++, ptr++) ptr->number=0;

	while (pixels<=limit-7) {
	    number=0;
	    for (pixel=0; pixel<limit; pixel++)
		if (cell[pixel].number>number && cell[pixel].size>1) {
		    ptr=(&(cell[pixel])); number=ptr->number;
		}
	    if (number==0) break;

	    ptr->number=0; lo=ptr->lo; up=ptr->up;
	    		   size=ptr->size>>1; rP=(r0=ptr->r0)+size;
					      gP=(g0=ptr->g0)+size;
					      bP=(b0=ptr->b0)+size; --pixels;
	    bC=Divide(lo,up,org->b,(word)bP);
	    if (bC>lo) {
		gC=Divide(lo,bC,org->g,(word)gP);
		if (gC>lo) {
		    rC=Divide(lo,gC,org->r,(word)rP);
		    if (rC>lo) Store(lo,rC,size,r0,g0,b0);
		    if (rC<gC) Store(rC,gC,size,rP,g0,b0);
		}
		if (gC<bC) {
		    rC=Divide(gC,bC,org->r,(word)rP);
		    if (rC>gC) Store(gC,rC,size,r0,gP,b0);
		    if (rC<bC) Store(rC,bC,size,rP,gP,b0);
		}
	    }
	    if (bC<up) {
		gC=Divide(bC,up,org->g,(word)gP);
		if (gC>bC) {
		    rC=Divide(bC,gC,org->r,(word)rP);
		    if (rC>bC) Store(bC,rC,size,r0,g0,bP);
		    if (rC<gC) Store(rC,gC,size,rP,g0,bP);
		}
		if (gC<up) {
		    rC=Divide(gC,up,org->r,(word)rP);
		    if (rC>gC) Store(gC,rC,size,r0,gP,bP);
		    if (rC<up) Store(rC,up,size,rP,gP,bP);
		}
	    }
	}
	if ((new=AllocCM(pixels))==NULL) return(NULL);

	for (ptr=cell, pixel=0; pixel<pixels; pixel++, ptr++) {
	    r=g=b=0.0;
	    for (i=ptr->lo; i<ptr->up; i++) {
		j=lut[i]; cnt=(double)pxl[j]; r+=((double)org->r[j]*cnt);
					      g+=((double)org->g[j]*cnt);
					      b+=((double)org->b[j]*cnt);
			  pxl[j]=pixel;
	    }
	    cnt=(double)ptr->number; new->r[pixel]=(word)(r/cnt);
				     new->g[pixel]=(word)(g/cnt);
				     new->b[pixel]=(word)(b/cnt);
	}
	for (x=0; x<image->xSize; x++)
	for (y=0; y<image->ySize; y++) {
	    pixel=pxl[GETPIXEL(image,x,y)]; PUTPIXEL(image,x,y,pixel);
	}

	free(cell); free(pxl); free(lut); FreeCM(org); return(new);
}
