
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"
#include "FFT.h"

#ifndef Float
#define Float	float
#endif

#ifndef	BPS
#define BPS	8
#endif

static int	Distance(arg,Nxyz)
char		*arg;
int		Nxyz;
{
	return (*arg=='-')?atoi(arg+1):(int)(atof(arg)*(double)Nxyz+0.5);
}

static Cell	PV(ac)
double		ac;
{
	return (Cell)((double)((1<<BPS)-1)*((ac<0.0)?0.0:(ac>1.0)?1.0:ac)+0.5);
}

#define LEN	512

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,Dx,Dy,Dz,
		Lx,Ly,H,V,Ix,z,y,x,z0,dz,Dx1,Dy1,len,H1,V1,Dz1;
	char	**names,desc[LEN],form[LEN],*name[1],str[LEN],
		*mae="memory allocation error.";
	Float	***P,***Pz,**Py,*Px,**Qy,*Qx;
	Complex	*Ex,*Ey,*W,**Fg,**Fh,*w;
	Cell
#ifdef	OM
		**new,***org,**o,
#else
		**org,**new,
#endif
		**v,*h;
	double	r,i,ac0,Nz0,Nyz0;

	if (argc!=4 && argc!=7)
	    Error("usage : sliceAC org name {ratioX ratioY ratioZ} new");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==4) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else
	    if ((Dx=Distance(argv[3],Nx))<=0 || Dx>Nx ||
		(Dy=Distance(argv[4],Ny))<=0 || Dy>Ny ||
		(Dz=Distance(argv[5],Nz))<=0 || Dz>Nz)
		Error("bad correlation distance.");

	printf("%d\t%d\t%d\t",Dx,Dy,Dz);

	if ((P=Pz=ALLOC(Float **,Dz))==NULL ||
	    (Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (W=ALLOC(Complex,(Lx>Ly)?Lx:Ly))==NULL ||
#ifdef	OM
	    (Fg=ALLOC(Complex *,Ly))==NULL ||
#else
	    (org=(Cell **)(Fg=ALLOC(Complex *,Ly)))==NULL ||
#endif
	    (new=(Cell **)(Fh=ALLOC(Complex *,Ly)))==NULL) Error(mae);

	H=Dx*2-1;
	V=Dy*2-1;

	for (z=0; z<Dz; z++) {
	    if ((*(Pz++)=Py=ALLOC(Float *,V))==NULL) Error(mae);

	    for (y=0; y<V; y++) {
		if ((*(Py++)=Px=ALLOC(Float,H))==NULL) Error(mae);

		for (x=0; x<H; x++) *(Px++)=0.0;
	    }
	}
	Ix=Lx/2+1;

	for (y=0; y<Ly; y++)
	    if ((Fg[y]=ALLOC(Complex,Ix))==NULL ||
		(Fh[y]=ALLOC(Complex,Ix))==NULL) Error(mae);

#ifdef	OM
	org=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,org);
#endif
	Dx1=Dx-1;
	Dy1=Dy-1;

	for (z0=0; z0<Nz; z0++)
	    for (z=z0, dz=0; z<Nz && dz<Dz; z++, dz++) {
#ifdef	OM
		o=org[z];
#else
		ReadImageFile(argv[1],names+z,Nx,Ny,1,&org);
#endif
		for (y=0; y<Ny; y++) {
		    for (w=W, x=0; x<Nx; x++, w++) {
#ifdef	OM
			w->r=(double)o[y][x];
#else
			w->r=(double)org[y][x];
#endif
			w->i=0.0;
		    }
		    for (; x<Lx; x++, w++) w->r=w->i=0.0;

		    FFT(Lx,Ex,W,-1);

		    for (w=W, x=0; x<Ix; x++) Fg[y][x]=(*(w++));
		}
		if (dz==0)
		    for (x=0; x<Ix; x++) {
			for (w=W, y=0; y<Ny; y++) *(w++)=Fg[y][x];

			for (; y<Ly; y++, w++) w->r=w->i=0.0;

			FFT(Ly,Ey,W,-1);

			for (w=W, y=0; y<Ly; y++, w++) {
			    Fh[y][x]=(*w);

			    w->r=w->r*w->r+w->i*w->i;
			    w->i=0.0;
			}
			FFT(Ly,Ey,W,1);

			for (w=W, y=0; y<Ly; y++) Fg[y][x]=(*(w++));
		    }
		else
		    for (x=0; x<Ix; x++) {
			for (w=W, y=0; y<Ny; y++) *(w++)=Fg[y][x];

			for (; y<Ly; y++, w++) w->r=w->i=0.0;

			FFT(Ly,Ey,W,-1);

			for (w=W, y=0; y<Ly; y++, w++) {
			    r=w->r;
			    i=w->i; w->r=r*Fh[y][x].r+i*Fh[y][x].i;
				    w->i=i*Fh[y][x].r-r*Fh[y][x].i;
			}
			FFT(Ly,Ey,W,1);

			for (w=W, y=0; y<Ly; y++) Fg[y][x]=(*(w++));
		    }

		Qy=P[dz];
		for (y=Ly-Dy1; y!=Dy; y=(y+1)%Ly) {
		    for (w=W, x=0; x<Ix; x++) *(w++)=Fg[y][x];

		    for (x=Lx-Ix; x>0; x--, w++) {
			w->r=Fg[y][x].r; w->i=(-Fg[y][x].i);
		    }
		    FFT(Lx,Ex,W,1);

		    Qx=(*(Qy++));
		    for (w=W+(x=Lx-Dx1); x<Lx; x++, w++) *(Qx++)+=w->r;
		    for (w=W, x=0; x<=Dx1; x++, w++) *(Qx++)+=w->r;
		}
	    }

	ac0=P[0][Dy1][Dx1]/((double)Nx*(double)Ny*(double)Nz);
	sprintf(desc,"%lf",ac0/((double)Lx*(double)Ly));
	puts(desc);

	len=1; for (z=Dz*2-1; z>=10; z/=10) ++len;
	sprintf(form,"%%0%dd.tif",len);
	name[0]=str;

	H1=H-1;
	V1=V-1;
	Dz1=Dz-1;

	for (Pz=P+Dz1, z=(-Dz1); z<0; z++, Pz--) {
	    Nz0=(double)(Nz+z)*ac0;
	    for (Py=(*Pz)+V1, v=new, y=(-Dy1); y<=Dy1; y++, v++, Py--) {
		Nyz0=(double)(Ny-abs(y))*Nz0;
		for (Px=(*Py)+H1, h=(*v), x=(-Dx1); x<=Dx1; x++, h++, Px--)
		    *h=PV(*Px/((double)(Nx-abs(x))*Nyz0));
	    }
	    sprintf(*name,form,z+Dz1);
	    StoreImageFile(argv[argc-1],name,&new,0,0,0,H1,V1,0,
			   GetImageBPS(&new,0,0,0,H1,V1,0),NULL,desc);
	}
	for (; z<=Dz1; z++, Pz++) {
	    Nz0=(double)(Nz-z)*ac0;
	    for (Py=(*Pz), v=new, y=(-Dy1); y<=Dy1; y++, v++, Py++) {
		Nyz0=(double)(Ny-abs(y))*Nz0;
		for (Px=(*Py), h=(*v), x=(-Dx1); x<=Dx1; x++, h++, Px++)
		    *h=PV(*Px/((double)(Nx-abs(x))*Nyz0));
	    }
	    sprintf(*name,form,z+Dz1);
	    StoreImageFile(argv[argc-1],name,&new,0,0,0,H1,V1,0,
			   GetImageBPS(&new,0,0,0,H1,V1,0),NULL,desc);
	}
	return 0;
}
