
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"
#include "FFT.h"

#ifndef Float
#define Float	float
#endif

#ifndef	BPS
#define BPS	8
#endif

static int	Distance(arg,Nxyz)
char		*arg;
int		Nxyz;
{
	return (*arg=='-')?atoi(arg+1):(int)(atof(arg)*(double)Nxyz+0.5);
}

static Cell	PV(ac)
double		ac;
{
	return (Cell)((double)((1<<BPS)-1)*((ac<0.0)?0.0:(ac>1.0)?1.0:ac)+0.5);
}

#define LEN	512

#define COMPLEX	struct complex

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,Dx,Dy,Dz,
		H,V,Lx,Ly,Lz,z,Ix,y,x,len,Dx1,Dy1,Dz1,H1,V1,bps;
	char	**names,str[LEN],desc[LEN],form[LEN],*name[1],
		*mae="memory allocation error.";
	Cell	**cell,**v,*h,c;
	Complex	*Ex,*Ey,*Ez,*W,*w;
	COMPLEX {
			Float	r,i;
		} **F;
	Float	**P;
	FILE	*file;
	long	offset;
	double	ac0,Nz0,Nyz0;

	if (argc!=5 && argc!=8)
	    Error("usage : sliceAC.VM org name {ratioX ratioY ratioZ} tmp new");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==5) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else
	    if ((Dx=Distance(argv[3],Nx))<=0 || Dx>Nx ||
		(Dy=Distance(argv[4],Ny))<=0 || Dy>Ny ||
		(Dz=Distance(argv[5],Nz))<=0 || Dz>Nz)
		Error("bad correlation distance.");

	printf("%d\t%d\t%d\t",Dx,Dy,Dz);

	H=Dx*2-1;
	V=Dy*2-1;

	cell=AllocateSliceMemory((Nx>H)?Nx:H,(Ny>V)?Ny:V);

	if ((Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (Lz=SetUpFFT(Nz+Dz,&Ez))==0 ||
	    (W=ALLOC(Complex,(Lx>Ly)?(Lx>Lz)?Lx:Lz:(Ly>Lz)?Ly:Lz))==NULL ||
	    (P=(Float **)(F=ALLOC(COMPLEX *,
				  z=(Ly>Nz)?Ly:Nz)))==NULL) Error(mae);

	Ix=Lx/2+1;

	for (y=0; y<z; y++)
	    if ((F[y]=ALLOC(COMPLEX,Ix))==NULL) Error(mae);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++) {
		for (w=W, x=0; x<Nx; x++, w++) {
		    w->r=(double)cell[y][x]; w->i=0.0;
		}
		for (; x<Lx; x++, w++) w->r=w->i=0.0;

		FFT(Lx,Ex,W,-1);

		for (w=W, x=0; x<Ix; x++, w++) {
		    F[y][x].r=w->r; F[y][x].i=w->i;
		}
	    }
	    for (x=0; x<Ix; x++) {
		for (w=W, y=0; y<Ny; y++, w++) {
		    w->r=F[y][x].r; w->i=F[y][x].i;
		}
		for (; y<Ly; y++, w++) w->r=w->i=0.0;

		FFT(Ly,Ey,W,-1);

		for (w=W, y=0; y<Ly; y++, w++) {
		    F[y][x].r=w->r; F[y][x].i=w->i;
		}
	    }
	    sprintf(str,"%s/%d",argv[argc-2],z);
	    if ((file=fopen(str,"wb"))==NULL)
		Error("can not create temporary file.");

	    for (y=0; y<Ly; y++)
		if (fwrite(F[y],sizeof(COMPLEX),(size_t)Ix,file)!=Ix)
		    Error("can not write temporary data.");

	    fclose(file);
	}
	for (y=0; y<Ly; y++) {
	    offset=(long)y*(long)Ix*sizeof(COMPLEX);

	    for (z=0; z<Nz; z++) {
		sprintf(str,"%s/%d",argv[argc-2],z);
		file=fopen(str,"rb");
		fseek(file,offset,SEEK_SET);
		fread(F[z],sizeof(COMPLEX),(size_t)Ix,file);
		fclose(file);
	    }
	    for (x=0; x<Ix; x++) {
		for (w=W, z=0; z<Nz; z++, w++) {
		    w->r=F[z][x].r; w->i=F[z][x].i;
		}
		for (; z<Lz; z++, w++) w->r=w->i=0.0;

		FFT(Lz,Ez,W,-1);

		for (w=W, z=0; z<Lz; z++, w++) {
		    w->r=w->r*w->r+w->i*w->i; w->i=0.0;
		}
		FFT(Lz,Ez,W,1);

		for (w=W, z=0; z<Dz; z++, w++) {
		    F[z][x].r=w->r; F[z][x].i=w->i;
		}
	    }
	    for (z=0; z<Dz; z++) {
		sprintf(str,"%s/%d",argv[argc-2],z);
		file=fopen(str,"rb+");
		fseek(file,offset,SEEK_SET);
		(void)fwrite(F[z],sizeof(COMPLEX),(size_t)Ix,file);
		fclose(file);
	    }
	}
	len=1; for (z=Dz*2-2; z>=10; z/=10) ++len;
	sprintf(form,"%%0%dd.tif",len);
	name[0]=str;

	Dx1=Dx-1; H1=H-1;
	Dy1=Dy-1; V1=V-1;
	Dz1=Dz-1;

	for (z=0; z<Dz; z++) {
	    sprintf(str,"%s/%d",argv[argc-2],z);
	    file=fopen(str,"rb");
	    for (y=0; y<Ly; y++) fread(F[y],sizeof(COMPLEX),(size_t)Ix,file);
	    fclose(file);

	    for (x=0; x<Ix; x++) {
		for (w=W, y=0; y<Ly; y++, w++) {
		    w->r=F[y][x].r; w->i=F[y][x].i;
		}
		FFT(Ly,Ey,W,1);

		for (w=W, y=0; y<Ly; y++, w++) {
		    F[y][x].r=w->r; F[y][x].i=w->i;
		}
	    }
	    for (y=0; y<Ly; y++) {
		for (w=W, x=0; x<Ix; x++, w++) {
		    w->r=F[y][x].r; w->i=F[y][x].i;
		}
		for (; x<Lx; x++, w++) {
		    w->r=F[y][Lx-x].r; w->i=(-F[y][Lx-x].i);
		}
		FFT(Lx,Ex,W,1);

		for (w=W, x=0; x<Lx; x++, w++) P[y][x]=w->r;
	    }
	    if (z==0) {
		ac0=P[0][0]/((double)Nx*(double)Ny*(double)Nz);
		sprintf(desc,"%lf",ac0/((double)Lx*(double)Ly*(double)Lz));
		puts(desc);
	    }
	    Nz0=(double)(Nz-z)*ac0;
	    for (v=cell, y=(-Dy1); y<0; y++, v++) {
		Nyz0=(double)(Ny+y)*Nz0;
		for (h=(*v), x=(-Dx1); x<0; x++, h++)
		    *h=PV(P[Ly+y][Lx+x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[Ly+y][x]/((double)(Nx-x)*Nyz0));
	    }
	    for (; y<=Dy1; y++, v++) {
		Nyz0=(double)(Ny-y)*Nz0;
		for (h=(*v), x=(-Dx1); x<0; x++, h++)
		    *h=PV(P[y][Lx+x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[y][x]/((double)(Nx-x)*Nyz0));
	    }
	    sprintf(*name,form,z+Dz1);
	    StoreImageFile(argv[argc-1],name,&cell,0,0,0,H1,V1,0,
			   bps=GetImageBPS(&cell,0,0,0,H1,V1,0),NULL,desc);

	    if (z!=0) {
		for (v=cell+V1, y=0; y<Dy1; y++, v--)
		for (h=(*v)+H1, x=0; x<=H1; x++, h--) {
		    c=cell[y][x]; cell[y][x]=(*h); *h=c;
		}
		for (h=(*v)+H1, x=0; x<Dx1; x++, h--) {
		    c=cell[y][x]; cell[y][x]=(*h); *h=c;
		}
		sprintf(*name,form,-z+Dz1);
		StoreImageFile(argv[argc-1],name,&cell,0,0,0,H1,V1,0,
			       bps,NULL,desc);
	    }
	}
	return 0;
}
