
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"
#include "FFT.h"

#ifndef Float
#define Float	float
#endif

#ifndef	BPS
#define BPS	8
#endif

static int	Distance(arg,Nxyz)
char		*arg;
int		Nxyz;
{
	return (*arg=='-')?atoi(arg+1):(int)(atof(arg)*(double)Nxyz+0.5);
}

static Cell	PV(ac)
double		ac;
{
	return (Cell)((double)((1<<BPS)-1)*((ac<0.0)?0.0:(ac>1.0)?1.0:ac)+0.5);
}

#define LEN	512

#define COMPLEX	struct complex

#define ALLOC(type,size)	(type *)malloc((size_t)(size)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,Dx,Dy,Dz,H,V,Lx,Ly,Lz,Ix,z,y,x,len,Dx1,Dy1,Dz1,H1,V1;
	char	**names,desc[LEN],form[LEN],*name[1],str[LEN],
		*mae="memory allocation error.";
	Cell	**cell,**v,*h;
	Complex	*Ex,*Ey,*Ez,*W,*w;
	COMPLEX	{
			Float	r,i;
		} ***F;
	Float	***P;
	double	ac0,Nz0,Nyz0;

	if (argc!=4 && argc!=7)
	    Error("usage : sliceAC org name {ratioX ratioY ratioZ} new");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==4) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else
	    if ((Dx=Distance(argv[3],Nx))<=0 || Dx>Nx ||
		(Dy=Distance(argv[4],Ny))<=0 || Dy>Ny ||
		(Dz=Distance(argv[5],Nz))<=0 || Dz>Nz)
		Error("bad correlation distance.");

	printf("%d\t%d\t%d\t",Dx,Dy,Dz);

	H=Dx*2-1;
	V=Dy*2-1;

	cell=AllocateSliceMemory((Nx>H)?Nx:H,(Ny>V)?Ny:V);

	if ((Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (Lz=SetUpFFT(Nz+Dz,&Ez))==0 ||
	    (W=ALLOC(Complex,(Lx>Ly)?(Lx>Lz)?Lx:Lz:(Ly>Lz)?Ly:Lz))==NULL ||
	    (P=(Float ***)(F=ALLOC(COMPLEX **,Nz)))==NULL) Error(mae);

	Ix=Lx/2+1;

	for (z=0; z<Nz; z++) {
	    if ((F[z]=ALLOC(COMPLEX *,Ly))==NULL) Error(mae);

	    for (y=0; y<Ly; y++)
		if ((F[z][y]=ALLOC(COMPLEX,Ix))==NULL) Error(mae);
	}
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++) {
		for (w=W, x=0; x<Nx; x++, w++) {
		    w->r=(double)cell[y][x]; w->i=0.0;
		}
		for (; x<Lx; x++, w++) w->r=w->i=0.0;

		FFT(Lx,Ex,W,-1);

		for (w=W, x=0; x<Ix; x++, w++) {
		    F[z][y][x].r=w->r; F[z][y][x].i=w->i;
		}
	    }
	    for (x=0; x<Ix; x++) {
		for (w=W, y=0; y<Ny; y++, w++) {
		    w->r=F[z][y][x].r; w->i=F[z][y][x].i;
		}
		for (; y<Ly; y++, w++) w->r=w->i=0.0;

		FFT(Ly,Ey,W,-1);

		for (w=W, y=0; y<Ly; y++, w++) {
		    F[z][y][x].r=w->r; F[z][y][x].i=w->i;
		}
	    }
	}
	for (y=0; y<Ly; y++)
	for (x=0; x<Ix; x++) {
	    for (w=W, z=0; z<Nz; z++, w++) {
		w->r=F[z][y][x].r; w->i=F[z][y][x].i;
	    }
	    for (; z<Lz; z++, w++) w->r=w->i=0.0;

	    FFT(Lz,Ez,W,-1);

	    for (w=W, z=0; z<Lz; z++, w++) {
		w->r=w->r*w->r+w->i*w->i; w->i=0.0;
	    }
	    FFT(Lz,Ez,W,1);

	    for (w=W, z=0; z<Dz; z++, w++) {
		F[z][y][x].r=w->r; F[z][y][x].i=w->i;
	    }
	}
	for (z=0; z<Dz; z++) {
	    for (x=0; x<Ix; x++) {
		for (w=W, y=0; y<Ly; y++, w++) {
		    w->r=F[z][y][x].r; w->i=F[z][y][x].i;
		}
		FFT(Ly,Ey,W,1);

		for (w=W, y=0; y<Ly; y++, w++) {
		    F[z][y][x].r=w->r; F[z][y][x].i=w->i;
		}
	    }
	    for (y=0; y<Ly; y++) {
		for (w=W, x=0; x<Ix; x++, w++) {
		    w->r=F[z][y][x].r; w->i=F[z][y][x].i;
		}
		for (; x<Lx; x++, w++) {
		    w->r=F[z][y][Lx-x].r; w->i=(-F[z][y][Lx-x].i);
		}
		FFT(Lx,Ex,W,1);

		for (w=W, x=0; x<Lx; x++, w++) P[z][y][x]=w->r;
	    }
	}
	ac0=P[0][0][0]/((double)Nx*(double)Ny*(double)Nz);
	sprintf(desc,"%lf",ac0/((double)Lx*(double)Ly*(double)Lz));
	puts(desc);

	len=1; for (z=Dz*2-2; z>=10; z/=10) ++len;
	sprintf(form,"%%0%dd.tif",len);
	name[0]=str;

	Dx1=Dx-1; H1=H-1;
	Dy1=Dy-1; V1=V-1;
	Dz1=Dz-1;

	for (z=(-Dz1); z<0; z++) {
	    Nz0=(double)(Nz+z)*ac0;
	    for (v=cell, y=(-Dy1); y<=0; y++, v++) {
		Nyz0=(double)(Ny+y)*Nz0;
		for (h=(*v), x=(-Dx1); x<=0; x++, h++)
		    *h=PV(P[-z][-y][-x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[-z][-y][Lx-x]/((double)(Nx-x)*Nyz0));
	    }
	    for (; y<=Dy1; y++, v++) {
		Nyz0=(double)(Ny-y)*Nz0;
		for (h=(*v), x=(-Dx1); x<=0; x++, h++)
		    *h=PV(P[-z][Ly-y][-x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[-z][Ly-y][Lx-x]/((double)(Nx-x)*Nyz0));
	    }
	    sprintf(*name,form,z+Dz1);
	    StoreImageFile(argv[argc-1],name,&cell,0,0,0,H1,V1,0,
			   GetImageBPS(&cell,0,0,0,H1,V1,0),NULL,desc);
	}
	for (; z<=Dz1; z++) {
	    Nz0=(double)(Nz-z)*ac0;
	    for (v=cell, y=(-Dy1); y<0; y++, v++) {
		Nyz0=(double)(Ny+y)*Nz0;
		for (h=(*v), x=(-Dx1); x<0; x++, h++)
		    *h=PV(P[z][Ly+y][Lx+x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[z][Ly+y][x]/((double)(Nx-x)*Nyz0));
	    }
	    for (; y<=Dy1; y++, v++) {
		Nyz0=(double)(Ny-y)*Nz0;
		for (h=(*v), x=(-Dx1); x<0; x++, h++)
		    *h=PV(P[z][y][Lx+x]/((double)(Nx+x)*Nyz0));

		for (; x<=Dx1; x++, h++)
		    *h=PV(P[z][y][x]/((double)(Nx-x)*Nyz0));
	    }
	    sprintf(*name,form,z+Dz1);
	    StoreImageFile(argv[argc-1],name,&cell,0,0,0,H1,V1,0,
			   GetImageBPS(&cell,0,0,0,H1,V1,0),NULL,desc);
	}
	return 0;
}
