
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "slice.h"
#include "bev.h"

#define EXTENSION_DEPTH	"d.tif"
#define EXTENSION_LABEL	"l.tif"
#define EXTENSION_SHADE	"s.tif"
#define EXTENSION_EDGE	"e.tif"

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

typedef unsigned short	Word;
typedef unsigned char	Byte;

static Word	**AllocateWordMemory(wid,hig)
int		wid,hig;
{
	Word	**word;
	int	x,y;
	char	*msg="word memory allocation error.";

	if ((word=(Word **)MA(hig,word))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((word[y]=(Word *)MA(wid,*word))==NULL) Error(msg);

	    for (x=0; x<wid; x++) word[y][x]=0;
	}
	return word;
}

#define AllocateCellMemory	AllocateSliceMemory

static Byte	**AllocateByteMemory(wid,hig)
int		wid,hig;
{
	Byte	**byte;
	int	x,y;
	char	*msg="byte memory allocation error.";

	if ((byte=(Byte **)MA(hig,byte))==NULL) Error(msg);

	for (y=0; y<hig; y++) {
	    if ((byte[y]=(Byte *)MA(wid,*byte))==NULL) Error(msg);

	    for (x=0; x<wid; x++) byte[y][x]=0;
	}
	return byte;
}

static void	ScanEdge(h1,v1,h2,v2,hMin,hMax)
int		h1,v1,h2,v2,*hMin,*hMax;
{
	int	wh,wv,sh,sv,m,p,e,h,v;

	if (h2>=h1) { wh=h2-h1; sh=1; } else { wh=h1-h2; sh=(-1); }
	if (v2>=v1) { wv=v2-v1; sv=1; } else { wv=v1-v2; sv=(-1); }
	if (wh>=wv) {
	    h2+=sh; m=2*wv; e=m-wh; p=e-wh;
	    for (v=v1, h=h1; h!=h2; h+=sh) {
		if (h<hMin[v]) hMin[v]=h;
		if (h>hMax[v]) hMax[v]=h;
		if (e>=0) { v+=sv; e+=p; } else e+=m;
	    }
	}
	else {
	    v2+=sv; m=2*wh; e=m-wv; p=e-wv;
	    for (h=h1, v=v1; v!=v2; v+=sv) {
		if (h<hMin[v]) hMin[v]=h;
		if (h>hMax[v]) hMax[v]=h;
		if (e>=0) { h+=sh; e+=p; } else e+=m;
	    }
	}
}

static void	PaintParallelogram(a,b,c,h0,v0,dh1,dv1,dh2,dv2,hMin,hMax,
#ifndef	DS
				   depth,cluster,gray,label,shade)
#else
				   depth)
#endif
double		a,b,c,h0,v0,dh1,dv1,dh2,dv2;
int		*hMin,*hMax;
Word		**depth;
#ifndef	DS
Cell		cluster,**label;
Byte		gray,**shade;
#endif
{
	int	hmin,hmax,vMin,vMax,h,v,d,
		h1=(int)floor(h0        )+1,v1=(int)floor(v0        )+1,
		h2=(int)floor(h0+dh1    )+1,v2=(int)floor(v0+dv1    )+1,
		h3=(int)floor(h0+dh1+dh2)+1,v3=(int)floor(v0+dv1+dv2)+1,
		h4=(int)floor(h0    +dh2)+1,v4=(int)floor(v0    +dv2)+1,
		h12=h2-h1,h23=h3-h2,h34=h4-h3,h41=h1-h4,
		v12=v2-v1,v23=v3-v2,v34=v4-v3,v41=v1-v4;
	double	dh,dv;

	hmin=hmax=h1; if (h2<hmin) hmin=h2; if (h2>hmax) hmax=h2;
		      if (h3<hmin) hmin=h3; if (h3>hmax) hmax=h3;
		      if (h4<hmin) hmin=h4; if (h4>hmax) hmax=h4;
	vMin=vMax=v1; if (v2<vMin) vMin=v2; if (v2>vMax) vMax=v2;
		      if (v3<vMin) vMin=v3; if (v3>vMax) vMax=v3;
		      if (v4<vMin) vMin=v4; if (v4>vMax) vMax=v4;

	for (v=vMin; v<=vMax; v++) { hMin[v]=hmax; hMax[v]=hmin; }

	ScanEdge(h1,v1,h2,v2,hMin,hMax);
	ScanEdge(h2,v2,h3,v3,hMin,hMax);
	ScanEdge(h3,v3,h4,v4,hMin,hMax);
	ScanEdge(h4,v4,h1,v1,hMin,hMax);

	for (dv=c+b*(double)((v=vMin)-1),
	     v1=h12*(v-v1)+v12*h1,
	     v2=h23*(v-v2)+v23*h2,
	     v3=h34*(v-v3)+v34*h3,
	     v4=h41*(v-v4)+v41*h4; v<=vMax; v1+=h12,
					    v2+=h23,
					    v3+=h34,
					    v4+=h41, dv+=b, v++)
	for (dh=dv+a*(double)((h=hMin[v])-1),
	     h1=v1-v12*h,
	     h2=v2-v23*h,
	     h3=v3-v34*h,
	     h4=v4-v41*h; h<=hMax[v]; h1-=v12,
				      h2-=v23,
				      h3-=v34,
				      h4-=v41, dh+=a, h++)
	    if ((h1>=0 && h2>=0 && h3>=0 && h4>=0) ||
		(h1<=0 && h2<=0 && h3<=0 && h4<=0))
#ifndef	DS
		if ((d=(int)floor(dh))>(int)depth[v][h]) {
		    depth[v][h]=(Word)d; label[v][h]=cluster; shade[v][h]=gray;
		}
#else
		depth[v][h]=((d=(int)floor(dh))>1)?(Word)d:1;
#endif
}

#ifndef	DS
static void	DrawLine(h0,v0,d0,dh,dv,dd,depth,edge)
double		h0,v0,d0,dh,dv,dd;
Word		**depth;
Byte		**edge;
{
	int	wh,sh,wv,sv,m,p,e,h,v,
		h1=(int)floor(h0)+1,h2=(int)floor(h0+dh)+1,
		v1=(int)floor(v0)+1,v2=(int)floor(v0+dv)+1,
		d1=(int)floor(d0)+1,d2=(int)floor(d0+dd)+1;
	double	g,d;

	if (h1==h2 && v1==v2) {
	    if (d1+1>=(int)depth[v1][h1] ||
		d2+1>=(int)depth[v1][h1]) edge[v1][h1]=1;
	}
	else {
	    if (h2>h1) { wh=h2-h1; sh=1; } else { wh=h1-h2; sh=(-1); }
	    if (v2>v1) { wv=v2-v1; sv=1; } else { wv=v1-v2; sv=(-1); }
	    g=(double)(d2-d1); d=(double)d1+1.0;
	    if (wh>=wv) {
		h2+=sh; m=2*wv; e=m-wh; p=e-wh; g/=(double)wh;
		for (v=v1, h=h1; h!=h2; h+=sh, d+=g) {
		    if ((int)ceil(d)>=(int)depth[v][h]) edge[v][h]=1;
		    if (e>=0) { v+=sv; e+=p; } else e+=m;
		}
	    }
	    else {
		v2+=sv; m=2*wh; e=m-wv; p=e-wv; g/=(double)wv;
		for (h=h1, v=v1; v!=v2; v+=sv, d+=g) {
		    if ((int)ceil(d)>=(int)depth[v][h]) edge[v][h]=1;
		    if (e>=0) { h+=sh; e+=p; } else e+=m;
		}
	    }
	}
}
#endif

#define NH	bev.Nh
#define NV	bev.Nv
#define ND	bev.Nd
#define HX	bev.hx
#define HY	bev.hy
#define HZ	bev.hz
#define VX	bev.vx
#define VY	bev.vy
#define VZ	bev.vz
#define DX	bev.dx
#define DY	bev.dy
#define DZ	bev.dz
#define H0	bev.h0
#define V0	bev.v0
#define D0	bev.d0

typedef struct {
		int	Nh,Nv;
		Word	**depth;
		Cell	**label;
		Byte	**shade,**edge;
	} Map;

#define DEPTH	(map.depth)
#define LABEL	(map.label)
#define SHADE	(map.shade)
#define EDGE	(map.edge)

static Map	BirdEyeViewMapper(Nx,Ny,Nz,x1,y1,z1,x2,y2,z2,cell,Sx,Sy,Sz,
				  vLo,vLa,lLo,lLa)
int		Nx,Ny,Nz,x1,y1,z1,x2,y2,z2;
Cell		***cell;
double		Sx,Sy,Sz,vLo,vLa,lLo,lLa;
{
	BEV	bev;
	Map	map;
	int	*hMin,*hMax,x,y,z;
	double	lx,ly,lz,hx,hy,hz,vx,vy,vz,dx,dy,dz,
		xx,yy,zz,g,a,b,h0,v0,dh1,dv1,dh2,dv2,h1,v1,h2,v2,hh,vv,dd;
#ifndef	DS
	int	gray,x3,y3,z3;
	double	dh3,dv3,dh4,dv4,c,d0,d1,d2;
#else
	Word	**depth,d,D,dE,dW,dN,dS;
	double	c1,c2,xh,yh,zh,xv,yv,zv,xd,yd,zd,dh,dv,nx,ny,nz,ip;
	int	h,v;
#endif

	bev=SetUpBEV(Sx*(-0.5),
		     Sy*(-0.5),
		     Sz*(-0.5),Sx*((double)Nx-0.5),
			       Sy*((double)Ny-0.5),
			       Sz*((double)Nz-0.5),vLo,vLa);

	if (ND>(1<<(8*sizeof(Word)))) Error("too large image depth.");

	map.Nh=NH;
	map.Nv=NV; DEPTH=AllocateWordMemory(NH,NV);
		   LABEL=AllocateCellMemory(NH,NV);
		   SHADE=AllocateByteMemory(NH,NV);
		   EDGE =AllocateByteMemory(NH,NV);

	if ((hMin=(int *)MA(NV,hMin))==NULL ||
	    (hMax=(int *)MA(NV,hMax))==NULL)
	    Error("raster scan memory allocation error.");

	LightDirection(lLo,lLa,&lx,&ly,&lz);

	hx=Sx*HX; vx=Sx*VX; dx=Sx*DX;
	hy=Sy*HY; vy=Sy*VY; dy=Sy*DY;
	hz=Sz*HZ; vz=Sz*VZ; dz=Sz*DZ;

#ifndef	DS
#define PP()	PaintParallelogram(a,b,c,h2,v2,dh1,dv1,dh2,dv2,hMin,hMax, \
				   DEPTH,cell[z][y][x],(Byte)gray,LABEL,SHADE)

if (DX!=0.0 && (g=HY*VZ-HZ*VY)!=0.0) {
    a=(VZ*DY-VY*DZ)/g; b=(DZ*HY-DY*HZ)/g;
    yy=(double)y1-0.5; dh3=yy*(dh1=hy); dv3=yy*(dv1=vy);
    zz=(double)z1-0.5; dh4=zz*(dh2=hz); dv4=zz*(dv2=vz);
    if (DX<0.0) {
	gray=(lx<0.0)?(int)(0.5-255.0*lx):0;
	for (x=x1; x<=x2; x++) {
	    xx=(double)x-0.5; c=xx*dx-D0-a*(h0=xx*hx-H0)-b*(v0=xx*vx-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, y=y1; y<=y2; y++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, z=z1; z<=z2; z++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (x==x1 || !cell[z][y][x-1])) PP();
	}
    }
    else {
	gray=(lx>0.0)?(int)(0.5+255.0*lx):0;
	for (x=x2; x>=x1; x--) {
	    xx=(double)x+0.5; c=xx*dx-D0-a*(h0=xx*hx-H0)-b*(v0=xx*vx-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, y=y1; y<=y2; y++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, z=z1; z<=z2; z++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (x==x2 || !cell[z][y][x+1])) PP();
	}
    }
}
if (DY!=0.0 && (g=HZ*VX-HX*VZ)!=0.0) {
    a=(VX*DZ-VZ*DX)/g; b=(DX*HZ-DZ*HX)/g;
    zz=(double)z1-0.5; dh3=zz*(dh1=hz); dv3=zz*(dv1=vz);
    xx=(double)x1-0.5; dh4=xx*(dh2=hx); dv4=xx*(dv2=vx);
    if (DY<0.0) {
	gray=(ly<0.0)?(int)(0.5-255.0*ly):0;
	for (y=y1; y<=y2; y++) {
	    yy=(double)y-0.5; c=yy*dy-D0-a*(h0=yy*hy-H0)-b*(v0=yy*vy-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, z=z1; z<=z2; z++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, x=x1; x<=x2; x++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (y==y1 || !cell[z][y-1][x])) PP();
	}
    }
    else {
	gray=(ly>0.0)?(int)(0.5+255.0*ly):0;
	for (y=y2; y>=y1; y--) {
	    yy=(double)y+0.5; c=yy*dy-D0-a*(h0=yy*hy-H0)-b*(v0=yy*vy-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, z=z1; z<=z2; z++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, x=x1; x<=x2; x++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (y==y2 || !cell[z][y+1][x])) PP();
	}
    }
}
if (DZ!=0.0 && (g=HX*VY-HY*VX)!=0.0) {
    a=(VY*DX-VX*DY)/g; b=(DY*HX-DX*HY)/g;
    xx=(double)x1-0.5; dh3=xx*(dh1=hx); dv3=xx*(dv1=vx);
    yy=(double)y1-0.5; dh4=yy*(dh2=hy); dv4=yy*(dv2=vy);
    if (DZ<0.0) {
	gray=(lz<0.0)?(int)(0.5-255.0*lz):0;
	for (z=z1; z<=z2; z++) {
	    zz=(double)z-0.5; c=zz*dz-D0-a*(h0=zz*hz-H0)-b*(v0=zz*vz-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, x=x1; x<=x2; x++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, y=y1; y<=y2; y++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (z==z1 || !cell[z-1][y][x])) PP();
	}
    }
    else {
	gray=(lz>0.0)?(int)(0.5+255.0*lz):0;
	for (z=z2; z>=z1; z--) {
	    zz=(double)z+0.5; c=zz*dz-D0-a*(h0=zz*hz-H0)-b*(v0=zz*vz-V0)+1.0;
	    for (h1=h0+dh3, v1=v0+dv3, x=x1; x<=x2; x++, h1+=dh1, v1+=dv1)
	    for (h2=h1+dh4, v2=v1+dv4, y=y1; y<=y2; y++, h2+=dh2, v2+=dv2)
		if (cell[z][y][x] && (z==z2 || !cell[z+1][y][x])) PP();
	}
    }
}

	x3=x2+1; xx=(double)x1-0.5;
	y3=y2+1; yy=(double)y1-0.5;
	z3=z2+1; zz=(double)z1-0.5; hh=xx*hx+yy*hy+zz*hz-H0;
				    vv=xx*vx+yy*vy+zz*vz-V0;
				    dd=xx*dx+yy*dy+zz*dz-D0;

	for (h0=hh, v0=vv, d0=dd, z=z1; z<=z2; z++, h0+=hz, v0+=vz, d0+=dz)
	for (h1=h0, v1=v0, d1=d0, y=y1; y<=y3; y++, h1+=hy, v1+=vy, d1+=dy)
	for (h2=h1, v2=v1, d2=d1, x=x1; x<=x3; x++, h2+=hx, v2+=vx, d2+=dx)
	    switch(((x!=x1 && y!=y1 && cell[z][y-1][x-1])?1:0)+
		   ((x!=x1 && y!=y3 && cell[z][y  ][x-1])?2:0)+
		   ((x!=x3 && y!=y1 && cell[z][y-1][x  ])?4:0)+
		   ((x!=x3 && y!=y3 && cell[z][y  ][x  ])?8:0)) {
		case 1 : case 2 : case  4 : case  6 : case  7 :
		case 8 : case 9 : case 11 : case 13 : case 14 :
		    DrawLine(h2,v2,d2,hz,vz,dz,DEPTH,EDGE); break;
	    }

	for (h0=hh, v0=vv, d0=dd, y=y1; y<=y2; y++, h0+=hy, v0+=vy, d0+=dy)
	for (h1=h0, v1=v0, d1=d0, x=x1; x<=x3; x++, h1+=hx, v1+=vx, d1+=dx)
	for (h2=h1, v2=v1, d2=d1, z=z1; z<=z3; z++, h2+=hz, v2+=vz, d2+=dz)
	    switch(((z!=z1 && x!=x1 && cell[z-1][y][x-1])?1:0)+
		   ((z!=z1 && x!=x3 && cell[z-1][y][x  ])?2:0)+
		   ((z!=z3 && x!=x1 && cell[z  ][y][x-1])?4:0)+
		   ((z!=z3 && x!=x3 && cell[z  ][y][x  ])?8:0)) {
		case 1 : case 2 : case  4 : case  6 : case  7 :
		case 8 : case 9 : case 11 : case 13 : case 14 :
		    DrawLine(h2,v2,d2,hy,vy,dy,DEPTH,EDGE); break;
	    }

	for (h0=hh, v0=vv, d0=dd, x=x1; x<=x2; x++, h0+=hx, v0+=vx, d0+=dx)
	for (h1=h0, v1=v0, d1=d0, z=z1; z<=z3; z++, h1+=hz, v1+=vz, d1+=dz)
	for (h2=h1, v2=v1, d2=d1, y=y1; y<=y3; y++, h2+=hy, v2+=vy, d2+=dy)
	    switch(((y!=y1 && z!=z1 && cell[z-1][y-1][x])?1:0)+
		   ((y!=y1 && z!=z3 && cell[z  ][y-1][x])?2:0)+
		   ((y!=y3 && z!=z1 && cell[z-1][y  ][x])?4:0)+
		   ((y!=y3 && z!=z3 && cell[z  ][y  ][x])?8:0)) {
		case 1 : case 2 : case  4 : case  6 : case  7 :
		case 8 : case 9 : case 11 : case 13 : case 14 :
		    DrawLine(h2,v2,d2,hx,vx,dx,DEPTH,EDGE); break;
	    }

#else
#define PP(c,h0,v0,depth)	\
	PaintParallelogram(a,b,c,h0,v0,dh1,dv1,dh2,dv2,hMin,hMax,depth)

	depth=AllocateWordMemory(NH,NV);

	if (DX!=0.0 && (g=HY*VZ-HZ*VY)!=0.0) {
	    a=(VZ*DY-VY*DZ)/g; b=(DZ*HY-DY*HZ)/g;
	    yy=(double)(y2-y1+1); dh1=yy*hy; dv1=yy*vy;
	    zz=(double)(z2-z1+1); dh2=zz*hz; dv2=zz*vz;
	    yy=(double)y1-0.5;
	    zz=(double)z1-0.5; h0=yy*hy+zz*hz; v0=yy*vy+zz*vz;
	    xx=(double)x1-0.5; c1=xx*dx-D0-a*(h1=xx*hx-H0)-b*(v1=xx*vx-V0)+1.0;
	    xx=(double)x2+0.5; c2=xx*dx-D0-a*(h2=xx*hx-H0)-b*(v2=xx*vx-V0)+1.0;
	    if (DX<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	if (DY!=0.0 && (g=HZ*VX-HX*VZ)!=0.0) {
	    a=(VX*DZ-VZ*DX)/g; b=(DX*HZ-DZ*HX)/g;
	    zz=(double)(z2-z1+1); dh1=zz*hz; dv1=zz*vz;
	    xx=(double)(x2-x1+1); dh2=xx*hx; dv2=xx*vx;
	    zz=(double)z1-0.5;
	    xx=(double)x1-0.5; h0=zz*hz+xx*hx; v0=zz*vz+xx*vx;
	    yy=(double)y1-0.5; c1=yy*dy-D0-a*(h1=yy*hy-H0)-b*(v1=yy*vy-V0)+1.0;
	    yy=(double)y2+0.5; c2=yy*dy-D0-a*(h2=yy*hy-H0)-b*(v2=yy*vy-V0)+1.0;
	    if (DY<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	if (DZ!=0.0 && (g=HX*VY-HY*VX)!=0.0) {
	    a=(VY*DX-VX*DY)/g; b=(DY*HX-DX*HY)/g;
	    xx=(double)(x2-x1+1); dh1=xx*hx; dv1=xx*vx;
	    yy=(double)(y2-y1+1); dh2=yy*hy; dv2=yy*vy;
	    xx=(double)x1-0.5;
	    yy=(double)y1-0.5; h0=xx*hx+yy*hy; v0=xx*vx+yy*vy;
	    zz=(double)z1-0.5; c1=zz*dz-D0-a*(h1=zz*hz-H0)-b*(v1=zz*vz-V0)+1.0;
	    zz=(double)z2+0.5; c2=zz*dz-D0-a*(h2=zz*hz-H0)-b*(v2=zz*vz-V0)+1.0;
	    if (DZ<0.0) {
		PP(c1,h0+h1,v0+v1,DEPTH); PP(c2,h0+h2,v0+v2,depth);
	    }
	    else {
		PP(c1,h0+h1,v0+v1,depth); PP(c2,h0+h2,v0+v2,DEPTH);
	    }
	}
	for (v=NV-2; v>0; v--)
	for (h=NH-2; h>0; h--)
	    if ((d=DEPTH[v][h])>(D=depth[v][h]) && D==0)
		depth[v][h]=d;
	    else if (d<D)
		DEPTH[v][h]=D;

	hx=HX/Sx; hy=HY/Sy; hz=HZ/Sz; hh=H0+(double)(NH-2)-0.5;
	vx=VX/Sx; vy=VY/Sy; vz=VZ/Sz; vv=V0+(double)(NV-2)-0.5;
	dx=DX/Sx; dy=DY/Sy; dz=DZ/Sz; dd=D0               -1.0;

	for (xv=hx*hh+vx*vv+dx*dd+0.5,
	     yv=hy*hh+vy*vv+dy*dd+0.5,
	     zv=hz*hh+vz*vv+dz*dd+0.5,
	     v=NV-2; v>0; v--, xv-=vx, yv-=vy, zv-=vz) {
	    for (xh=xv, yh=yv, zh=zv, h=NH-2; h>0; h--, xh-=hx, yh-=hy, zh-=hz)
		if ((d=DEPTH[v][h])>0) {
		    DEPTH[v][h]=0;
		    for (xd=xh+dx*(double)d,
			 yd=yh+dy*(double)d,
			 zd=zh+dz*(double)d,
			 D=depth[v][h]; d>=D; d--, xd-=dx, yd-=dy, zd-=dz)
			if ((x=(int)floor(xd))>=x1 && x<=x2 &&
			    (y=(int)floor(yd))>=y1 && y<=y2 &&
			    (z=(int)floor(zd))>=z1 && z<=z2 && cell[z][y][x]) {
			    DEPTH[v][h]=d; LABEL[v][h]=cell[z][y][x]; break;
			}
		}

	    free(depth[v]);
	}
	free(depth);

	for (v=NV-2; v>0; v--)
	for (h=NH-2; h>0; h--)
	    if (DEPTH[v][h]) {
		dh=((double)(dE=DEPTH[v][h+1])-(double)(dW=DEPTH[v][h-1]))/2.0;
		dv=((double)(dN=DEPTH[v+1][h])-(double)(dS=DEPTH[v-1][h]))/2.0;

		if ((ip=(nx=DX-HX*dh-VX*dv)*lx+
			(ny=DY-HY*dh-VY*dv)*ly+
			(nz=DZ-HZ*dh-VZ*dv)*lz)>0.0)
		    SHADE[v][h]=(Byte)(255.0*ip/sqrt(nx*nx+ny*ny+nz*nz));

		if (dE==0 || dW==0 || dN==0 || dS==0) EDGE[v][h]=1;
	    }
#endif
	free(hMin); free(hMax); return map;
}

#define PI	((bps==1)?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK)
#define SPP	1
#define PC	PLANARCONFIG_CONTIG
#define TSF	TIFFSetField

static TIFF	*OpenTIFF(path,wid,hig,bps,desc,tsl)
char		*path,*desc;
int		wid,hig,bps;
uint8		**tsl;
{
	TIFF	*tiff=TIFFOpen(path,"w");
	int	tss;

	if (tiff==NULL ||
	    !TSF(tiff,TIFFTAG_IMAGEWIDTH,     (uint32)wid) ||
	    !TSF(tiff,TIFFTAG_IMAGELENGTH,    (uint32)hig) ||
	    !TSF(tiff,TIFFTAG_PHOTOMETRIC,    (uint16)PI ) ||
	    !TSF(tiff,TIFFTAG_BITSPERSAMPLE,  (uint16)bps) ||
	    !TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)SPP) ||
	    !TSF(tiff,TIFFTAG_PLANARCONFIG,   (uint16)PC ) ||
	    (*tsl=(uint8 *)malloc(tss=TIFFScanlineSize(tiff)))==NULL) {
	    fprintf(stderr,"%s : ",path); Error("image file not stored.");
	}
	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(uint32)((tss>8192)?1:8192/tss));
#ifdef	LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	TSF(tiff,TIFFTAG_IMAGEDESCRIPTION,desc);

	return tiff;
}

#define WriteTIFF(tiff,tsl,v,path)	\
	if (TIFFWriteScanline(tiff,tsl,(uint32)v,(uint16)0)<0) { \
	    fprintf(stderr,"%s : ",path); Error("write data error."); \
	}

static void	StoreWordImage(path,word,Nh,Nv,bps,desc)
char		*path,*desc;
Word		**word;
int		Nh,Nv,bps;
{
	int	v,h;
	Word	*data,mask,msb=(Word)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=word[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(word[v]);
	}
	free(word); free(tsl); TIFFClose(tiff);
}

static void	StoreCellImage(path,cell,Nh,Nv,bps,desc)
char		*path,*desc;
Cell		**cell;
int		Nh,Nv,bps;
{
	int	v,h;
	Cell	*data,mask,msb=(Cell)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=cell[v];
	    if (bps==8)
		for (ptr=tsl, h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, h=0; h<Nh; h++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, h=0; h<Nh; h++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(cell[v]);
	}
	free(cell); free(tsl); TIFFClose(tiff);
}

static void	StoreByteImage(path,byte,Nh,Nv,bps,desc)
char		*path,*desc;
Byte		**byte;
int		Nh,Nv,bps;
{
	int	v,h;
	Byte	*data,mask,msb=(Byte)(1<<(bps-1));
	uint8	*tsl,*ptr,bit;
	TIFF	*tiff=OpenTIFF(path,Nh,Nv,bps,desc,&tsl);

	for (v=0; v<Nv; v++) {
	    data=byte[v]; ptr=tsl;
	    if (bps==8)
		for (h=0; h<Nh; h++) *(ptr++)=(uint8)*(data++);
	    else
		for (bit=0x80, h=0; h<Nh; h++, data++)
		    for (mask=msb; mask!=0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    WriteTIFF(tiff,tsl,v,path); free(byte[v]);
	}
	free(byte); free(tsl); TIFFClose(tiff);
}

#define STR_LEN		512
#define XYZ_LEN		256
#define NAME_LEN	512
#define DESC_LEN	256
#define PATH_LEN	512

#define RADIAN	(M_PI/180.0)

#define BPS(limit)	\
	(limit<2)?1:(limit<4)?2:(limit<16)?4:(limit<256)?8:(limit<65536)?16:32

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,bps,xMin,yMin,zMin,xMax,yMax,zMax,x,y,z,
		idx,x1,y1,z1,x2,y2,z2,h,v,limit;
	char	**names,str[STR_LEN],xyz[6][XYZ_LEN],name[NAME_LEN],
		desc[DESC_LEN],path[PATH_LEN];
	double	Sx,Sy,Sz,vLo,vLa,lLo,lLa;
	Cell	***cell;
	Map	map;

	if (argc!=6)
	    Error("usage : sliceBEVM directory nameFile scaleX scaleY scaleZ");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,NULL);

	if ((Sx=atof(argv[3]))<=0.0 ||
	    (Sy=atof(argv[4]))<=0.0 ||
	    (Sz=atof(argv[5]))<=0.0) Error("bad scaling factor.");

	cell=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	xMin=Nx-1; yMin=Ny-1; zMin=Nz-1; xMax=0; yMax=0; zMax=0;
	for (z=0; z<Nz; z++) for (y=0; y<Ny; y++) for (x=0; x<Nx; x++)
	    if (cell[z][y][x]) {
		if (x<xMin) xMin=x; if (x>xMax) xMax=x;
		if (y<yMin) yMin=y; if (y>yMax) yMax=y;
		if (z<zMin) zMin=z; if (z>zMax) zMax=z;
	    }

while (fgets(str,STR_LEN,stdin)!=NULL) {
    switch(sscanf(str,"%lf %lf %lf %lf %s %s %s %s %s %s %s",
		      &vLo,&vLa,&lLo,&lLa,
		      xyz[0],xyz[1],xyz[2],xyz[3],xyz[4],xyz[5],name)) {
	case  5 :
	    x1=xMin; y1=yMin; z1=zMin; x2=xMax; y2=yMax; z2=zMax;
	    for (idx=0; (name[idx]=xyz[0][idx])!='\0'; idx++) ; break;
	case 11 :
	    if ((x1=atoi(xyz[0]))<xMin) x1=xMin;
	    if ((y1=atoi(xyz[1]))<yMin) y1=yMin;
	    if ((z1=atoi(xyz[2]))<zMin) z1=zMin;
	    if ((x2=atoi(xyz[3]))>xMax) x2=xMax;
	    if ((y2=atoi(xyz[4]))>yMax) y2=yMax;
	    if ((z2=atoi(xyz[5]))>zMax) z2=zMax;
	    if (x1>x2 || y1>y2 || z1>z2) continue; break;
	default :
	    continue;
    }
    map=BirdEyeViewMapper(Nx,Ny,Nz,x1,y1,z1,x2,y2,z2,cell,Sx,Sy,Sz,
			  vLo*RADIAN,vLa*RADIAN,lLo*RADIAN,lLa*RADIAN);

    sprintf(desc,
	    "%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\n",
	    -0.5,-0.5,-0.5,(double)Nx-0.5,(double)Ny-0.5,(double)Nz-0.5,
	    Sx,Sy,Sz,vLo,vLa);

    limit=0;
    for (v=0; v<map.Nv; v++)
    for (h=0; h<map.Nh; h++)
	if ((int)DEPTH[v][h]>limit) limit=(int)DEPTH[v][h];

    sprintf(path,"%s.%s",name,EXTENSION_DEPTH);
    StoreWordImage(path,DEPTH,map.Nh,map.Nv,BPS(limit),desc);

    sprintf(path,"%s.%s",name,EXTENSION_LABEL);
    StoreCellImage(path,LABEL,map.Nh,map.Nv,bps,       desc);

    sprintf(path,"%s.%s",name,EXTENSION_SHADE);
    StoreByteImage(path,SHADE,map.Nh,map.Nv,8,         desc);

    sprintf(path,"%s.%s",name,EXTENSION_EDGE );
    StoreByteImage(path,EDGE, map.Nh,map.Nv,1,         desc);
}
	return 0;
}
