
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "slice.h"

#define STR_LEN	512
#define RADIAN	(M_PI/180.0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,ooi,i,j,k,Nh1,Nv1,Nd1,h1,v1,d1,h2,v2,d2,l,m,n;
	char	**names,*name[1],str[STR_LEN],format[STR_LEN],desc[STR_LEN];
	double	lon,cLon,sLon,lat,cLat,sLat,rot,cRot,sRot,
		hx,hy,hz,vx,vy,vz,dx,dy,dz,x,y,z,h0,v0,d0,h,v,d,
		xh,yh,zh,xv,yv,zv,xd,yd,zd,Sx=1.0,Sy=1.0,Sz=1.0;
	Cell	***cell,**slice;

	if (argc!=8 && argc!=11)
Error("usage : sliceIR orgDir nameFile {Sx Sy Sz} lon lat rot ooi newDir");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==11)
	    if ((Sx=atof(argv[3]))<=0.0 ||
		(Sy=atof(argv[4]))<=0.0 ||
		(Sz=atof(argv[5]))<=0.0) Error("bad scaling factor.");

	if ((ooi=atoi(argv[argc-2]))<0 || ooi>=(1<<(8*sizeof(Cell))))
	    Error("bad pixel value for out of image area.");

	cell=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	lon=atof(argv[argc-5]); cLon=cos(lon*RADIAN); sLon=sin(lon*RADIAN);
	lat=atof(argv[argc-4]); cLat=cos(lat*RADIAN); sLat=sin(lat*RADIAN);
	rot=atof(argv[argc-3]); cRot=cos(rot*RADIAN); sRot=sin(rot*RADIAN);
	hx=  sLon*cRot+cLon*sLat*sRot ;
	hy=(-cLon*cRot+sLon*sLat*sRot);
	hz=(               -cLat*sRot);
	vx=(-sLon*sRot+cLon*sLat*cRot);
	vy=  cLon*sRot+sLon*sLat*cRot ;
	vz=(               -cLat*cRot);
	dx=	       cLon*cLat      ;
	dy=	       sLon*cLat      ;
	dz=		    sLat      ;

	for (i=0; i<=Nx; i+=Nx) {
	    x=Sx*((double)i-0.5); xh=x*hx; xv=x*vx; xd=x*dx;
	    for (j=0; j<=Ny; j+=Ny) {
		y=Sy*((double)j-0.5); yh=xh+y*hy; yv=xv+y*vy; yd=xd+y*dy;
		for (k=0; k<=Nz; k+=Nz) {
		    z=Sz*((double)k-0.5); zh=yh+z*hz; zv=yv+z*vz; zd=yd+z*dz;

		    if (i==0 && j==0 && k==0) {
			h0=h=zh; v0=v=zv; d0=d=zd;
		    }
		    else {
			if (zh<h0) h0=zh; else if (zh>h) h=zh;
			if (zv<v0) v0=zv; else if (zv>v) v=zv;
			if (zd<d0) d0=zd; else if (zd>d) d=zd;
		    }
		}
	    }
	}
	Nh1=h1=(int)ceil(h-h0)-1;
	Nv1=v1=(int)ceil(v-v0)-1;
	Nd1=d1=(int)ceil(d-d0)-1; h2=v2=d2=0;

	for (i=0; i<=Nd1; i++) {
	    d=d0+0.5+(double)i; xd=d*dx; yd=d*dy; zd=d*dz;
	    for (j=0; j<=Nv1; j++) {
		v=v0+0.5+(double)j; xv=xd+v*vx; yv=yd+v*vy; zv=zd+v*vz;
		for (k=0; k<=Nh1; k++) {
		    h=h0+0.5+(double)k; xh=xv+h*hx; yh=yv+h*hy; zh=zv+h*hz;

		    if ((l=(int)floor(zh/Sz+0.5))>=0 && l<Nz &&
			(m=(int)floor(yh/Sy+0.5))>=0 && m<Ny &&
			(n=(int)floor(xh/Sx+0.5))>=0 && n<Nx) {
			if (i<d1) d1=i; if (i>d2) d2=i;
			if (j<v1) v1=j; if (j>v2) v2=j;
			if (k<h1) h1=k; if (k>h2) h2=k;
		    }
		}
	    }
	}
	h0+=(double)h1; Nh1=h2-h1;
	v0+=(double)v1; Nv1=v2-v1;
	d0+=(double)d1; Nd1=d2-d1;

	slice=AllocateSliceMemory(Nh1+1,Nv1+1);

	name[0]=str;
	sprintf(*name,"%d",Nd1);
	sprintf(format,"%%0%dd.tif",strlen(*name));

	sprintf(desc,"%d\t%d\t%d\t%lg\t%lg\t%lg\n%lg\t%lg\t%lg\t%d\t%d\t%d\n",
		     Nx,Ny,Nz,Sx,Sy,Sz,lon,lat,rot,Nh1+1,Nv1+1,Nd1+1);

	for (i=0; i<=Nd1; i++) {
	    d=d0+0.5+(double)i; xd=d*dx; yd=d*dy; zd=d*dz;
	    for (j=0; j<=Nv1; j++) {
		v=v0+0.5+(double)j; xv=xd+v*vx; yv=yd+v*vy; zv=zd+v*vz;
		for (k=0; k<=Nh1; k++) {
		    h=h0+0.5+(double)k; xh=xv+h*hx; yh=yv+h*hy; zh=zv+h*hz;

slice[j][k]=((l=(int)floor(zh/Sz+0.5))>=0 && l<Nz &&
	     (m=(int)floor(yh/Sy+0.5))>=0 && m<Ny &&
	     (n=(int)floor(xh/Sx+0.5))>=0 && n<Nx)?cell[l][m][n]:(Cell)ooi;

		}
	    }
	    sprintf(*name,format,i);
	    StoreImageFile(argv[argc-1],name,&slice,0,0,0,Nh1,Nv1,0,
			   GetImageBPS(&slice,0,0,0,Nh1,Nv1,0),NULL,desc);
	}
	return 0;
}
