
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

static double	Min(x,y,z)
double		x,y,z;
{
	double	min=x+y+z,xy=x+y,yz=y+z,zx=z+x;

	if (min>xy) min=xy; if (min>x) min=x;
	if (min>yz) min=yz; if (min>y) min=y;
	if (min>zx) min=zx; if (min>z) min=z; return (min<0.0)?min:0.0;
}

static double	Max(x,y,z)
double		x,y,z;
{
	double	max=x+y+z,xy=x+y,yz=y+z,zx=z+x;

	if (max<xy) max=xy; if (max<x) max=x;
	if (max<yz) max=yz; if (max<y) max=y;
	if (max<zx) max=zx; if (max<z) max=z; return (max>0.0)?max:0.0;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,lower,upper,z,y,x,start=1;
	char	**names;
	Cell	**cell;
	double	x0,y0,z0,dx,dy,dz,hx,hy,hz,vx,vy,vz,h,v,d,h1,v1,d1,h2,v2,d2;

	if (argc!=11 && argc!=14)
Error("usage : sliceIRC dir name x0 y0 z0 dx dy dz {Mx My Mz} lower upper");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);
	cell=AllocateSliceMemory(Nx,Ny);

	x0=atof(argv[3]); y0=atof(argv[4]); z0=atof(argv[5]);
	dx=atof(argv[6]); dy=atof(argv[7]); dz=atof(argv[8]);

	if (dx==0.0 && dy==0.0) {
	    if (dz==0.0) Error("bad axial vector.");

	    hx=1.0; vy=dz=(dz>0.0)?1.0:-1.0; hy=hz=vx=vz=0.0;
	}
	else
{
	double	xy=dx*dx+dy*dy,
		rxy=sqrt(xy),son=dy/rxy,con=dx/rxy,
		rxyz=sqrt(xy+dz*dz),sat=dz/rxyz,cat=rxy/rxyz;

	hx=(-son)    ; hy=  con     ; hz=0.0;
	vx=(-con)*sat; vy=(-son)*sat; vz=cat;
	dx=  con *cat; dy=  son *cat; dz=sat;
}
	if (argc==14)
{
	double	Mx,My,Mz;

	if ((Mx=atof(argv[ 9]))<=0.0 ||
	    (My=atof(argv[10]))<=0.0 ||
	    (Mz=atof(argv[11]))<=0.0) Error("bad magnification factor.");

	hx*=Mx; hy*=My; hz*=Mz;
	vx*=Mx; vy*=My; vz*=Mz;
	dx*=Mx; dy*=My; dz*=Mz;
}
	if ((lower=atoi(argv[argc-2]))>
	    (upper=atoi(argv[argc-1]))) Error("bad value range.");

	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++)
		if (cell[y][x]>=lower && cell[y][x]<=upper) {
		    h=hx*(double)x+hy*(double)y+hz*(double)z;
		    v=vx*(double)x+vy*(double)y+vz*(double)z;
		    d=dx*(double)x+dy*(double)y+dz*(double)z;
		    if (start) {
			h1=h2=h; v1=v2=v; d1=d2=d; start=0;
		    }
		    else {
			if (h<h1) h1=h; else if (h>h2) h2=h;
			if (v<v1) v1=v; else if (v>v2) v2=v;
			if (d<d1) d1=d; else if (d>d2) d2=d;
		    }
		}
	}
	if (start) Error("no pixel.");

	h=hx*x0+hy*y0+hz*z0;
	v=vx*x0+vy*y0+vz*z0;
	d=dx*x0+dy*y0+dz*z0;

	printf("%g\t%g\t%g\t%g\t%g\t%g\n",
	       h1+Min(hx,hy,hz)-h,v1+Min(vx,vy,vz)-v,d1+Min(dx,dy,dz)-d,
	       h2+Max(hx,hy,hz)-h,v2+Max(vx,vy,vz)-v,d2+Max(dx,dy,dz)-d);

	return 0;
}
