
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"

#ifndef TABLE_RATIO
#define TABLE_RATIO	(27.0/256.0)
#endif

#define TABLE_SIZE	((~(1<<(8*sizeof(int)-1)))/sizeof(int))
#define PV_STORAGE	(Cell)(~0)

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]); return label;
}

typedef struct {
		int	pixels;
		union {
			int	label;
			struct {
				short	x1,y1,z1,x2,y2,z2;
			} area;
		} u;
	} Data;

static int	Compare(ptr1,ptr2)
Data		*ptr1,*ptr2;
{
	return (ptr1->pixels==ptr2->pixels)?ptr1->u.label-ptr2->u.label:
					    ptr2->pixels -ptr1->pixels;
}

#define MA(noe,top)	malloc((size_t)(noe)*sizeof(*(top)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,bps,lo,up,**last,y,x,size,*table,labels,z,
		pixels,new,old,*tmp,clusters,label,cluster;
	char	**names,**descs,*amae="array memory allocation error.";
	Cell	**cell,pv;
	double	table_size;
	Data	*data,*ptr;

	if (argc!=5 && argc!=6)
	    Error("usage : sliceMCL orgDir nameFile lower upper {newDir}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,&descs);

	if ((lo=atoi(argv[3]))<0 || (up=atoi(argv[4]))>=1<<bps || lo>up)
	    Error("bad pixel value range.");

	cell=AllocateSliceMemory(Nx,Ny);

	if ((last=(int **)MA(Ny+1,last))==NULL) Error(amae);

	++last;
	for (y=(-1); y<Ny; y++) {
	    if ((last[y]=(int *)MA(Nx+1,last[y]))==NULL) Error(amae);

	    ++last[y];
	    for (x=(-1); x<Nx; x++) last[y][x]=0;
	}

	table_size=TABLE_RATIO*(double)Nx*(double)Ny*(double)Nz+1.0;
	size=(table_size>(double)TABLE_SIZE)?TABLE_SIZE:(int)table_size;

	for (; size>0 && (table=(int *)MA(size,table))==NULL; size/=2)
	fprintf(stderr,"no allocatable memory for table of size %d.\n",size);

	if (size<=0) Error("table memory allocation error.");

	table[0]=0; labels=1;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		pixels=1;
		if ((int)cell[y][x]<lo || (int)cell[y][x]>up)
		    pixels+=table[new=0];
		else
		    if (last[y][x]) {
			pixels+=table[new=Search(table,last[y][x])];
			if (last[y-1][x] &&
			   (old=Search(table,last[y-1][x]))!=new) {
			    pixels+=table[old]; table[old]=(-new);
			}
			if (last[y][x-1] &&
			   (old=Search(table,last[y][x-1]))!=new) {
			    pixels+=table[old]; table[old]=(-new);
			}
		    }
		    else
			if (last[y-1][x]) {
			    pixels+=table[new=Search(table,last[y-1][x])];
			    if (last[y][x-1] && (old=last[y][x-1])!=new) {
				pixels+=table[old]; table[old]=(-new);
			    }
			}
			else
			    if (last[y][x-1])
				pixels+=table[new=last[y][x-1]];
			    else {
				if (labels==size) Error("too many labels.");

				new=labels++;
			    }

		if ((table[last[y][x]=new]=pixels)<0) Error("too many pixels.");
	    }
	}

	if ((tmp=(int *)realloc(table,labels*sizeof(int)))!=NULL) table=tmp;

	clusters=1;
	for (label=1; label<labels; label++) if (table[label]>0) ++clusters;

	if ((data=(Data *)MA(clusters,data))==NULL)
	    Error("data memory allocation error.");

	ptr=data;
	for (label=0; label<labels; label++)
	    if (table[label]>=0) {
		ptr->pixels=table[ptr->u.label=label]; ++ptr;
	    }
	    else
		table[label]=(-Search(table,label));

	if (clusters>1) qsort(data+1,clusters-1,sizeof(*data),Compare);

	for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++) {
	    table[ptr->u.label]=cluster;
	    ptr->u.area.x1=(short)(Nx-1); ptr->u.area.x2=0;
	    ptr->u.area.y1=(short)(Ny-1); ptr->u.area.y2=0;
	    ptr->u.area.z1=(short)(Nz-1); ptr->u.area.z2=0;
	}
	for (label=1; label<labels; label++)
	    if (table[label]<0) table[label]=table[-table[label]];

	if (argc==6) { 
	    pv=(clusters-1>PV_STORAGE)?PV_STORAGE:(Cell)(clusters-1);
	    bps=(pv<2)?1:(pv<4)?2:(pv<16)?4:(pv<256)?8:(pv<65536)?16:32;
	}

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) last[y][x]=0;

	label=1;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		if ((int)cell[y][x]<lo || (int)cell[y][x]>up)
		    cluster=0;
		else if (last[y][x])
		    cluster=last[y][x];
		else if (last[y-1][x])
		    cluster=last[y-1][x];
		else if (last[y][x-1])
		    cluster=last[y][x-1];
		else
		    cluster=table[label++];

		ptr=data+(last[y][x]=cluster);
		if (x<(int)ptr->u.area.x1) ptr->u.area.x1=(short)x;
		if (x>(int)ptr->u.area.x2) ptr->u.area.x2=(short)x;
		if (y<(int)ptr->u.area.y1) ptr->u.area.y1=(short)y;
		if (y>(int)ptr->u.area.y2) ptr->u.area.y2=(short)y;
		if (z<(int)ptr->u.area.z1) ptr->u.area.z1=(short)z;
		if (z>(int)ptr->u.area.z2) ptr->u.area.z2=(short)z;

		if (argc==6)
		    cell[y][x]=(cluster>PV_STORAGE)?PV_STORAGE:(Cell)cluster;
	    }
	    if (argc==6)
		StoreImageFile(argv[5],names+z,&cell,
			       0,0,0,Nx-1,Ny-1,0,bps,descs+z,NULL);
	}

	printf("%d\n",clusters);
	for (ptr=data, cluster=0; cluster<clusters; cluster++, ptr++)
	    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
		   ptr->pixels,(int)ptr->u.area.x1,
			       (int)ptr->u.area.y1,
			       (int)ptr->u.area.z1,
			       (int)ptr->u.area.x2,
			       (int)ptr->u.area.y2,
			       (int)ptr->u.area.z2);

	return 0;
}
