
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

typedef struct {
		int	Nh,Nv;
		double	hx,hy,hz,h0,
			vx,vy,vz,v0,
			dx,dy,dz;
	} TP;	/* transform parameters */

#define NH	tp.Nh
#define NV	tp.Nv
#define HX	tp.hx
#define HY	tp.hy
#define HZ	tp.hz
#define H0	tp.h0
#define VX	tp.vx
#define VY	tp.vy
#define VZ	tp.vz
#define V0	tp.v0
#define DX	tp.dx
#define DY	tp.dy
#define DZ	tp.dz

#define MINMAX(exp)	hv=exp; if (hv<min) min=hv; if (hv>max) max=hv

static TP	SetUpTP(Nx,Ny,Nz,Sx,Sy,Sz,lon,lat,rot)
int		Nx,Ny,Nz;
double		Sx,Sy,Sz,lon,lat,rot;
{
	TP	tp;
	double	min,max,hv,clon=cos(lon),clat=cos(lat),crot=cos(rot),
			   slon=sin(lon),slat=sin(lat),srot=sin(rot),
			   xMin=Sx*(-0.5),xMax=Sx*((double)Nx-0.5),
			   yMin=Sy*(-0.5),yMax=Sy*((double)Ny-0.5),
			   zMin=Sz*(-0.5),zMax=Sz*((double)Nz-0.5);

	HX=  slon*crot+clon*slat*srot ;
	HY=(-clon*crot+slon*slat*srot);
	HZ=(               -clat*srot);

	min=max=xMin*HX+yMin*HY+zMin*HZ;
	MINMAX(xMax*HX+yMin*HY+zMin*HZ);
	MINMAX(xMin*HX+yMax*HY+zMin*HZ);
	MINMAX(xMax*HX+yMax*HY+zMin*HZ);
	MINMAX(xMin*HX+yMin*HY+zMax*HZ);
	MINMAX(xMax*HX+yMin*HY+zMax*HZ);
	MINMAX(xMin*HX+yMax*HY+zMax*HZ);
	MINMAX(xMax*HX+yMax*HY+zMax*HZ); H0=min; NH=(int)ceil(max-min);

	VX=(-slon*srot+clon*slat*crot);
	VY=  clon*srot+slon*slat*crot ;
	VZ=(               -clat*crot);

	min=max=xMin*VX+yMin*VY+zMin*VZ;
	MINMAX(xMax*VX+yMin*VY+zMin*VZ);
	MINMAX(xMin*VX+yMax*VY+zMin*VZ);
	MINMAX(xMax*VX+yMax*VY+zMin*VZ);
	MINMAX(xMin*VX+yMin*VY+zMax*VZ);
	MINMAX(xMax*VX+yMin*VY+zMax*VZ);
	MINMAX(xMin*VX+yMax*VY+zMax*VZ);
	MINMAX(xMax*VX+yMax*VY+zMax*VZ); V0=min; NV=(int)ceil(max-min);

	DX=clon*clat;
	DY=slon*clat;
	DZ=     slat; return tp;
}

typedef	struct {
		int	Nh,Nv,h1,v1,h2,v2;
		Cell	**cell;
	} Slice;

#define H1	slice.h1
#define V1	slice.v1
#define H2	slice.h2
#define V2	slice.v2
#define CELL	(slice.cell)

static Slice	GenerateNewSlice(Nx,Ny,Nz,cell,Sx,Sy,Sz,ooi,x,y,z,lon,lat,rot)
int		Nx,Ny,Nz,ooi;
Cell		***cell;
double		Sx,Sy,Sz,x,y,z,lon,lat,rot;
{
	TP	tp;
	Slice	slice;
	int	i,j,k,l,m;
	double	d,xd,yd,zd,v,xv,yv,zv,h,xh,yh,zh;

	tp=SetUpTP(Nx,Ny,Nz,Sx,Sy,Sz,lon,lat,rot);

	H1=NH; H2=(-1); slice.Nh=NH;
	V1=NV; V2=(-1); slice.Nv=NV; CELL=AllocateSliceMemory(NH,NV);

	d=Sx*x*DX+Sy*y*DY+Sz*z*DZ; xd=d*DX; yd=d*DY; zd=d*DZ;
	for (i=0; i<NV; i++) {
	    v=V0+0.5+(double)i; xv=xd+v*VX; yv=yd+v*VY; zv=zd+v*VZ;
	    for (j=0; j<NH; j++) {
		h=H0+0.5+(double)j; xh=xv+h*HX; yh=yv+h*HY; zh=zv+h*HZ;

		if ((k=(int)floor(zh/Sz+0.5))>=0 && k<Nz &&
		    (l=(int)floor(yh/Sy+0.5))>=0 && l<Ny &&
		    (m=(int)floor(xh/Sx+0.5))>=0 && m<Nx) {
		    CELL[i][j]=cell[k][l][m]; if (j<H1) H1=j; if (j>H2) H2=j;
					      if (i<V1) V1=i; if (i>V2) V2=i;
		}
		else
		    CELL[i][j]=(Cell)ooi;
	    }
	}
	return slice;
}

static void	FreeSliceMemory(ySize,cell)
int		ySize;
Cell		**cell;
{
	int	y;

	for (y=0; y<ySize; y++) free(cell[y]); free(cell);
}

#define LINE_LEN	1024
#define PATH_LEN	512
#define DESC_LEN	256

#define RAD	(M_PI/180.0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,ooi;
	char	**names,line[LINE_LEN],path[PATH_LEN],desc[DESC_LEN];
	double	x,y,z,lon,lat,rot,Sx=1.0,Sy=1.0,Sz=1.0;
	Cell	***cell;
	Slice	slice;

	if (argc!=4 && argc!=7)
Error("usage : sliceNSG directory nameFile {scaleX scaleY scaleZ} outOfImage");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if (argc==7)
	    if ((Sx=atof(argv[3]))<=0.0 ||
		(Sy=atof(argv[4]))<=0.0 ||
		(Sz=atof(argv[5]))<=0.0) Error("bad scaling factor.");

	if ((ooi=atoi(argv[argc-1]))<0 || ooi>=(1<<(8*sizeof(Cell))))
	    Error("bad out of image pixel value.");

	cell=AllocateImageMemory(Nx,Ny,Nz);

	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	while (fgets(line,LINE_LEN,stdin)!=NULL)
	    if (sscanf(line,"%lf %lf %lf %lf %lf %lf %s",
		       &x,&y,&z,&lon,&lat,&rot,path)==7)
{
	if (x<=(-0.5) || x>=(double)Nx-0.5 ||
	    y<=(-0.5) || y>=(double)Ny-0.5 ||
	    z<=(-0.5) || z>=(double)Nz-0.5) {
	    fprintf(stderr,"%s : bad reference point.\n",line); continue;
	}

	slice=GenerateNewSlice(Nx,Ny,Nz,cell,Sx,Sy,Sz,ooi,
			       x,y,z,lon*RAD,lat*RAD,rot*RAD);

	sprintf(desc,
		"%d\t%d\t%d\t%lg\t%lg\t%lg\n%lg\t%lg\t%lg\t%lg\t%lg\t%lg\n",
		Nx,Ny,Nz,Sx,Sy,Sz,x,y,z,lon,lat,rot);
	StoreSliceFile(path,CELL,H1,V1,H2,V2,
		       GetImageBPS(&CELL,H1,V1,0,H2,V2,0),desc);

	FreeSliceMemory(slice.Nv,CELL);
}
	return 0;
}
