
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

static double	OvoidExtentXYZ(A,B,C)
double		A,B,C;
{
	return sqrt(A*A+B*B+C*C);
}

#define FORMAT_LEN	512

#define RADIAN	(M_PI/180.0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,label,z,y,x,length,x1,y1,z1,x2,y2,z2;
	char	**names,format[FORMAT_LEN],**descs=NULL,
		*nmae="name memory allocation error.";
	Cell	***cell;
	double	x0,y0,z0,lon,lat,rot,A,B,C,
		cLon,sLon,cLat,sLat,cRot,sRot,xa,xb,xc,ya,yb,yc,za,zb,zc,
		r,ax,bx,cx,ay,by,cy,az,bz,cz,Sx=1.0,Sy=1.0,Sz=1.0;

	if (argc!=4 && argc!=6 && argc!=7 && argc!=9)
Error("usage : sliceO (orgDir nameFile)|(Nx Ny Nz bg) {Sx Sy Sz} newDir");

	if (argc==4 || argc==7) {
	    ScanNameFile(argv[1],argv[2],&Nz,&names);
	    CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,&descs);
	    cell=AllocateImageMemory(Nx,Ny,Nz);
	    ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);
	}
	else {
	    if ((Nx=atoi(argv[1]))<=0 ||
		(Ny=atoi(argv[2]))<=0 ||
		(Nz=atoi(argv[3]))<=0) Error("bad image size.");

	    if ((label=atoi(argv[4]))<0 || label>=1<<(8*sizeof(Cell)))
		Error("bad background label.");

	    cell=AllocateImageMemory(Nx,Ny,Nz);
	    for (z=0; z<Nz; z++)
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) cell[z][y][x]=(Cell)label;

	    if ((names=(char **)malloc(Nz*sizeof(*names)))==NULL) Error(nmae);

	    for (length=1, z=10; z<Nz; z*=10, length++);
	    sprintf(format,"%%0%dd.tif",length);
	    length+=5; /* ".tif\0" */

	    for (z=0; z<Nz; z++) {
		if ((names[z]=(char *)malloc(length))==NULL) Error(nmae);

		sprintf(names[z],format,z);
	    }
	}

	if (argc==7 || argc==9)
	    if ((Sx=atof(argv[argc-4]))<=0.0 ||
		(Sy=atof(argv[argc-3]))<=0.0 ||
		(Sz=atof(argv[argc-2]))<=0.0) Error("bad scaling factor.");

	while (scanf("%d %lf %lf %lf %lf %lf %lf %lf %lf %lf",
		     &label,&x0,&y0,&z0,&lon,&lat,&rot,&A,&B,&C)==10) {
	    if (label<0 || label>=1<<(8*sizeof(Cell))) {
		fputs("bad foreground (ovoid) label.\n",stderr); continue;
	    }
	    if (A<=0.0 || B<=0.0 || C<=0.0) {
		fputs("bad ovoid radius.\n",stderr); continue;
	    }
	    cLon=cos(lon*RADIAN); sLon=sin(lon*RADIAN);
	    cLat=cos(lat*RADIAN); sLat=sin(lat*RADIAN);
	    cRot=cos(rot*RADIAN); sRot=sin(rot*RADIAN);
	    xa=  sLon*cRot+cLon*sLat*sRot ;
	    xb=(-sLon*sRot+cLon*sLat*cRot);
	    xc=		   cLon*cLat      ;
	    ya=(-cLon*cRot+sLon*sLat*sRot);
	    yb=( cLon*sRot+sLon*sLat*cRot);
	    yc=		   sLon*cLat      ;
	    za= 	       -cLat*sRot ;
	    zb= 	       -cLat*cRot ;
	    zc=			sLat	  ;

	    x0/=Sx; r=OvoidExtentXYZ(A*xa,B*xb,C*xc)/Sx;
	    x1=(int)floor(x0-r); if (x1< 0 ) x1=0;
	    x2=(int)ceil( x0+r); if (x2>=Nx) x2=Nx-1;
	    y0/=Sy; r=OvoidExtentXYZ(A*ya,B*yb,C*yc)/Sy;
	    y1=(int)floor(y0-r); if (y1< 0 ) y1=0;
	    y2=(int)ceil( y0+r); if (y2>=Ny) y2=Ny-1;
	    z0/=Sz; r=OvoidExtentXYZ(A*za,B*zb,C*zc)/Sz;
	    z1=(int)floor(z0-r); if (z1< 0 ) z1=0;
	    z2=(int)ceil( z0+r); if (z2>=Nz) z2=Nz-1;

	    xa*=(Sx/A); xb*=(Sx/B); xc*=(Sx/C);
	    ya*=(Sy/A); yb*=(Sy/B); yc*=(Sy/C);
	    za*=(Sz/A); zb*=(Sz/B); zc*=(Sz/C);

	    for (x=x1; x<=x2; x++) {
		r=(double)x-x0; ax=xa*r; bx=xb*r; cx=xc*r;
		for (y=y1; y<=y2; y++) {
		    r=(double)y-y0; ay=ax+ya*r; by=bx+yb*r; cy=cx+yc*r;
		    for (z=z1; z<=z2; z++) {
			r=(double)z-z0; az=ay+za*r; bz=by+zb*r; cz=cy+zc*r;
			if (az*az+bz*bz+cz*cz<=1.0) cell[z][y][x]=(Cell)label;
		    }
		}
	    }
	}
	StoreImageFile(argv[argc-1],names,
		       cell,0,0,0,Nx-1,Ny-1,Nz-1,
		       GetImageBPS(cell,0,0,0,Nx-1,Ny-1,Nz-1),
		       descs,NULL);
	return 0;
}
