
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

#ifndef	SEGMENT
#define SEGMENT		180
#endif

#ifndef	ITERATION
#define ITERATION	1024
#endif

static double	xx,yy,zz,xy,yz,zx,L0;

#define VAL	vec[0]
#define LON	vec[1]
#define LAT	vec[2]

#define LON1	(-M_PI)
#define LON2	M_PI
#define LAT1	0.0
#define LAT2	(M_PI/2.0)

static void	EvaluateAxis(vec)
double		*vec;
{
	double	co,so,co2,so2,cso,ca,sa,ca2,sa2,csa;

	if (LON>LON1 && LON<=LON2 && LAT>=LAT1 && LAT<=LAT2) {
	    co=cos(LON); so=sin(LON); co2=co*co; so2=so*so; cso=co*so; 
	    ca=cos(LAT); sa=sin(LAT); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
#ifdef	OBLATE
	    VAL=((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#else
	    VAL=L0-
		((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#endif
	}
	else
	    VAL=L0;
}

typedef	struct {
		int	N;
		double	x0,y0,z0,xx,yy,zz,xy,yz,zx;
	} Moment;

typedef struct {
		double	a,b,c,lon,lat,rot;
	} Axis;

static double	can0[3],can1[3],can2[3],*can[3]={can0,can1,can2};

static Axis	FitOvoid(Sx,Sy,Sz,moment)
double		Sx,Sy,Sz;
Moment		moment;
{
	Axis	axis;
	double	val,lon,lat,vec[3],step=M_PI/(double)SEGMENT,
		co,so,co2,so2,cso,
		ca,sa,ca2,sa2,csa,H2,V2,HV,HmV,rot2,dHV,HV0,a,b;

	xx=Sx*Sx*moment.xx;
	yy=Sy*Sy*moment.yy;
	zz=Sz*Sz*moment.zz;
	xy=Sx*Sy*moment.xy;
	yz=Sy*Sz*moment.yz;
	zx=Sz*Sx*moment.zx;

	val=L0=xx+yy+zz; lon=LON2; lat=LAT2;
	for (LON=LON2; LON> LON1; LON-=step)
	for (LAT=LAT2; LAT>=LAT1; LAT-=step) {
	    EvaluateAxis(vec);
	    if (VAL<val) {
		val=VAL; lon=LON; lat=LAT;
	    }
	}
	can0[1]=can2[1]=lon; can1[1]=lon+step;
	can0[2]=can1[2]=lat; can2[2]=lat-step;

	Simplex(2,can,EvaluateAxis,ITERATION,vec);

	axis.lon=LON;
	axis.lat=LAT;
#ifdef	OBLATE
	axis.c=sqrt(5.0*VAL/(double)moment.N);
#else
	axis.c=sqrt(5.0*(L0-VAL)/(double)moment.N);
#endif
	co=cos(axis.lon); so=sin(axis.lon); co2=co*co; so2=so*so; cso=co*so;
	ca=cos(axis.lat); sa=sin(axis.lat); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
	H2=xx*so2+yy*co2-2.0*xy*cso;
	V2=(xx*co2+yy*so2+2.0*xy*cso)*sa2-2.0*(zx*co+yz*so)*csa+zz*ca2;
	HV=((xx-yy)*cso-xy*(co2-so2))*sa-(zx*so-yz*co)*ca;
	HmV=(H2-V2)/2.0;
	rot2=atan2(HV,HmV);
	dHV=HmV*cos(rot2)+HV*sin(rot2);
	HV0=(H2+V2)/2.0;

#ifdef	OBLATE
	if (dHV>=0.0) {
#else
	if (dHV<=0.0) {
#endif
	    b=HV0-dHV;
	    a=HV0+dHV; if (rot2<0.0) rot2+=(M_PI*2.0);
	}
	else {
	    b=HV0+dHV;
	    a=HV0-dHV; rot2+=M_PI;
	}
	axis.rot=rot2/2.0; axis.b=sqrt(5.0*b/(double)moment.N);
			   axis.a=sqrt(5.0*a/(double)moment.N);

	return axis;
}

#define DEGREE	(180.0/M_PI)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,bps,depth,index,z,x,y;
	char	**names;
	double	Sx=1.0,Sy=1.0,Sz=1.0;
	Cell	**cell;
	Moment	*moment,*ptr;
	Axis	axis;

	if (argc!=3 && argc!=6)
	    Error("usage : sliceOF directory nameFile {scaleX scaleY scaleZ}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,NULL);

	if (argc==6)
	    if ((Sx=atof(argv[3]))<=0.0 ||
		(Sy=atof(argv[4]))<=0.0 ||
		(Sz=atof(argv[5]))<=0.0) Error("bad scaling factor.");

	cell=AllocateSliceMemory(Nx,Ny);

	depth=1<<bps;
	if ((moment=(Moment *)malloc(depth*sizeof(*moment)))==NULL)
	    Error("moment memory allocation error.");

	for (ptr=moment, index=0; index<depth; index++, ptr++) {
	    ptr->N=0; ptr->x0=ptr->y0=ptr->z0=
		      ptr->xx=ptr->yy=ptr->zz=
		      ptr->xy=ptr->yz=ptr->zx=0.0;
	}
	depth=0;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		index=(int)cell[y][x]; if (index>depth) depth=index;

		ptr=moment+index;
		++(ptr->N);
		ptr->x0+=(double)x;
		ptr->y0+=(double)y;
		ptr->z0+=(double)z;
		ptr->xx+=((double)x*(double)x+1.0/12.0);
		ptr->yy+=((double)y*(double)y+1.0/12.0);
		ptr->zz+=((double)z*(double)z+1.0/12.0);
		ptr->xy+=((double)x*(double)y);
		ptr->yz+=((double)y*(double)z);
		ptr->zx+=((double)z*(double)x);
	    }
	}

	for (ptr=moment, index=0; index<=depth; index++, ptr++)
	    if (ptr->N) {
		ptr->x0/=(double)ptr->N;
		ptr->y0/=(double)ptr->N;
		ptr->z0/=(double)ptr->N;
		ptr->xx-=(ptr->x0*ptr->x0*(double)ptr->N);
		ptr->yy-=(ptr->y0*ptr->y0*(double)ptr->N);
		ptr->zz-=(ptr->z0*ptr->z0*(double)ptr->N);
		ptr->xy-=(ptr->x0*ptr->y0*(double)ptr->N);
		ptr->yz-=(ptr->y0*ptr->z0*(double)ptr->N);
		ptr->zx-=(ptr->z0*ptr->x0*(double)ptr->N);

		axis=FitOvoid(Sx,Sy,Sz,*ptr);

	printf(
"%d\t%d\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lg\t%lf\n",
	       index,
	       ptr->N,
	       ptr->x0,ptr->y0,ptr->z0,
	       Sx*ptr->x0,Sy*ptr->y0,Sz*ptr->z0,
	       axis.lon*DEGREE,axis.lat*DEGREE,axis.rot*DEGREE,
	       axis.a,axis.b,axis.c,
	       axis.a*axis.b*axis.c*M_PI*4.0/3.0);
	    }

	return 0;
}
