
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"

static char	**AllocateFlagMemory(Nx,Ny)
int		Nx,Ny;
{
	char	**flag,*fmae="flag memory allocation error.";
	int	x,y;

	if ((flag=(char **)malloc(Ny*sizeof(*flag)))==NULL) Error(fmae);

	for (y=0; y<Ny; y++) {
	    if ((flag[y]=(char *)malloc(Nx))==NULL) Error(fmae);

	    for (x=0; x<Nx; x++) flag[y][x]=0;
	}
	return flag;
}

#define CUE_LENGTH_RATIO	4.4

typedef struct {
		int	x,y;
	} Cue;

static int	len,cnt;
static Cue	*top,*end,*head;

static void	PushCue(x,y)
int		x,y;
{
	if (++cnt>len) Error("cue overflow.");

	head->x=x; head->y=y; if (++head==end) head=top;
}

static int	Nx,Ny;
static char	**flag;

#define OCCUPIED(x,y)	(x>=0 && x<Nx && y>=0 && y<Ny && flag[y][x]==1)

static void	LabelBackground(x,y)
int		x,y;
{
	Cue	*tail;
	int	yP,yM,S,N,dx,S0,N0,x0,xC,SC,NC;

	head=tail=top;
	for (cnt=0; cnt>=0; cnt--) {
	    if (OCCUPIED(x,y)) {
		flag[y][x]=0; yP=y+1; S=OCCUPIED(x,yP);
			      yM=y-1; N=OCCUPIED(x,yM);
		for (dx=1; dx>=(-1); dx-=2) {
		    S0=S; N0=N;
		    for (xC=x, x0=x+dx; OCCUPIED(x0,y); xC=x0, x0+=dx) {
			flag[y][x0]=0; SC=S0; S0=OCCUPIED(x0,yP);
				       NC=N0; N0=OCCUPIED(x0,yM);
			if (SC && !S0) PushCue(xC,yP);
			if (NC && !N0) PushCue(xC,yM);
		    }
		    if (S0) PushCue(xC,yP);
		    if (N0) PushCue(xC,yM);
		}
	    }
	    x=tail->x; y=tail->y; if (++tail==end) tail=top;
	}
}

#define LABEL(x,y)	if (flag[y][x]==1) LabelBackground(x,y)

static void	PareObjectSkin(cell,thres,thick,bg)
Cell		**cell;
int		thres,thick,bg;
{
	int	x,y,t,h=Nx-1,v=Ny-1;

	if (bg) {
	    for (x=0; x<Nx; x++)
	    for (y=0; y<Ny; y++) flag[y][x]=((int)cell[y][x]<thres)?1:2;

	    for (x=0; x<Nx; x++) { LABEL(x,0); LABEL(x,v); }
	    for (y=0; y<Ny; y++) { LABEL(0,y); LABEL(h,y); }

	    for (x=0; x<Nx; x++)
	    for (y=0; y<Ny; y++) if (flag[y][x]==2) flag[y][x]=1;
	}
	else
	    for (x=0; x<Nx; x++)
	    for (y=0; y<Ny; y++) flag[y][x]=((int)cell[y][x]<thres)?0:1;

	for (t=0; t<thick; t++, h--, v--) {
	    for (x=t; x<=h; x++)
	    for (y=t; y<=v; y++)
		if (flag[y][x])
		    if (x==t || !flag[y][x-1] ||
			x==h || !flag[y][x+1] ||
			y==t || !flag[y-1][x] ||
			y==v || !flag[y+1][x]) flag[y][x]=2;

	    for (x=t; x<=h; x++)
	    for (y=t; y<=v; y++) if (flag[y][x]==2) flag[y][x]=0;
	}
}

#define DESC_LEN	512

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,bps,thres,bg,thick,z,lim,tot,h1,v1,d1,h2,v2,d2,num,y,x;
	char	**names,desc[DESC_LEN],***mask,**ptr,bit;
	Cell	**cell;

	if (argc!=5 && argc!=6)
	Error("usage : sliceOSP orgDir nameFile threshold thickness {newDir}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,NULL);

	if ((thres=atoi(argv[3]))<=0 || thres>=1<<bps) Error("bad threshold.");

	if (argv[4][0]=='-') {
	    bg=0; thick=atoi(&(argv[4][1]));
	}
	else {
	    bg=1; thick=atoi(argv[4]);
	}
	if (thick<0) Error("bad thickness.");

	cell=AllocateSliceMemory(Nx,Ny);

	flag=AllocateFlagMemory(Nx,Ny);

	if (bg) {
	    len=(int)(CUE_LENGTH_RATIO*(double)(Nx+Ny)+1.0);
	    if ((top=(Cue *)malloc(len*sizeof(*top)))==NULL)
		Error("cue memory allocation error.");

	    end=(&(top[len]));
	}

	if (argc==6) {
	    if ((mask=(char ***)malloc(((Nz+7)/8)*sizeof(*mask)))==NULL)
		Error("mask memory allocation error.");

	    for (z=0; z<(Nz+7)/8; z++) mask[z]=AllocateFlagMemory(Nx,Ny);

	    lim=0;
	}
	tot=0; h1=Nx-1; v1=Ny-1; d1=Nz-1; h2=v2=d2=0;

	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    PareObjectSkin(cell,thres,thick,bg);

	    num=0;
	    if (argc==6) {
		ptr=mask[z/8]; bit=1<<(z%8);
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++)
		    if (flag[y][x]) {
			++num; if (x<h1) h1=x; if (x>h2) h2=x;
			       if (y<v1) v1=y; if (y>v2) v2=y;

			if (cell[y][x]>lim) lim=cell[y][x];

			ptr[y][x]&=(~bit);
		    }
		    else
			ptr[y][x]|=bit;
	    }
	    else
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++)
		    if (flag[y][x]) {
			++num; if (x<h1) h1=x; if (x>h2) h2=x;
			       if (y<v1) v1=y; if (y>v2) v2=y;
		    }

	    if (num!=0) {
		tot+=num; if (z<d1) d1=z; if (z>d2) d2=z;
	    }
	}
	if (tot==0) Error("no occupied pixel exists.");

	printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\n",tot,h1,v1,d1,h2,v2,d2);

	if (argc==6) {
	    sprintf(desc,"%d\t%d\t%d\t%d",h1,v1,h2,v2);
/*
	    for (bps=1; 1<<bps<=lim; bps++) ;
*/
	    bps=(lim<2)?1:(lim<4)?2:(lim<16)?4:(lim<256)?8:(lim<65536)?16:32;

	    for (z=d1; z<=d2; z++) {
	        ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

		ptr=mask[z/8]; bit=1<<(z%8);
		for (y=v1; y<=v2; y++)
		for (x=h1; x<=h2; x++) if (ptr[y][x]&bit) cell[y][x]=0;

		StoreImageFile(argv[5],names+z,&cell,h1,v1,0,h2,v2,0,bps,
			       NULL,desc);
	    }
	}
	return 0;
}
