
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"

#define BITS	(sizeof(Cell)*8)

static Cell	***flag;

static int	Occupied(x,y,z)
int		x,y,z;
{
	return flag[z/BITS][y][x]&(Cell)(1<<(z%BITS));
}

static void	Erase(x,y,z)
int		x,y,z;
{
	flag[z/BITS][y][x]&=(~(Cell)(1<<(z%BITS)));
}

#define CUE_LENGTH_RATIO	2.8

typedef struct {
		int	x,y,z;
	} Cue;

static int	len,cnt;
static Cue	*top,*end,*head;

static void	Push(x,y,z)
int		x,y,z;
{
	if (++cnt>len) Error("cue overflow.");

	head->x=x; head->y=y; head->z=z; if (++head==end) head=top;
}

static int	x1,y1,z1,x2,y2,z2;

#define OCCUPIED(x,y,z)	(x>=x1 && x<=x2 && \
			 y>=y1 && y<=y2 && \
			 z>=z1 && z<=z2 && Occupied(x,y,z))

static void	Fill(x,y,z)
int		x,y,z;
{
	Cue	*tail;
	int	yP,yM,zP,zM,S,N,U,L,dx,S0,N0,U0,L0,x0,xC,SC,NC,UC,LC;

	head=tail=top;
	for (cnt=0; cnt>=0; cnt--) {
	    if (OCCUPIED(x,y,z)) {
		Erase(x,y,z); yP=y+1; S=OCCUPIED(x,yP,z);
			      yM=y-1; N=OCCUPIED(x,yM,z);
			      zP=z+1; U=OCCUPIED(x,y,zP);
			      zM=z-1; L=OCCUPIED(x,y,zM);
		for (dx=1; dx>=(-1); dx-=2) {
		    S0=S; N0=N; U0=U; L0=L;
		    for (xC=x, x0=x+dx; OCCUPIED(x0,y,z); xC=x0, x0+=dx) {
			Erase(x0,y,z); SC=S0; S0=OCCUPIED(x0,yP,z);
				       NC=N0; N0=OCCUPIED(x0,yM,z);
				       UC=U0; U0=OCCUPIED(x0,y,zP);
				       LC=L0; L0=OCCUPIED(x0,y,zM);
			if (SC && !S0) Push(xC,yP,z);
			if (NC && !N0) Push(xC,yM,z);
			if (UC && !U0) Push(xC,y,zP);
			if (LC && !L0) Push(xC,y,zM);
		    }
		    if (S0) Push(xC,yP,z);
		    if (N0) Push(xC,yM,z);
		    if (U0) Push(xC,y,zP);
		    if (L0) Push(xC,y,zM);
		}
	    }
	    x=tail->x; y=tail->y; z=tail->z; if (++tail==end) tail=top;
	}
}

#define FILL(x,y,z)	if (Occupied(x,y,z)) Fill(x,y,z)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,bps,thres,thick,bg,Z,z,x,y,t,x3,y3,z3,x4,y4,z4,num;
	char	**names,**descs;
	Cell	**cell,***mask,**ptr,bit,**Ptr,Bit;

	if (argc!=5 && argc!=6)
Error("usage : sliceOSP3 orgDir nameFile threshold thickness {newDir}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,&bps,&descs);

	if ((thres=atoi(argv[3]))<=0 || thres>=1<<bps) Error("bad threshold.");

	if (argv[4][0]=='-') {
	    bg=0; thick=atoi(&(argv[4][1]));
	}
	else {
	    bg=1; thick=atoi(argv[4]);
	}
	if (thick<0) Error("bad thickness.");

	cell=AllocateSliceMemory(Nx,Ny);

	Z=(Nz+BITS-1)/BITS;

	if ((mask=(Cell ***)malloc(Z*sizeof(*mask)))==NULL)
	    Error("mask memory allocation error.");

	for (z=0; z<Z; z++) mask[z]=AllocateSliceMemory(Nx,Ny);

	if (bg) {
	    if ((flag=(Cell ***)malloc(Z*sizeof(*flag)))==NULL)
		Error("flag memory allocation error.");

	    for (z=0; z<Z; z++) flag[z]=AllocateSliceMemory(Nx,Ny);

	    len=(int)(CUE_LENGTH_RATIO*(double)(Nx*Ny+Ny*Nz+Nz*Nx)+1.0);
	    if ((top=(Cue *)malloc(len*sizeof(*top)))==NULL)
		Error("cue memory allocation errror.");

	    end=(&(top[len]));
	}

	x1=Nx-1; y1=Ny-1; z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<Nz; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    ptr=mask[z/BITS]; bit=1<<(z%BITS);
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		if ((int)cell[y][x]>=thres) {
		    if (x<x1) x1=x; if (x>x2) x2=x;
		    if (y<y1) y1=y; if (y>y2) y2=y;
		    if (z<z1) z1=z; if (z>z2) z2=z; ptr[y][x]|=bit;
		}
		else
		    ptr[y][x]&=(~bit);
	    }
	}

	if (bg) {
	    for (z=z1/BITS; z<=z2/BITS; z++)
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++) flag[z][y][x]=(~mask[z][y][x]);

	    for (x=x1; x<=x2; x++)
	    for (y=y1; y<=y2; y++) { FILL(x,y,z1); FILL(x,y,z2); }
	    for (y=y1; y<=y2; y++)
	    for (z=z1; z<=z2; z++) { FILL(x1,y,z); FILL(x2,y,z); }
	    for (z=z1; z<=z2; z++)
	    for (x=x1; x<=x2; x++) { FILL(x,y1,z); FILL(x,y2,z); }

	    for (z=z1/BITS; z<=z2/BITS; z++)
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++) mask[z][y][x]|=flag[z][y][x];
	}

	for (t=0; t<thick; t++, x1++, y1++, z1++, x2--, y2--, z2--)
	    for (z=z1; z<=z2; z++) {
		if (z==z1) {
		    Ptr=mask[z/BITS]; Bit=1<<(z%BITS);
		    for (y=y1; y<=y2; y++)
		    for (x=x1; x<=x2; x++)
			if (Ptr[y][x]&Bit) cell[y][x]=2; else cell[y][x]=0;
		}
		else
		    for (y=y1; y<=y2; y++)
		    for (x=x1; x<=x2; x++) cell[y][x]<<=1;

		ptr=Ptr; bit=Bit;
		if (z!=z2) {
		    Ptr=mask[(z+1)/BITS]; Bit=1<<((z+1)%BITS);
		    for (y=y1; y<=y2; y++)
		    for (x=x1; x<=x2; x++) if (Ptr[y][x]&Bit) cell[y][x]|=1;
		}
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
		    if (cell[y][x]&2)
			if (x==x1 || !(cell[y  ][x-1]&2) ||
			    x==x2 || !(cell[y  ][x+1]&2) ||
			    y==y1 || !(cell[y-1][x  ]&2) ||
			    y==y2 || !(cell[y+1][x  ]&2) ||
			    z==z1 || !(cell[y  ][x  ]&4) ||
			    z==z2 || !(cell[y  ][x  ]&1)) ptr[y][x]&=(~bit);
	    }

	x3=x2; y3=y2; z3=z2; x4=x1; y4=y1; z4=z1; num=0;
	for (z=z1; z<=z2; z++) {
	    ptr=mask[z/BITS]; bit=1<<(z%BITS);
	    for (y=y1; y<=y2; y++)
	    for (x=x1; x<=x2; x++)
		if (ptr[y][x]&bit) {
		    if (x<x3) x3=x; if (x>x4) x4=x;
		    if (y<y3) y3=y; if (y>y4) y4=y;
		    if (z<z3) z3=z; if (z>z4) z4=z; ++num;
		}
	}
	if (num==0) Error("no occupied pixel exists.");

	printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\n",num,x3,y3,z3,x4,y4,z4);

	if (argc==6)
{
	int	lim=0;

	for (z=z3; z<=z4; z++) {
	    ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

	    ptr=mask[z/BITS]; bit=1<<(z%BITS);
	    for (y=y3; y<=y4; y++)
	    for (x=x3; x<=x4; x++)
		if (ptr[y][x]&bit && cell[y][x]>lim) lim=cell[y][x];
	}
/*
	for (bps=1; 1<<bps<=lim; bps++) ;
*/
	bps=(lim<2)?1:(lim<4)?2:(lim<16)?4:(lim<256)?8:(lim<65536)?16:32;

	for (z=0; z<Nz; z++) {
	    ptr=mask[z/BITS]; bit=1<<(z%BITS);
	    if (z<z3 || z>z4)
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) cell[y][x]=0;
	    else {
		ReadImageFile(argv[1],names+z,Nx,Ny,1,&cell);

		for (y=0; y<Ny; y++)
		    if (y<y3 || y>y4)
			for (x=0; x<Nx; x++) cell[y][x]=0;
		    else
			for (x=0; x<Nx; x++)
			    if (x<x3 || x>x4)
				cell[y][x]=0;
			    else
				if (!(ptr[y][x]&bit)) cell[y][x]=0;
	    }
	    StoreImageFile(argv[5],names+z,&cell,0,0,0,Nx-1,Ny-1,0,bps,
			   descs+z,NULL);
	}
}
	return 0;
}
