
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"

#ifndef TABLE_RATIO
#define TABLE_RATIO	(27.0/256.0)
#endif

#define TABLE_SIZE	((~(1<<(8*sizeof(int)-1)))/sizeof(int))

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]); return label;
}

typedef struct {
		Cell	cell;
		union {
			int	label;
			int	pixels;
		} u;
		double	x0,y0,z0;
		short	x1,y1,z1,x2,y2,z2,x3,y3,z3;
	} Peak;

#define MA(noe,top)	malloc((noe)*sizeof(*(top)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,**last,y,x,size,*table,X,Y,Z,labels,z,
		lB,lN,lW,label,*tmp,clusters,cluster;
	char	**names,**descs,*amae="array memory allocation error.";
	Cell	**prev,**curr,**succ,**mask,**work,cell,cB,cN,cW;
	double	table_size;
	Peak	*peak,*ptr;

	if (argc!=3 && argc!=4)
	    Error("usage : slicePS orgDir nameFile {newDir}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,&descs);

	prev=AllocateSliceMemory(Nx,Ny);
	curr=AllocateSliceMemory(Nx,Ny);
	succ=AllocateSliceMemory(Nx,Ny);

	if (argc==4) mask=AllocateSliceMemory(Nx,Ny);

	if ((last=(int **)MA(Ny+1,last))==NULL) Error(amae);

	++last;
	for (y=(-1); y<Ny; y++) {
	    if ((last[y]=(int *)MA(Nx+1,last[y]))==NULL) Error(amae);

	    ++last[y];
	    for (x=(-1); x<Nx; x++) last[y][x]=0;
	}

	table_size=TABLE_RATIO*(double)Nx*(double)Ny*(double)Nz+2.0;
	size=(table_size>(double)TABLE_SIZE)?TABLE_SIZE:(int)table_size;

	for (; size>1 && (table=(int *)MA(size,table))==NULL; size/=2)
	fprintf(stderr,"no allocatable memory for table of size %d.\n",size);

	if (size<=1) Error("table memory allocation error.");

	X=Nx-1; Y=Ny-1; Z=Nz-1;

	table[0]=table[1]=0; labels=2;
	for (z=0; z<Nz; z++) {
	    if (z==0)
		ReadImageFile(argv[1],names,Nx,Ny,1,&curr);
	    else {
		work=prev; prev=curr; curr=succ; succ=work;
	    }
	    if (z<Z) ReadImageFile(argv[1],names+z+1,Nx,Ny,1,&succ);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++)
		if ((cell=curr[y][x])==0)
		    last[y][x]=0;
		else {
		    cB=(z)?prev[y  ][x  ]:0; lB=last[y  ][x  ];
		    cN=(y)?curr[y-1][x  ]:0; lN=last[y-1][x  ];
		    cW=(x)?curr[y  ][x-1]:0; lW=last[y  ][x-1];

		    if (cB>cell || (z<Z && succ[y  ][x  ]>cell) ||
			cN>cell || (y<Y && curr[y+1][x  ]>cell) ||
			cW>cell || (x<X && curr[y  ][x+1]>cell)) {
			if (cB==cell && (lB=Search(table,lB))>1) table[lB]=(-1);
			if (cN==cell && (lN=Search(table,lN))>1) table[lN]=(-1);
			if (cW==cell && (lW=Search(table,lW))>1) table[lW]=(-1);

			last[y][x]=0;
		    }
		    else if ((cB==cell && lB==0) ||
			     (cN==cell && lN==0) ||
			     (cW==cell && lW==0)) {
			if ((lB=Search(table,lB))>1) table[lB]=(-1);
			if ((lN=Search(table,lN))>1) table[lN]=(-1);
			if ((lW=Search(table,lW))>1) table[lW]=(-1);

			last[y][x]=1;
		    }
		    else {
			if (lB) {
			    label=Search(table,lB);

			    if (lN && (lN=Search(table,lN))!=label)
				if (lN>1)
				    table[lN]=(-label);
				else {
				    table[label]=(-1); label=1;
				}

			    if (lW && (lW=Search(table,lW))!=label)
				if (lW>1)
				    table[lW]=(-label);
				else {
				    table[label]=(-1); label=1;
				}
			}
			else
			    if (lN) {
				label=Search(table,lN);

				if (lW && (lW=Search(table,lW))!=label)
				    if (lW>1)
					table[lW]=(-label);
				    else {
					table[label]=(-1); label=1;
				    }
			    }
			    else
				if (lW)
				    label=Search(table,lW);
				else {
				    if (labels==size) Error("too many labels.");

				    table[label=labels++]=(int)cell;
				}

			last[y][x]=label;
		    }
		}
	}

	if ((tmp=(int *)realloc(table,labels*sizeof(int)))!=NULL) table=tmp;

	clusters=2;
	for (label=2; label<labels; label++) if (table[label]>0) ++clusters;

	if ((peak=(Peak *)MA(clusters,peak))==NULL)
	    Error("peak memory allocation error.");

	ptr=peak;
	for (label=0; label<labels; label++)
	    if (table[label]>=0) {
		ptr->cell=(Cell)table[ptr->u.label=label]; ++ptr;
	    }
	    else
		table[label]=(-Search(table,label));

	for (ptr=peak, cluster=0; cluster<clusters; cluster++, ptr++) {
	    table[ptr->u.label]=cluster;
	    ptr->u.pixels=0; ptr->x0=0.0; ptr->x1=(short)X; ptr->x2=0;
			     ptr->y0=0.0; ptr->y1=(short)Y; ptr->y2=0;
			     ptr->z0=0.0; ptr->z1=(short)Z; ptr->z2=0;
	}
	for (label=2; label<labels; label++)
	    if (table[label]<0) table[label]=table[-table[label]];

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) last[y][x]=0;

	label=2;
	for (z=0; z<Nz; z++) {
	    if (z==0)
		ReadImageFile(argv[1],names,Nx,Ny,1,&curr);
	    else {
		work=prev; prev=curr; curr=succ; succ=work;
	    }
	    if (z<Z) ReadImageFile(argv[1],names+z+1,Nx,Ny,1,&succ);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		if ((cell=curr[y][x])==0)
		    cluster=0;
		else {
		    cB=(z)?prev[y  ][x  ]:0;
		    cN=(y)?curr[y-1][x  ]:0;
		    cW=(x)?curr[y  ][x-1]:0;
		
		    if (cB>cell || (z<Z && succ[y  ][x  ]>cell) ||
			cN>cell || (y<Y && curr[y+1][x  ]>cell) ||
			cW>cell || (x<X && curr[y  ][x+1]>cell))
			cluster=0;
		    else if ((cB==cell && last[y  ][x  ]==0) ||
			     (cN==cell && last[y-1][x  ]==0) ||
			     (cW==cell && last[y  ][x-1]==0))
			cluster=1;
		     else
			if (last[y][x])
			    cluster=last[y][x];
			else if (last[y-1][x])
			    cluster=last[y-1][x];
			else if (last[y][x-1])
			    cluster=last[y][x-1];
			else
			    cluster=table[label++];
		}
		if ((last[y][x]=cluster)>1) {
		    ptr=peak+cluster;
		    ++(ptr->u.pixels);
		    ptr->x0+=(double)x; if (x<(int)ptr->x1) ptr->x1=(short)x;
					if (x>(int)ptr->x2) ptr->x2=(short)x;
		    ptr->y0+=(double)y; if (y<(int)ptr->y1) ptr->y1=(short)y;
					if (y>(int)ptr->y2) ptr->y2=(short)y;
		    ptr->z0+=(double)z; if (z<(int)ptr->z1) ptr->z1=(short)z;
					if (z>(int)ptr->z2) ptr->z2=(short)z;
		}
		if (argc==4) mask[y][x]=(cluster>1)?cell:0;
	    }
	    if (argc==4) StoreImageFile(argv[3],names+z,&mask,0,0,0,X,Y,0,
					GetImageBPS(&mask,0,0,0,X,Y,0),
					descs+z,NULL);
	}

	printf("%d\n",clusters-2);
	for (ptr=peak+(cluster=2); cluster<clusters; cluster++, ptr++)
	    printf("%hu\t%d\t%lf\t%lf\t%lf\t%hd\t%hd\t%hd\t%hd\t%hd\t%hd\n",
		   ptr->cell,ptr->u.pixels,ptr->x0/(double)ptr->u.pixels,
					   ptr->y0/(double)ptr->u.pixels,
					   ptr->z0/(double)ptr->u.pixels,
					   ptr->x1,ptr->y1,ptr->z1,
					   ptr->x2,ptr->y2,ptr->z2);

	return 0;
}
