
#include <stdio.h>
#include <stdlib.h>
#include "slice.h"
#include "FFT.h"

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,y,x,arg,Lz,Lz0,z,No=0;
	char	**names;
	double	dz,*P,f,d;
	Cell	***cell,**mask=NULL;
	Complex	*E,*C;

	if (argc<4 || (argc>6 && argc%2==1))
	Error("usage : slicePSD directory nameFile {mask}|{(x y) ...} dz");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);

	if ((dz=atof(argv[argc-1]))<=0.0) Error("bad slice interval.");

	if (argc==4)
	    No=Nx*Ny;
	else {
	    mask=AllocateSliceMemory(Nx,Ny);
	    if (argc==5) {
		ReadSliceFile(argv[3],Nx,Ny,mask);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) if (mask[y][x]) ++No;

		if (No==0) Error("no sampling point.");
	    }
	    else {
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) mask[y][x]=0;

		for (arg=3; arg<argc-1; arg+=2) {
		    if ((x=atoi(argv[arg  ]))<0 || x>=Nx ||
			(y=atoi(argv[arg+1]))<0 || y>=Ny || mask[y][x])
			Error("bad position of sampling point.");

		    mask[y][x]=1; ++No;
		}
	    }
	}

	if ((Lz=SetUpFFT(Nz,&E))==0 ||
	    (C=(Complex *)malloc(Lz*sizeof(*C)))==NULL)
	    Error("FFT memory allocation error.");

	Lz0=Lz/2+1;
	if ((P=(double *)malloc(Lz0*sizeof(*P)))==NULL)
	    Error("PSD memory allocation error.");

	for (z=0; z<Lz0; z++) P[z]=0.0;

	cell=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	printf("%lf\t%d\t%d\t%d\n",dz,Nz,Lz,No);

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (mask==NULL || mask[y][x]) {
		for (z=0; z<Nz; z++) {
		    C[z].r=(double)cell[z][y][x]; C[z].i=0.0;
		}
		for (z=Nz; z<Lz; z++) C[z].r=C[z].i=0.0;

		FFT(Lz,E,C,-1);

		for (z=0; z<Lz0; z++) P[z]+=(C[z].r*C[z].r+C[z].i*C[z].i);
	    }

	f=1.0/(dz*(double)Lz); d=dz/((double)Lz*(double)No);
	for (z=0; z<Lz0; z++) printf("%lf\t%le\n",(double)z*f,P[z]*d);

	return 0;
}
