
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "slice.h"

static int	CheckAxis(str)
char		*str;
{
	int	code;

	switch(str[0]) {
	    case '+' : code=0; break;
	    case '-' : code=1; break; default : return -1;
	}
	switch(str[1]) {
	    case 'x' : case 'X' : code+=0; break;
	    case 'y' : case 'Y' : code+=2; break;
	    case 'z' : case 'Z' : code+=4; break; default : return -1;
	}
	return code;
}

typedef struct {
		int	Nh,Nv,Nd,xh,yh,zh,xv,yv,zv,xd,yd,zd,lo,la,ro,x0,y0,z0;
	} TP;

#define NH	tp.Nh
#define NV	tp.Nv
#define ND	tp.Nd
#define XH	tp.xh
#define YH	tp.yh
#define ZH	tp.zh
#define XV	tp.xv
#define YV	tp.yv
#define ZV	tp.zv
#define XD	tp.xd
#define YD	tp.yd
#define ZD	tp.zd
#define LO	tp.lo
#define LA	tp.la
#define RO	tp.ro
#define X0	tp.x0
#define Y0	tp.y0
#define Z0	tp.z0

static TP	SetUpTP(Nx,Ny,Nz,hCode,vCode)
int		Nx,Ny,Nz,hCode,vCode;
{
	TP	tp;

	XH=YH=ZH=XV=YV=ZV=XD=YD=ZD=0;

	switch(hCode) {
	    case  0 : NH=Nx; XH=  1; break;
	    case  1 : NH=Nx; XH= -1; break;
	    case  2 : NH=Ny; YH=  1; break;
	    case  3 : NH=Ny; YH= -1; break;
	    case  4 : NH=Nz; ZH=  1; break;
	    case  5 : NH=Nz; ZH= -1; break;
	}
	switch(vCode) {
	    case  0 : NV=Nx; XV=  1; break;
	    case  1 : NV=Nx; XV= -1; break;
	    case  2 : NV=Ny; YV=  1; break;
	    case  3 : NV=Ny; YV= -1; break;
	    case  4 : NV=Nz; ZV=  1; break;
	    case  5 : NV=Nz; ZV= -1; break;
	}
	switch(hCode*10+vCode) {
	    case 02 : ND=Nz; ZD=  1; LO=  90; LA=  90; RO=   0; break;
	    case 03 : ND=Nz; ZD= -1; LO=  90; LA= -90; RO=   0; break;
	    case 04 : ND=Ny; YD= -1; LO= -90; LA=   0; RO= 180; break;
	    case 05 : ND=Ny; YD=  1; LO=  90; LA=   0; RO=   0; break;
	    case 12 : ND=Nz; ZD= -1; LO= -90; LA= -90; RO=   0; break;
	    case 13 : ND=Nz; ZD=  1; LO= -90; LA=  90; RO=   0; break;
	    case 14 : ND=Ny; YD=  1; LO=  90; LA=   0; RO= 180; break;
	    case 15 : ND=Ny; YD= -1; LO= -90; LA=   0; RO=   0; break;
	    case 20 : ND=Nz; ZD= -1; LO= 180; LA= -90; RO=   0; break;
	    case 21 : ND=Nz; ZD=  1; LO= 180; LA=  90; RO=   0; break;
	    case 24 : ND=Nx; XD=  1; LO=   0; LA=   0; RO= 180; break;
	    case 25 : ND=Nx; XD= -1; LO= 180; LA=   0; RO=   0; break;
	    case 30 : ND=Nz; ZD=  1; LO=   0; LA=  90; RO=   0; break;
	    case 31 : ND=Nz; ZD= -1; LO=   0; LA= -90; RO=   0; break;
	    case 34 : ND=Nx; XD= -1; LO= 180; LA=   0; RO= 180; break;
	    case 35 : ND=Nx; XD=  1; LO=   0; LA=   0; RO=   0; break;
	    case 40 : ND=Ny; YD=  1; LO=  90; LA=   0; RO= -90; break;
	    case 41 : ND=Ny; YD= -1; LO= -90; LA=   0; RO= -90; break;
	    case 42 : ND=Nx; XD= -1; LO= 180; LA=   0; RO= -90; break;
	    case 43 : ND=Nx; XD=  1; LO=   0; LA=   0; RO= -90; break;
	    case 50 : ND=Ny; YD= -1; LO= -90; LA=   0; RO=  90; break;
	    case 51 : ND=Ny; YD=  1; LO=  90; LA=   0; RO=  90; break;
	    case 52 : ND=Nx; XD=  1; LO=   0; LA=   0; RO=  90; break;
	    case 53 : ND=Nx; XD= -1; LO= 180; LA=   0; RO=  90; break;
	}
	X0=(XH+XV+XD==1)?0:Nx-1;
	Y0=(YH+YV+YD==1)?0:Ny-1;
	Z0=(ZH+ZV+ZD==1)?0:Nz-1;

	return tp;
}

#define DESC_LEN	256

static void	GenerateNewSlice(Nx,Ny,Nz,org,tp,new,d,path)
int		Nx,Ny,Nz,d;
Cell		***org,**new;
TP		tp;
char		*path;
{
	int	h,v,x=X0+XD*d,y=Y0+YD*d,z=Z0+ZD*d;
	char	desc[DESC_LEN];

	for (v=0; v<NV; v++)
	for (h=0; h<NH; h++)
	    new[v][h]=org[ZH*h+ZV*v+z][YH*h+YV*v+y][XH*h+XV*v+x];

	sprintf(desc,"%d\t%d\t%d\t1\t1\t1\n%d\t%d\t%d\t%d\t%d\t%d\n",
		     Nx,Ny,Nz,x,y,z,LO,LA,RO);
	StoreSliceFile(path,new,0,0,NH-1,NV-1,
		       GetImageBPS(&new,0,0,0,NH-1,NV-1,0),desc);
}

void	FreeSliceMemory(ySize,cell)
int	ySize;
Cell	**cell;
{
	int	y;

	for (y=0; y<ySize; y++) free(cell[y]); free(cell);
}

#define LINE_LEN	1024
#define STR_LEN		512
#define PATH_LEN	512

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,hCode,vCode,d;
	char	**names,line[LINE_LEN],hStr[STR_LEN],vStr[STR_LEN],
		dirOrNo[PATH_LEN],path[PATH_LEN],format[PATH_LEN],
		*bdoa="bad direction of axis.",
		*tdsp="too deep slice position.";
	Cell	***org,**new;
	TP	tp;

	if (argc!=3 && argc!=6 && argc!=7)
Error("usage : sliceRAR orgDir nameFile {hAxis vAxis newDir|(sliceNo path)}");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,NULL);
	org=AllocateImageMemory(Nx,Ny,Nz);
	ReadImageFile(argv[1],names,Nx,Ny,Nz,org);

switch(argc) {
    case 3 :
	while (fgets(line,LINE_LEN,stdin)!=NULL)
switch(sscanf(line,"%s %s %s %s",hStr,vStr,dirOrNo,path)) {
    case 3 :
	if ((hCode=CheckAxis(hStr))<0 ||
	    (vCode=CheckAxis(vStr))<0 || hCode/2==vCode/2) {
	    fprintf(stderr,"%s : %s\n",line,bdoa); continue;
	}
	tp=SetUpTP(Nx,Ny,Nz,hCode,vCode);

	sprintf(path,"%d",ND-1);
	sprintf(format,"%s/%%0%dd.tif",dirOrNo,strlen(path));

	new=AllocateSliceMemory(NH,NV);
	for (d=0; d<ND; d++) {
	    sprintf(path,format,d);
	    GenerateNewSlice(Nx,Ny,Nz,org,tp,new,d,path);
	}
	FreeSliceMemory(NV,new);
	break;
    case 4 :
	if ((hCode=CheckAxis(hStr))<0 ||
	    (vCode=CheckAxis(vStr))<0 || hCode/2==vCode/2) {
	    fprintf(stderr,"%s : %s\n",line,bdoa); continue;
	}
	tp=SetUpTP(Nx,Ny,Nz,hCode,vCode);

	if ((d=atoi(dirOrNo))<0 || d>=ND) {
	    fprintf(stderr,"%s : %s\n",line,tdsp); continue;
	}
	new=AllocateSliceMemory(NH,NV);
	GenerateNewSlice(Nx,Ny,Nz,org,tp,new,d,path);
	FreeSliceMemory(NV,new);
	break;
    default :
	fprintf(stderr,"%s : bad number of paramters.\n",line); continue;
}
	break;
    case 6 :
	if ((hCode=CheckAxis(argv[3]))<0 ||
	    (vCode=CheckAxis(argv[4]))<0 || hCode/2==vCode/2) Error(bdoa);

	tp=SetUpTP(Nx,Ny,Nz,hCode,vCode);

	sprintf(path,"%d",ND-1);
	sprintf(format,"%s/%%0%dd.tif",argv[5],strlen(path));

	new=AllocateSliceMemory(NH,NV);
	for (d=0; d<ND; d++) {
	    sprintf(path,format,d);
	    GenerateNewSlice(Nx,Ny,Nz,org,tp,new,d,path);
	}
	FreeSliceMemory(NV,new);
	break;
    case 7 :
	if ((hCode=CheckAxis(argv[3]))<0 ||
	    (vCode=CheckAxis(argv[4]))<0 || hCode/2==vCode/2) Error(bdoa);

	tp=SetUpTP(Nx,Ny,Nz,hCode,vCode);
	if ((d=atoi(argv[5]))<0 || d>=ND) Error(tdsp);

	new=AllocateSliceMemory(NH,NV);
	GenerateNewSlice(Nx,Ny,Nz,org,tp,new,d,argv[6]);
	FreeSliceMemory(NV,new);
	break;
}
	return 0;
}
