
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "slice.h"

#define CUE_RATIO	2.8

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

static char	***AllocateMaskMemory(wid,hig,num)
int		wid,hig,num;
{
	char	***mask;
	int	z,y,x;
	char	*msg="mask memory allocation error.";

	if ((mask=(char ***)MA(num,mask))==NULL) Error(msg);

	for (z=0; z<num; z++) {
	    if ((mask[z]=(char **)MA(hig,*mask))==NULL) Error(msg);

	    for (y=0; y<hig; y++) {
		if ((mask[z][y]=(char *)MA(wid,**mask))==NULL) Error(msg);

		for (x=0; x<wid; x++) mask[z][y][x]=0;
	    }
	}
	return mask;
}

static int	ScanImage(Nx,Ny,Nz,cell,lo,up,xC,yC,zC,mask)
int		Nx,Ny,Nz,lo,up,*xC,*yC,*zC;
Cell		***cell;
char		***mask;
{
	int	x,y,z,dx,dy,dz,xB=(*xC),yB=(*yC),zB=(*zC),d0=0,cnt=0;

	for (x=0; x<Nx; x++) { dx=x-xB; dx=   dx*dx;
	for (y=0; y<Ny; y++) { dy=y-yB; dy=dx+dy*dy;
	for (z=0; z<Nz; z++) { dz=z-zB; dz=dy+dz*dz;
	    if (cell[z][y][x]>=lo && cell[z][y][x]<=up) {
		mask[z][y][x]=1;
		if (++cnt==1 || dz<d0) { *xC=x; *yC=y; *zC=z; d0=dz; }
	    }
	    else
		mask[z][y][x]=0;
	}}}
	return cnt;
}

typedef unsigned short	Word;

typedef struct {
		Word	x,y,z;
	} Cue;

static int	Len,Cnt;
static Cue	*Top,*End;

static void	PutCue(head,x,y,z)
Cue		**head;
int		x,y,z;
{
	if (++Cnt>Len) Error("cue overflow.");

	(*head)->x=x;
	(*head)->y=y;
	(*head)->z=z; if (++(*head)==End) *head=Top;
}

#define OCCUPIED(x,y,z)	(x>=0 && x<Nx && \
			 y>=0 && y<Ny && \
			 z>=0 && z<Nz && mask[z][y][x]==1)
#define RESET(x,y,z)	if (++cnt==1) { \
			    *h1=(*h2)=x; \
			    *v1=(*v2)=y; \
			    *d1=(*d2)=z; \
			} \
			else { \
			    if (x<*h1) *h1=x; if (x>*h2) *h2=x; \
			    if (y<*v1) *v1=y; if (y>*v2) *v2=y; \
			    if (z<*d1) *d1=z; if (z>*d2) *d2=z; \
			} \
			mask[z][y][x]=2

static int	LabelCluster(Nx,Ny,Nz,mask,x,y,z,h1,v1,d1,h2,v2,d2)
int		Nx,Ny,Nz,x,y,z,*h1,*v1,*d1,*h2,*v2,*d2;
char		***mask;
{
	int	yP,yM,zP,zM,S,N,U,L,dx,S0,N0,U0,L0,x0,x1,S1,N1,U1,L1,cnt=0;
	Cue	*head=Top,*tail=Top;

	for (Cnt=0; Cnt>=0; Cnt--) {
	    if (OCCUPIED(x,y,z)) {
		RESET(x,y,z); yP=y+1; S=OCCUPIED(x,yP,z);
			      yM=y-1; N=OCCUPIED(x,yM,z);
			      zP=z+1; U=OCCUPIED(x,y,zP);
			      zM=z-1; L=OCCUPIED(x,y,zM);
		for (dx=1; dx>=(-1); dx-=2) {
		    S0=S; N0=N; U0=U; L0=L;
		    for (x1=x, x0=x+dx; OCCUPIED(x0,y,z); x1=x0, x0+=dx) {
			RESET(x0,y,z); S1=S0; S0=OCCUPIED(x0,yP,z);
				       N1=N0; N0=OCCUPIED(x0,yM,z);
				       U1=U0; U0=OCCUPIED(x0,y,zP);
				       L1=L0; L0=OCCUPIED(x0,y,zM);
			if (S1 && !S0) PutCue(&head,x1,yP,z);
			if (N1 && !N0) PutCue(&head,x1,yM,z);
			if (U1 && !U0) PutCue(&head,x1,y,zP);
			if (L1 && !L0) PutCue(&head,x1,y,zM);
		    }
		    if (S0) PutCue(&head,x1,yP,z);
		    if (N0) PutCue(&head,x1,yM,z);
		    if (U0) PutCue(&head,x1,y,zP);
		    if (L0) PutCue(&head,x1,y,zM);
		}
	    }
	    x=tail->x; y=tail->y; z=tail->z; if (++tail==End) tail=Top;
	}
	return cnt;
}

#define LINE_LEN	1024
#define DIR_LEN		512

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nz,Nx,Ny,lo,up,pn,cnt,x,y,z,h1,v1,d1,h2,v2,d2;
	char	**names,**descs,***mask,line[LINE_LEN],dir[DIR_LEN];
	Cell	***cell,**slice=NULL;

	if (argc!=3) Error("usage : sliceSCL orgDir nameFile");

	ScanNameFile(argv[1],argv[2],&Nz,&names);
	CheckImageFile(argv[1],Nz,names,&Nx,&Ny,NULL,&descs);

	cell=AllocateImageMemory(Nx,Ny,Nz);
	mask=AllocateMaskMemory(Nx,Ny,Nz);

	Len=(int)(CUE_RATIO*pow((double)Nx*(double)Ny*(double)Nz,2.0/3.0)+1.0);
	if ((Top=(Cue *)MA(Len,Top))==NULL)
	    Error("cue memory allocation error.");

	End=(&(Top[Len]));

	ReadImageFile(argv[1],names,Nx,Ny,Nz,cell);

	while (fgets(line,LINE_LEN,stdin)!=NULL)
	if ((pn=sscanf(line,"%d %d %d %d %d %s",&lo,&up,&x,&y,&z,dir))>=5) {
	    if ((cnt=ScanImage(Nx,Ny,Nz,cell,lo,up,&x,&y,&z,mask))==0) {
		fprintf(stderr,
			"%d %d %d %d %d : no occupied pixel exists.\n",
			lo,up,x,y,z);
		continue;
	    }

	    cnt=LabelCluster(Nx,Ny,Nz,mask,x,y,z,&h1,&v1,&d1,&h2,&v2,&d2);

	    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
		   lo,up,x,y,z,cnt,h1,v1,d1,h2,v2,d2);

	    if (pn==6) {
		if (slice==NULL) slice=AllocateSliceMemory(Nx,Ny);

		for (z=d1; z<=d2; z++) {
		    for (y=0; y<Ny; y++)
		    for (x=0; x<Nx; x++) slice[y][x]=(mask[z][y][x]==2)?1:0;

		    StoreImageFile(dir,names+z,
				   &slice,0,0,0,Nx-1,Ny-1,0,1,descs+z,NULL);
		}
	    }
	}
	return 0;
}
