
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiffio.h"
#include "slice.h"

int	GetImageBPS(cell,x1,y1,z1,x2,y2,z2)
Cell	***cell;
int	x1,y1,z1,x2,y2,z2;
{
	int	z,y,x,depth=0;

	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) if (cell[z][y][x]>depth) depth=cell[z][y][x];

return (depth<2)?1:(depth<4)?2:(depth<16)?4:(depth<256)?8:(depth<65536)?16:32;
}

#define PI	((bps==1)?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK)
#define SPP	1
#define PC	PLANARCONFIG_CONTIG
#define TSF	TIFFSetField

void	StoreSliceFile(path,cell,x1,y1,x2,y2,bps,desc)
char	*path,*desc;
Cell	**cell;
int	x1,y1,x2,y2,bps;
{
	TIFF	*tiff;
	int	tss,y,x,wid=x2-x1+1,hig=y2-y1+1;
	uint8	*tsl,*ptr,bit;
	Cell	*data,mask,msb=(Cell)(1<<(bps-1));
	uint16	*ptr16;
	uint32	*ptr32;

	if ((tiff=TIFFOpen(path,"w"))==NULL ||
	    !TSF(tiff,TIFFTAG_IMAGEWIDTH,     (uint32)wid) ||
	    !TSF(tiff,TIFFTAG_IMAGELENGTH,    (uint32)hig) ||
	    !TSF(tiff,TIFFTAG_PHOTOMETRIC,    (uint16)PI ) ||
	    !TSF(tiff,TIFFTAG_BITSPERSAMPLE,  (uint16)bps) ||
	    !TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)SPP) ||
	    !TSF(tiff,TIFFTAG_PLANARCONFIG,   (uint16)PC ) ||
	    (tsl=(uint8 *)malloc(tss=TIFFScanlineSize(tiff)))==NULL) {
	    fprintf(stderr,"%s : ",path); Error("image file not stored.");
	}
	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(uint32)((tss>8192)?1:8192/tss));
#ifdef	LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	if (desc!=NULL) TSF(tiff,TIFFTAG_IMAGEDESCRIPTION,desc);

	for (y=y1; y<=y2; y++) {
	    data=(&(cell[y][x1]));
	    if (bps==8)
		for (ptr=tsl, x=x1; x<=x2; x++) *(ptr++)=(uint8)*(data++);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=x1; x<=x2; x++)
		    *(ptr16++)=(uint16)*(data++);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=x1; x<=x2; x++)
		    *(ptr32++)=(uint32)*(data++);
	    else
		for (ptr=tsl, bit=0x80, x=x1; x<=x2; x++, data++)
		    for (mask=msb; mask>0; mask>>=1) {
			if (*data&mask) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }

	    if (TIFFWriteScanline(tiff,tsl,(uint32)(y-y1),(uint16)0)<0) {
		fprintf(stderr,"%s : ",path); Error("write data error.");
	    }
	}
	free(tsl); TIFFClose(tiff);
}

#define PATH_LEN	512

#define STORE(desc)	StoreSliceFile(path,cell[z],x1,y1,x2,y2,bps,desc)

void	StoreImageFile(dir,names,cell,x1,y1,z1,x2,y2,z2,bps,descEach,descAll)
char	*dir,**names,**descEach,*descAll;
Cell	***cell;
int	x1,y1,z1,x2,y2,z2,bps;
{
	int	z,lenE,lenA;
	char	path[PATH_LEN],*desc;

	for (z=z1; z<=z2; z++) {
	    sprintf(path,"%s/%s",dir,names[z]);
	    if ((descEach!=NULL && descEach[z]!=NULL) || descAll!=NULL)
		if (descEach==NULL || descEach[z]==NULL)
		    STORE(descAll);
		else if (descAll==NULL)
		    STORE(descEach[z]);
		else {
		    lenE=strlen(descEach[z]);
		    lenA=strlen(descAll);
		    if ((desc=(char *)malloc(lenE+lenA+3))==NULL) {
			fprintf(stderr,"%s : description not stored.\n",path);

		        STORE(NULL);
		    }
		    else {
			strcpy(desc,descEach[z]);
			if (descEach[z][lenE-1]!='\n') strcat(desc,"\n");
			strcat(desc,descAll);
			if (descAll[lenA-1]    !='\n') strcat(desc,"\n");

		        STORE(desc); free(desc);
		    }
		}
	    else
		STORE(NULL);
	}
}
