
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

typedef struct {
		double	sum1,sum2;
	} Sum;

typedef struct {
		Sum	r,g,b;
	} RGB;

#define STRLEN	256
#define PATH	argv[arg]
#define PIXELS	(xSize*ySize)
#define MA(ptr)	malloc(PIXELS*sizeof(*ptr))
#define R	image->r
#define G	image->g
#define B	image->b
#define PATH2	argv[argc-2]
#define PATH1	argv[argc-1]
#define INT(z)	floor((z)+0.5)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	arg,code,xSize,ySize,x,y,v;
	TIFF	*tiff;
	char	msg[STRLEN],*wmae="work memory allocation error.",
			    *itnm="image type not match.",
			    *tfns="TIFF file not stored.";
	Image	*image;
	CM	*cm,*largest=NULL;
	Sum	*iP,iA,*i=NULL;
	RGB	*cP,cA,*c=NULL;
	double	z,sum0=(double)(argc-3);
#ifdef	N_SD
	double	sI,sR,sG,sB;
#endif
	if (argc<4) Error("usage : tiff2msd TIFFs meanTIFF SDTIFF");

	for (arg=1; arg<=argc-3; arg++) {
	    if ((tiff=TIFFOpen(PATH,"r"))==NULL) {
		sprintf(msg,"%s : TIFF file not found.",PATH); Error(msg);
	    }
	    if ((code=ReadTIFF(tiff,&image,&cm))>0) {
		sprintf(msg,"%s : TIFF %s",PATH,ReadTIFFError(code));
		Error(msg);
	    }
	    TIFFClose(tiff);

	    if (arg==1) {
		xSize=image->xSize; ySize=image->ySize;

		if (cm!=NULL) {
		    if ((i=(Sum *)MA(i))==NULL) Error(wmae);

		    for (iP=i, x=0; x<xSize; x++)
		    for (y=0; y<ySize; y++, iP++) {
			iP->sum1=z=(double)GETPIXEL(image,x,y); iP->sum2=(z*z);
		    }
		    largest=cm;
		}
		else {
		    if ((c=(RGB *)MA(c))==NULL) Error(wmae);

		    for (cP=c, x=0; x<xSize; x++)
		    for (y=0; y<ySize; y++, cP++) {
			cP->r.sum1=z=(double)R[x][y]; cP->r.sum2=(z*z);
			cP->g.sum1=z=(double)G[x][y]; cP->g.sum2=(z*z);
			cP->b.sum1=z=(double)B[x][y]; cP->b.sum2=(z*z);
		    }
		}
	    }
	    else {
		if (image->xSize!=xSize || image->ySize!=ySize) {
		    sprintf(msg,"%s : image size not match.",PATH); Error(msg);
		}
		if (cm!=NULL) {
		    if (i==NULL) {
			sprintf(msg,"%s : %s",PATH,itnm); Error(msg);
		    }
		    for (iP=i, x=0; x<xSize; x++)
		    for (y=0; y<ySize; y++, iP++) {
			iP->sum1+=(z=(double)GETPIXEL(image,x,y));
			iP->sum2+=(z*z);
		    }
		    if (cm->pixels>largest->pixels) {
			FreeCM(largest); largest=cm;
		    }
		    else
			FreeCM(cm);
		}
		else {
		    if (c==NULL) {
			sprintf(msg,"%s : %s",PATH,itnm); Error(msg);
		    }
		    for (cP=c, x=0; x<xSize; x++)
		    for (y=0; y<ySize; y++, cP++) {
			cP->r.sum1+=(z=(double)R[x][y]); cP->r.sum2+=(z*z);
			cP->g.sum1+=(z=(double)G[x][y]); cP->g.sum2+=(z*z);
			cP->b.sum1+=(z=(double)B[x][y]); cP->b.sum2+=(z*z);
		    }
		}
	    }
	    if (arg<argc-3) FreeImage(image);
	}

	printf("%d\t%d\t%d\t%d\n",xSize,ySize,argc-3,(i!=NULL)?1:3);

	if (i!=NULL) {
	    iA.sum1=iA.sum2=0.0;
#ifdef	N_SD
	    sI=0.0;
#endif
	    for (iP=i, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, iP++) {
		iA.sum1+=(z=(iP->sum1/=sum0));
		iP->sum2=sqrt(z=iP->sum2/sum0-z*z); iA.sum2+=z;
#ifdef	N_SD
		if (iP->sum2>sI) sI=iP->sum2;
#endif
	    }
	    iA.sum1/=(double)PIXELS; iA.sum2=sqrt(iA.sum2/(double)PIXELS);
#ifdef	N_SD
	    sI=(sI<1.0)?(double)(largest->pixels-1):
			(double)(largest->pixels-1)/sI;

	    printf("%lg\t%lg\t%lg\n",iA.sum1,iA.sum2,sI);
#else
	    printf("%lg\t%lg\n",iA.sum1,iA.sum2);
#endif
	}
	else {
	    cA.r.sum1=cA.r.sum2=
	    cA.g.sum1=cA.g.sum2=
	    cA.b.sum1=cA.b.sum2=0.0;
#ifdef	N_SD
	    sR=sG=sB=0.0;
#endif
	    for (cP=c, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, cP++) {
		cA.r.sum1+=(z=(cP->r.sum1/=sum0));
		cP->r.sum2=sqrt(z=cP->r.sum2/sum0-z*z); cA.r.sum2+=z;
		cA.g.sum1+=(z=(cP->g.sum1/=sum0));
		cP->g.sum2=sqrt(z=cP->g.sum2/sum0-z*z); cA.g.sum2+=z;
		cA.b.sum1+=(z=(cP->b.sum1/=sum0));
		cP->b.sum2=sqrt(z=cP->b.sum2/sum0-z*z); cA.b.sum2+=z;
#ifdef	N_SD
		if (cP->r.sum2>sR) sR=cP->r.sum2;
		if (cP->g.sum2>sG) sG=cP->g.sum2;
		if (cP->b.sum2>sB) sB=cP->b.sum2;
#endif
	    }
	    cA.r.sum1/=(double)PIXELS; cA.r.sum2=sqrt(cA.r.sum2/(double)PIXELS);
	    cA.g.sum1/=(double)PIXELS; cA.g.sum2=sqrt(cA.g.sum2/(double)PIXELS);
	    cA.b.sum1/=(double)PIXELS; cA.b.sum2=sqrt(cA.b.sum2/(double)PIXELS);
#ifdef	N_SD
	    sR=(sR<1.0)?255.0:255.0/sR;
	    sG=(sG<1.0)?255.0:255.0/sG;
	    sB=(sB<1.0)?255.0:255.0/sB;

	    printf("%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n",
		   cA.r.sum1,cA.r.sum2,sR,
		   cA.g.sum1,cA.g.sum2,sG,
		   cA.b.sum1,cA.b.sum2,sB);
#else
	    printf("%lg\t%lg\n%lg\t%lg\n%lg\t%lg\n",cA.r.sum1,cA.r.sum2,
						    cA.g.sum1,cA.g.sum2,
						    cA.b.sum1,cA.b.sum2);
#endif
	}

	if ((tiff=TIFFOpen(PATH2,"w"))==NULL) {
	    sprintf(msg,"%s : %s",PATH2,tfns); Error(msg);
	}
	if (i!=NULL) {
	    for (v=0; v<largest->pixels; v++)
		largest->r[v]=
		largest->g[v]=
		largest->b[v]=(word)((0xffff*v)/(largest->pixels-1));

	    for (iP=i, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, iP++) {
		v=(int)INT(iP->sum1); PUTPIXEL(image,x,y,v);
	    }
	    sprintf(msg,"%lg",iA.sum1);
	}
	else {
	    for (cP=c, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, cP++) {
		R[x][y]=(byte)INT(cP->r.sum1);
		G[x][y]=(byte)INT(cP->g.sum1);
		B[x][y]=(byte)INT(cP->b.sum1);
	    }
	    sprintf(msg,"%lg\t%lg\t%lg",cA.r.sum1,cA.g.sum1,cA.b.sum1);
	}
	TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,msg);

	if ((code=WriteTIFF(tiff,image,largest))>0) {
	    sprintf(msg,"%s : TIFF %s",PATH2,WriteTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	if ((tiff=TIFFOpen(PATH1,"w"))==NULL) {
	    sprintf(msg,"%s : %s",PATH1,tfns); Error(msg);
	}
	if (i!=NULL) {
	    for (iP=i, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, iP++) {
#ifdef	N_SD
		v=(int)INT(iP->sum2*sI);
#else
		v=(int)INT(iP->sum2);
#endif
		PUTPIXEL(image,x,y,v);
	    }
#ifdef	N_SD
	    sprintf(msg,"%lg\t%lg",iA.sum2,sI);
#else
	    sprintf(msg,"%lg",iA.sum2);
#endif
	}
	else {
	    for (cP=c, x=0; x<xSize; x++)
	    for (y=0; y<ySize; y++, cP++) {
#ifdef	N_SD
		R[x][y]=(byte)INT(cP->r.sum2*sR);
		G[x][y]=(byte)INT(cP->g.sum2*sG);
		B[x][y]=(byte)INT(cP->b.sum2*sB);
#else
		R[x][y]=(byte)INT(cP->r.sum2);
		G[x][y]=(byte)INT(cP->g.sum2);
		B[x][y]=(byte)INT(cP->b.sum2);
#endif
	    }
#ifdef	N_SD
	    sprintf(msg,"%lg\t%lg\n%lg\t%lg\n%lg\t%lg",cA.r.sum2,sR,
						       cA.g.sum2,sG,
						       cA.b.sum2,sB);
#else
	    sprintf(msg,"%lg\t%lg\t%lg",cA.r.sum2,cA.g.sum2,cA.b.sum2);
#endif
	}
	TIFFSetField(tiff,TIFFTAG_IMAGEDESCRIPTION,msg);

	if ((code=WriteTIFF(tiff,image,largest))>0) {
	    sprintf(msg,"%s : TIFF %s",PATH1,WriteTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	return 0;
}
