
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STRLEN	256

#define XSIZE	image->xSize
#define YSIZE	image->ySize
#define IR	image->r
#define IG	image->g
#define IB	image->b
#define GP	GETPIXEL(image,x,y)
#define PIXELS	cm->pixels
#define CMR	cm->r
#define CMG	cm->g
#define CMB	cm->b

#define PBM(magic)	printf("P%d\n%d\t%d\n",magic,XSIZE,YSIZE)
#define PGPM(magic)	printf("P%d\n%d\t%d\t%d\n",magic,XSIZE,YSIZE,largest)
#define PRINT1(z)	printf("%d\n",(int)(z))
#define PUTINT1(z)	putchar((int)(z))
#define PRINT3(r,g,b)	printf("%d\t%d\t%d\n",(int)(r),(int)(g),(int)(b))
#define PUTINT3(r,g,b)	PUTINT1(r); PUTINT1(g); PUTINT1(b)
#define FORXY		for (y=0; y<YSIZE; y++) for (x=0; x<XSIZE; x++)
#define IFRGB(R,G,B)	\
	if ((r=(int)R)>*largest) *largest=r; \
	if ((g=(int)G)>*largest) *largest=g; \
	if ((b=(int)B)>*largest) *largest=b; \
	if (r!=g || r!=b) *color=1; else if (r!=0 && r!=255) *gray=1
#define PBMRAW(z)		\
	PBM(4); \
	for (y=0; y<YSIZE; y++) { \
	    bit=0; mask=0x80; \
	    for (x=0; x<XSIZE; x++) { \
		if (z) bit|=mask; \
		if ((mask>>=1)==0) { putchar(bit); bit=0; mask=0x80; } \
	    } \
	    if (mask!=0x80) putchar(bit); \
	}

static int	CheckGray(image)
Image		*image;
{
	int	x,y,z,largest=0;

	FORXY if ((z=GP)>largest) largest=z; return(largest);
}

static void	CheckCM(image,cm,color,gray,largest)
Image		*image;
CM		*cm;
int		*color,*gray,*largest;
{
	int	x,y,z,r,g,b;

	*color=(*gray)=(*largest)=0;
	FORXY { z=GP; IFRGB(CMR[z],CMG[z],CMB[z]); }
}

static void	CheckRGB(image,color,gray,largest)
Image		*image;
int		*color,*gray,*largest;
{
	int	x,y,r,g,b;

	*color=(*gray)=(*largest)=0;
	FORXY { IFRGB(IR[x][y],IG[x][y],IB[x][y]); }
}

static void	Gray2PBM(image)
Image		*image;
{
	int	x,y;

	PBM(1); FORXY PRINT1(GP);
}

static void	Gray2PBMRaw(image)
Image		*image;
{
	int	x,y,bit,mask;

	PBMRAW(GP)
}

static void	Gray2PGM(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(2); FORXY PRINT1(GP);
}

static void	Gray2PGMRaw(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(5); FORXY PUTINT1(GP);
}

static void	CM2PBM(image,cm)
Image		*image;
CM		*cm;
{
	int	x,y;

	PBM(1); FORXY PRINT1((CMR[GP])?1:0);
}

static void	CM2PBMRaw(image,cm)
Image		*image;
CM		*cm;
{
	int	x,y,bit,mask;

	PBMRAW(CMR[GP])
}

static void	CM2PGM(image,cm,largest)
Image		*image;
CM		*cm;
int		largest;
{
	int	x,y;

	PGPM(2); FORXY PRINT1(CMR[GP]);
}

static void	CM2PGMRaw(image,cm,largest)
Image		*image;
CM		*cm;
int		largest;
{
	int	x,y;

	PGPM(5); FORXY PUTINT1(CMR[GP]);
}

static void	CM2PPM(image,cm,largest)
Image		*image;
CM		*cm;
int		largest;
{
	int	x,y,z;

	PGPM(3); FORXY { z=GP; PRINT3(CMR[z],CMG[z],CMB[z]); }
}

static void	CM2PPMRaw(image,cm,largest)
Image		*image;
CM		*cm;
int		largest;
{
	int	x,y,z;

	PGPM(6); FORXY { z=GP; PUTINT3(CMR[z],CMG[z],CMB[z]); }
}

static void	RGB2PBM(image)
Image		*image;
{
	int	x,y;

	PBM(1); FORXY PRINT1((IR[x][y])?1:0);
}

static void	RGB2PBMRaw(image)
Image		*image;
{
	int	x,y,bit,mask;

	PBMRAW(IR[x][y]);
}

static void	RGB2PGM(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(2); FORXY PRINT1(IR[x][y]);
}

static void	RGB2PGMRaw(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(5); FORXY PUTINT1(IR[x][y]);
}

static void	RGB2PPM(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(3); FORXY PRINT3(IR[x][y],IG[x][y],IB[x][y]);
}

static void	RGB2PPMRaw(image,largest)
Image		*image;
int		largest;
{
	int	x,y;

	PGPM(6); FORXY { PUTINT3(IR[x][y],IG[x][y],IB[x][y]); }
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	argp,code,z,color,gray,largest,raw=0,ppm=0,small=0;
	TIFF	*tiff=NULL;
	char	msg[STRLEN],
		*usage="usage : tiff2pnm {-ppm} {-raw} {-small} TIFF";
	Image	*image;
	CM	*cm;

	for (argp=1; argp<argc; argp++)
	    if (argv[argp][0]=='-')
		switch(argv[argp][1]) {
		    case 'p' : case 'P' :
			if (ppm) Error(usage); ppm=1; break;
		    case 'r' : case 'R' :
			if (raw) Error(usage); raw=1; break;
		    case 's' : case 'S' :
			if (small) Error(usage); small=1; break;
		    default :
			Error(usage);
		}
	    else
		if ((tiff=TIFFOpen(argv[argp],"r"))==NULL)
		    Error("TIFF file not found.");

	if (tiff==NULL) Error(usage);

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(msg,"TIFF %s",ReadTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	if (cm==NULL) {
	    CheckRGB(image,&color,&gray,&largest);
	    if (!small || color)
		if (raw)
		    RGB2PPMRaw(image,largest);
		else
		    RGB2PPM(image,largest);
	    else
	    if (gray)
		if (raw)
		    RGB2PGMRaw(image,largest);
		else
		    RGB2PGM(image,largest);
	    else
		if (raw)
		    RGB2PBMRaw(image);
		else
		    RGB2PBM(image);
	}
	else
	    if (ppm) {
		for (z=0; z<PIXELS; z++) {
		    CMR[z]>>=8; CMG[z]>>=8; CMB[z]>>=8;
		}
		CheckCM(image,cm,&color,&gray,&largest);
		if (!small || color)
		    if (raw)
			CM2PPMRaw(image,cm,largest);
		    else
			CM2PPM(image,cm,largest);
		else
		if (gray)
		    if (raw)
			CM2PGMRaw(image,cm,largest);
		    else
			CM2PGM(image,cm,largest);
		else
		    if (raw)
			CM2PBMRaw(image,cm);
		    else
			CM2PBM(image,cm);
	    }
	    else {
		largest=CheckGray(image);
		if (!small || largest>1)
		    if (raw)
			if (largest<256)
			    Gray2PGMRaw(image,largest);
			else {
	fputs("too many bits per sample. raw mode not possible.\n",stderr);
			    Gray2PGM(image,largest);
			}
		    else
			Gray2PGM(image,largest);
		else
		    if (raw)
			Gray2PBMRaw(image);
		    else
			Gray2PBM(image);
	    }

	return(0);
}
