
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STRLEN	256

#define R(x,y)	image->r[x][y]
#define G(x,y)	image->g[x][y]
#define B(x,y)	image->b[x][y]

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	Image	*image;
	CM	*cm;
	int	code,z,y,x,x2,xSize,ySize,i,j,r=255,g=255,b=255;
	char	msg[STRLEN];
	byte	*y0Buf,*y1Buf,*uBuf,*vBuf,*y0Ptr,*y1Ptr,*uPtr,*vPtr;
	FILE	*yFile,*uFile,*vFile;
	long	r0,g0,b0,r1,b1,g1,r2,g2,b2,r3,g3,b3,y0,y1,y2,y3,u,v;

	if (argc!=5 && argc!=8)
	    Error("usage : tiff2yuv TIFF {bgR bgG bgB} Y U V");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(msg,"TIFF %s",ReadTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	if (cm!=NULL) {
	    for (z=0; z<cm->pixels; z++) {
		cm->r[z]>>=8; cm->g[z]>>=8; cm->b[z]>>=8;
	    }
	    for (y=0; y<image->ySize; y++)
	    for (x=0; x<image->xSize; x++) {
		z=GETPIXEL(image,x,y); R(x,y)=(byte)cm->r[z];
				       G(x,y)=(byte)cm->g[z];
				       B(x,y)=(byte)cm->b[z];
	    }
	    FreeCM(cm);
	}

	if (argc==8)
	    if ((r=atoi(argv[2]))<0 || r>255 ||
		(g=atoi(argv[3]))<0 || g>255 ||
		(b=atoi(argv[4]))<0 || b>255) Error("bad background color.");

	xSize=(x2=(image->xSize+1)/2)*2;
	ySize=(   (image->ySize+1)/2)*2; printf("%d\t%d\n",xSize,ySize);

	if ((y0Buf=(byte *)malloc(xSize))==NULL ||
	    (y1Buf=(byte *)malloc(xSize))==NULL ||
	    (uBuf=(byte *)malloc(x2))==NULL ||
	    (vBuf=(byte *)malloc(x2))==NULL) Error("memory allocation error.");

	if ((yFile=fopen(argv[argc-3],"w"))==NULL ||
	    (uFile=fopen(argv[argc-2],"w"))==NULL ||
	    (vFile=fopen(argv[argc-1],"w"))==NULL) Error("file not stored.");

	for (y=0; y<ySize; y+=2) {
	    y0Ptr=y0Buf; y1Ptr=y1Buf; uPtr=uBuf; vPtr=vBuf;
	    for (x=0; x<xSize; x+=2) {
		r0=(long)R(x,y); g0=(long)G(x,y); b0=(long)B(x,y);
		if ((i=x+1)<image->xSize) {
		    r1=(long)R(i,y); g1=(long)G(i,y); b1=(long)B(i,y);
		}
		else {
		    r1=(long)r; g1=(long)g; b1=(long)b;
		}
		if ((j=y+1)<image->ySize) {
		    r2=(long)R(x,j); g2=(long)G(x,j); b2=(long)B(x,j);
		}
		else {
		    r2=(long)r; g2=(long)g; b2=(long)b;
		}
		if (i<image->xSize && j<image->ySize) {
		    r3=(long)R(i,j); g3=(long)G(i,j); b3=(long)B(i,j);
		}
		else {
		    r3=(long)r; g3=(long)g; b3=(long)b;
		}

#define Y(r,g,b)	( 19595L*r+38469L*g+ 7471L*b)
#define U(r,g,b)	(-11056L*r-21712L*g+32768L*b)
#define V(r,g,b)	( 32768L*r-27440L*g- 5328L*b)

		y0=Y(r0,g0,b0); y1=Y(r1,g1,b1);
		y2=Y(r2,g2,b2); y3=Y(r3,g3,b3);
		u=(U(r0,g0,b0)+U(r1,g1,b1)+U(r2,g2,b2)+U(r3,g3,b3))/4L;
		v=(V(r0,g0,b0)+V(r1,g1,b1)+V(r2,g2,b2)+V(r3,g3,b3))/4L;

		y0=(y0*219L)/255L+1048576L; y1=(y1*219L)/255L+1048576L;
		y2=(y2*219L)/255L+1048576L; y3=(y3*219L)/255L+1048576L;
		u=(u*224L)/255L;
		v=(v*224L)/255L;

		*(y0Ptr++)=(byte)(y0>>16); *(y0Ptr++)=(byte)(y1>>16);
		*(y1Ptr++)=(byte)(y2>>16); *(y1Ptr++)=(byte)(y3>>16);
		*(uPtr++)=(byte)((u>>16)+128);
		*(vPtr++)=(byte)((v>>16)+128);
	    }
	    if (fwrite(y0Buf,xSize,1,yFile)!=1 ||
		fwrite(y1Buf,xSize,1,yFile)!=1 ||
		fwrite(uBuf,x2,1,uFile)!=1 ||
		fwrite(vBuf,x2,1,vFile)!=1) Error("data write error.");
	}
	fclose(vFile); fclose(uFile); fclose(yFile); return 0;
}
