
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "copysl.h"

extern void	CopyTAGControl(),CopyTAG();

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static void	readCM(name,pixels,r,g,b)
char		*name;
int		*pixels;
uint16		**r,**g,**b;
{
	FILE	*file;
	int	pixel;

	if (*name=='-')
	    file=stdin;
	else
	    if ((file=fopen(name,"r"))==NULL) Error("CM file not found.");

	fscanf(file,"%d",pixels);

	if ((*r=(uint16 *)calloc(*pixels,sizeof(**r)))==NULL ||
	    (*g=(uint16 *)calloc(*pixels,sizeof(**g)))==NULL ||
	    (*b=(uint16 *)calloc(*pixels,sizeof(**b)))==NULL)
	    Error("CM data memory allocation error.");

	for (pixel=0; pixel<*pixels; pixel++)
	    fscanf(file,"%hu %hu %hu",&((*r)[pixel]),
				      &((*g)[pixel]),
				      &((*b)[pixel]));

	if (file!=stdin) fclose(file);
}

#define TGF	TIFFGetField
#define TGFD	TIFFGetFieldDefaulted
#define TSF	TIFFSetField

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*org,*new;
	uint16	pi,bps,*r,*g,*b;
	int	pixels,pixel,bw;
	char	*msg;

	if (argc<2 || argc>4) Error("usage : tiffcm orgTIFF {{CM} newTIFF}");

	if ((org=TIFFOpen(argv[1],"r"))==NULL)
	    Error("original TIFF file not found.");

	if (!TGF(org,TIFFTAG_PHOTOMETRIC,&pi))
	    Error("no photometric interpretation.");

	if (pi!=PHOTOMETRIC_MASK &&
	    pi!=PHOTOMETRIC_MINISBLACK &&
	    pi!=PHOTOMETRIC_MINISWHITE &&
	    pi!=PHOTOMETRIC_PALETTE) Error("bad photometric interpretation.");

	if (!TGF(org,TIFFTAG_BITSPERSAMPLE,&bps))
	    TGFD(org,TIFFTAG_BITSPERSAMPLE,&bps);

	if (argc==2) {
	    printf("%d\n",pixels=1<<bps);

	    switch(pi) {
		case PHOTOMETRIC_MASK :
		case PHOTOMETRIC_MINISBLACK :
		    for (pixel=0; pixel<pixels; pixel++) {
			bw=(0xffff*pixel)/(pixels-1);
			printf("%d\t%d\t%d\n",bw,bw,bw);
		    }
		    break;
		case PHOTOMETRIC_MINISWHITE :
		    for (pixel=0; pixel<pixels; pixel++) {
			bw=0xffff-(0xffff*pixel)/(pixels-1);
			printf("%d\t%d\t%d\n",bw,bw,bw);
		    }
		    break;
		case PHOTOMETRIC_PALETTE :
		    if (!TGF(org,TIFFTAG_COLORMAP,&r,&g,&b))
			Error("no colormap.");

		    for (pixel=0; pixel<pixels; pixel++)
			printf("%d\t%d\t%d\n",r[pixel],g[pixel],b[pixel]);

		    break;
	    }
	}
	else {
	    if (argc==4) {
		readCM(argv[2],&pixels,&r,&g,&b);
		if (pixels<1<<bps) Error("color number not match.");
	    }

	    if ((new=TIFFOpen(argv[argc-1],"w"))==NULL)
		Error("new TIFF file not created.");

	    if (argc==3) {
		CopyTAGControl(TIFFTAG_COLORMAP,0); CopyTAG(org,new);
	        TSF(new,TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_MINISBLACK);
	    }
	    else {
		CopyTAG(org,new);
		TSF(new,TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_PALETTE);
		TSF(new,TIFFTAG_COLORMAP,r,g,b);
	    }
	    if ((msg=CopyScanline(org,new))!=NULL) Error(msg);

	    TIFFClose(new);
	}
	TIFFClose(org); return(0);
}
