
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static void	CMHistogram(image,pixels,x1,y1,x2,y2)
Image		*image;
int		pixels,x1,y1,x2,y2;
{
	int	z,x,y,*count=(int *)calloc(pixels,sizeof(*count)),pixel=0;

	if (count==NULL) Error("memory allocation error for CM image.");

	for (z=0; z<pixels; z++) count[z]=0;

	for (x=x1; x<=x2; x++)
	for (y=y1; y<=y2; y++)
	    if ((++(count[GETPIXEL(image,x,y)]))==1) ++pixel;

	printf("%d\n",pixel);
	for (z=0; z<pixels; z++) if (count[z]) printf("%d\t%d\n",z,count[z]);
}

#ifndef	BIT
#define BIT	4
#endif

#ifndef	HASH
#define HASH	0
#endif

#define SIZE	(1<<BIT)

#if	HASH==0
#define MASK(rgb)	((rgb)>>(8-BIT))
#else
#define MASK(rgb)	((rgb)&(SIZE-1))
#endif

typedef union {
		int	i;
		struct {
			byte	r,g,b,noUse;
		} s;
	} RGB;

typedef struct {
		RGB	rgb;
		int	count;
	} Histogram;

static void	SortHistogram(histogram,pixel)
Histogram	*histogram;
int		pixel;
{
	int		lo=0,hi=pixel-1,key=histogram[(lo+hi)/2].count;
	Histogram	swap;

	do {
	    while (histogram[lo].count>key) ++lo;
	    while (histogram[hi].count<key) --hi;
	    if (lo<=hi) {
		if (lo<hi) {
		    swap=histogram[lo]; histogram[lo]=histogram[hi];
		    histogram[hi]=swap;
		}
		++lo; --hi;
	    }
	} while (lo<=hi);

	if (hi>0      ) SortHistogram(  histogram,     hi+1);
	if (lo<pixel-1) SortHistogram(&(histogram[lo]),pixel-lo);
}

#define List	struct LIST
	List {
		int		pixel;
		Histogram	histogram;
		List		*next;
	};

static void	RGBHistogram(image,x1,y1,x2,y2)
Image		*image;
int		x1,y1,x2,y2;
{
	int		r,g,b,x,y,z,pixel=0;
	List		*top[SIZE][SIZE][SIZE],**last,*list;
	RGB		rgb;
	Histogram	*histogram;

	for (r=0; r<SIZE; r++)
	for (g=0; g<SIZE; g++)
	for (b=0; b<SIZE; b++) top[r][g][b]=NULL; rgb.i=0;

	for (x=x1; x<=x2; x++)
	for (y=y1; y<=y2; y++) {
	    last=(&(top[MASK(rgb.s.r=image->r[x][y])]
		       [MASK(rgb.s.g=image->g[x][y])]
		       [MASK(rgb.s.b=image->b[x][y])]));

	    while ((list=(*last))!=NULL)
		if (list->histogram.rgb.i==rgb.i)
		    break;
		else
		    last=(&(list->next));

	    if (list==NULL) {
		if (((*last)=(List *)malloc(sizeof(**last)))==NULL)
		    Error("memory allocation error for RGB image (1).");

		(*last)->histogram.rgb.i=rgb.i;
		(*last)->histogram.count=1;
		(*last)->pixel=pixel++;
		(*last)->next=NULL;
	    }
	    else
		++(list->histogram.count);
	}

	if ((histogram=(Histogram *)calloc(pixel,sizeof(*histogram)))==NULL)
	    Error("memory allocation error for RGB image (2).");

	for (r=0; r<SIZE; r++)
	for (g=0; g<SIZE; g++)
	for (b=0; b<SIZE; b++)
	    for (list=top[r][g][b]; list!=NULL; list=list->next)
		histogram[list->pixel]=list->histogram;

	SortHistogram(histogram,pixel);

	printf("%d\n",pixel);
	for (z=0; z<pixel; z++)
	    printf("%d\t%d\t%d\t%d\n",histogram[z].rgb.s.r,
				      histogram[z].rgb.s.g,
				      histogram[z].rgb.s.b,histogram[z].count);
}

#define STRLEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	Image	*image;
	CM	*cm;
	int	code,x1,y1,x2,y2,bps;
	char	msg[STRLEN];

	if (argc!=2 && argc!=6) Error("usage : tiffhc TIFF {x1 y1 x2 y2}");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(msg,"TIFF %s",ReadTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	if (argc==2) {
	    x1=0; x2=image->xSize-1;
	    y1=0; y2=image->ySize-1;
	}
	else
	    if ((x1=atoi(argv[2]))<0 ||
		(x2=atoi(argv[4]))>=image->xSize || x1>x2 ||
		(y1=atoi(argv[3]))<0 ||
		(y2=atoi(argv[5]))>=image->ySize || y1>y2) Error("bad area.");

	if (cm!=NULL)
	    for (bps=1; cm->pixels>1<<bps; bps++) ;
	else
	    bps=8;

	printf("%d\t%d\t%d\t%d\n",(cm!=NULL)?1:3,bps,x2-x1+1,y2-y1+1);

	if (cm!=NULL)
	    CMHistogram(image,cm->pixels,x1,y1,x2,y2);
	else
	    RGBHistogram(image,x1,y1,x2,y2);

	return(0);
}
