
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

#define ERROR(msg)	fputs(msg,stderr); fputc('\n',stderr)

static void	Error(msg)
char		*msg;
{
	ERROR(msg); exit(1);
}

#define FORXY(image)	for (y=0; y<image->ySize; y++) \
			for (x=0; x<image->xSize; x++)

static char	*CopyMask(image)
Image		*image;
{
	char	*ptr,*top=(char *)malloc(image->xSize*image->ySize);
	int	x,y;
	byte	**r=image->r,**g=image->g,**b=image->b;

	if (top==NULL) Error("mask memory allocation error.");

	ptr=top; FORXY(image) *(ptr++)=r[x][y]|g[x][y]|b[x][y]; return top;
}

static CM	*MakeCM(limit)
int		limit;
{
	int	bps,pxl,idx;
	CM	*cm;

	for (bps=1; (1<<bps)<=limit; bps++); pxl=(1<<bps)-1;

	if ((cm=AllocCM(pxl+1))==NULL)
	    Error("color map memory allocation error.");

	for (idx=0; idx<=pxl; idx++)
	    cm->r[idx]=cm->g[idx]=cm->b[idx]=(word)((0xffff*idx)/pxl);

	return cm;
}

static CM	*MaskGrayII(mask,bg,fg)	/* image==bg */
char		*mask;
Image		*bg,*fg;
{
	int	x,y,z,limit=0;
	byte	**r=bg->r,**fgr=fg->r,
		**g=bg->g,**fgg=fg->g,
		**b=bg->b,**fgb=fg->b;

	FORXY(bg) {
	    if (*(mask++)) {
		z=GETPIXEL(fg,x,y);
		r[x][y]=fgr[x][y]; g[x][y]=fgg[x][y]; b[x][y]=fgb[x][y];
	    }
	    else
		z=GETPIXEL(bg,x,y);

	    if (z>limit) limit=z;
	}
	return MakeCM(limit);
}

static CM	*MaskGrayVI(mask,bg,fg)	/* image==fg */
char		*mask;
int		bg;
Image		*fg;
{
	int	x,y,z,limit=bg;

	FORXY(fg)
	    if (*(mask++)) {
		z=GETPIXEL(fg,x,y); if (z>limit) limit=z;
	    }
	    else {
		PUTPIXEL(fg,x,y,bg);
	    }

	return MakeCM(limit);
}

static CM	*MaskGrayIV(mask,bg,fg)	/* image=bg */
char		*mask;
Image		*bg;
int		fg;
{
	int	x,y,z,limit=fg;

	FORXY(bg)
	    if (*(mask++)) {
		PUTPIXEL(bg,x,y,fg);
	    }
	    else {
		z=GETPIXEL(bg,x,y); if (z>limit) limit=z;
	    }

	return MakeCM(limit);
}

static CM	*MaskGrayVV(mask,bg,fg,image)
char		*mask;
int		bg,fg;
Image		*image;
{
	int	x,y,z;

	FORXY(image) {
	    z=(*(mask++))?fg:bg; PUTPIXEL(image,x,y,z);
	}
	return MakeCM((fg>bg)?fg:bg);
}

static void	CM2RGB(image,cm)
Image		*image;
CM		*cm;
{
	int	x,y,z;
	word	*cr=cm->r,*cg=cm->g,*cb=cm->b;
	byte	**ir=image->r,**ig=image->g,**ib=image->b;

	for (z=0; z<cm->pixels; z++) {
	    cr[z]>>=8; cg[z]>>=8; cb[z]>>=8;
	}
	FORXY(image) {
	    z=GETPIXEL(image,x,y); ir[x][y]=(byte)cr[z];
				   ig[x][y]=(byte)cg[z];
				   ib[x][y]=(byte)cb[z];
	}
}

static CM	*RGB2Gray(image)
Image		*image;
{
	int	x,y;
	byte	**g=image->g,**b=image->b;

	FORXY(image) g[x][y]=b[x][y]=0; return MakeCM(255);
}

static CM	*MaskRGBII(mask,bg,fg)	/* image==bg */
char		*mask;
Image		*bg,*fg;
{
	int	x,y,gray=1;
	byte	**r=bg->r,**fgr=fg->r,
		**g=bg->g,**fgg=fg->g,
		**b=bg->b,**fgb=fg->b;

	FORXY(bg) {
	    if (*(mask++)) {
		r[x][y]=fgr[x][y]; g[x][y]=fgg[x][y]; b[x][y]=fgb[x][y];
	    }
	    if (gray) gray=(r[x][y]==g[x][y] && r[x][y]==b[x][y]);
	}
	return (gray)?RGB2Gray(bg):NULL;
}

static CM	*MaskRGBVI(mask,bgR,bgG,bgB,fg)	/* image==fg */
char		*mask;
int		bgR,bgG,bgB;
Image		*fg;
{
	int	x,y,gray=(bgR==bgG && bgR==bgB);
	byte	bgr=(byte)bgR,**r=fg->r,
		bgg=(byte)bgG,**g=fg->g,
		bgb=(byte)bgB,**b=fg->b;

	FORXY(fg)
	    if (*(mask++)) {
		if (gray) gray=(r[x][y]==g[x][y] && r[x][y]==b[x][y]);
	    }
	    else {
		r[x][y]=bgr; g[x][y]=bgg; b[x][y]=bgb;
	    }

	return (gray)?RGB2Gray(fg):NULL;
}

static CM	*MaskRGBIV(mask,bg,fgR,fgG,fgB)	/* image==bg */
char		*mask;
Image		*bg;
int		fgR,fgG,fgB;
{
	int	x,y,gray=(fgR==fgG && fgR==fgB);
	byte	**r=bg->r,fgr=(byte)fgR,
		**g=bg->g,fgg=(byte)fgG,
		**b=bg->b,fgb=(byte)fgB;

	FORXY(bg)
	    if (*(mask++)) {
		r[x][y]=fgr; g[x][y]=fgg; b[x][y]=fgb;
	    }
	    else
		if (gray) gray=(r[x][y]==g[x][y] && r[x][y]==b[x][y]);

	return (gray)?RGB2Gray(bg):NULL;
}

static CM	*MaskRGBVV(mask,bgR,bgG,bgB,fgR,fgG,fgB,image)
char		*mask;
int		bgR,bgG,bgB,fgR,fgG,fgB;
Image		*image;
{
	int	x,y;
	byte	bgr=(byte)bgR,fgr=(byte)fgR,**r=image->r,
		bgg=(byte)bgG,fgg=(byte)fgG,**g=image->g,
		bgb=(byte)bgB,fgb=(byte)fgB,**b=image->b;

	FORXY(image)
	    if (*(mask++)) {
		r[x][y]=fgr; g[x][y]=fgg; b[x][y]=fgb;
	    }
	    else {
		r[x][y]=bgr; g[x][y]=bgg; b[x][y]=bgb;
	    }

	return (bgR==bgG && bgR==bgB &&
		fgR==fgG && fgR==fgB)?RGB2Gray(image):NULL;
}

#define BAD_FILE(msg)	{ fprintf(stderr,"%s : TIFF ",path); Error(msg); }

static void	ReadTIFFFile(path,image,cm,xSize,ySize,needCM)
char		*path;
Image		**image;
CM		**cm;
int		xSize,ySize,needCM;
{
	TIFF	*tiff;
	int	code;

	if ((tiff=TIFFOpen(path,"r"))==NULL) BAD_FILE("file not found.");

	if ((code=ReadTIFF(tiff,image,cm))>0) BAD_FILE(ReadTIFFError(code));

	TIFFClose(tiff);

	if (xSize && ySize)
	    if (xSize!=(*image)->xSize ||
		ySize!=(*image)->ySize) BAD_FILE("image size not match.");

	if (needCM && cm==NULL) BAD_FILE("image type not match.");
}

static int	ScanPixelValue(arg,limit)
char		*arg;
int		limit;
{
	char	*str=(*arg=='-')?&(arg[1]):arg;
	int	pv;

	if ((pv=atoi(str))<0 || pv>=limit) {
	    fprintf(stderr,"%s : ",arg); Error("bad pixel value.");
	}
	return pv;
}

static void	Usage()
{
	ERROR("usage : tiffmask maskTIFF SPP BG FG newTIFF");
	ERROR("\tSPP = 1 or 3 (samples per pixel)");
	ERROR("\twhen SPP == 1 (gray scale image)");
	ERROR("\t\tBG = bgTIFF or -bg (background image or pixel value)");
	ERROR("\t\tFG = fgTIFF or -fg (foreground image or piexl value)");
	ERROR("\twhen SPP == 3 (RGB color image)");
	ERROR("\t\tBG = bgTIFF or {-bgR bgG bgB}");
	ERROR("\t\tFG = fgTIFF or { fgR fgG fgB}");
	exit(1);
}

#define LIMIT	(1<<24)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	spp,xSize,ySize,bg,fg,code;
	Image	*image,*iB,*iF;
	CM	*cm;
	char	*mask,*path=argv[argc-1];
	TIFF	*tiff;

	if (argc!=6 && argc!=8 && argc!=10) Usage();

	if ((spp=atoi(argv[2]))!=1 && spp!=3) Error("bad samples per pixel.");

	if (spp==1 && argc!=6) Usage();

	ReadTIFFFile(argv[1],&image,&cm,0,0,0); if (cm!=NULL) FreeCM(cm);

	mask=CopyMask(image); xSize=image->xSize;
			      ySize=image->ySize; FreeImage(image);

	if (spp==1) {
	    if (argv[3][0]=='-') {
		iB=NULL; bg=ScanPixelValue(argv[3],LIMIT);
	    }
	    else {
		ReadTIFFFile(argv[3],&iB,&cm,xSize,ySize,1); FreeCM(cm);
	    }
	    if (argv[4][0]=='-') {
		iF=NULL; fg=ScanPixelValue(argv[4],LIMIT);
	    }
	    else {
		ReadTIFFFile(argv[4],&iF,&cm,xSize,ySize,1); FreeCM(cm);
	    }
	    if (iB!=NULL && iF!=NULL)
		cm=MaskGrayII(mask,image=iB,iF);
	    else if (iF!=NULL)
		cm=MaskGrayVI(mask,bg,image=iF);
	    else if (iB!=NULL)
		cm=MaskGrayIV(mask,image=iB,fg);
	    else
		if ((image=AllocImage(xSize,ySize))!=NULL)
		    cm=MaskGrayVV(mask,bg,fg,image);
	}
	else
	    if (argc==6) {
		ReadTIFFFile(argv[3],&iB,&cm,xSize,ySize,0);
		if (cm!=NULL) { CM2RGB(iB,cm); FreeCM(cm); }
		ReadTIFFFile(argv[4],&iF,&cm,xSize,ySize,0);
		if (cm!=NULL) { CM2RGB(iF,cm); FreeCM(cm); }
		image=iB; cm=MaskRGBII(mask,iB,iF);
	    }
	    else if (argc==8)
		if (argv[3][0]=='-') {
		    ReadTIFFFile(argv[6],&image,&cm,xSize,ySize,0);
		    if (cm!=NULL) { CM2RGB(image,cm); FreeCM(cm); }
		    cm=MaskRGBVI(mask,ScanPixelValue(argv[3],256),
				      ScanPixelValue(argv[4],256),
				      ScanPixelValue(argv[5],256),image);
		}
		else {
		    ReadTIFFFile(argv[3],&image,&cm,xSize,ySize,0);
		    if (cm!=NULL) { CM2RGB(image,cm); FreeCM(cm); }
		    cm=MaskRGBIV(mask,image,ScanPixelValue(argv[4],256),
					    ScanPixelValue(argv[5],256),
					    ScanPixelValue(argv[6],256));
		}
	    else
		if ((image=AllocImage(xSize,ySize))!=NULL)
		    cm=MaskRGBVV(mask,ScanPixelValue(argv[3],256),
				      ScanPixelValue(argv[4],256),
				      ScanPixelValue(argv[5],256),
				      ScanPixelValue(argv[6],256),
				      ScanPixelValue(argv[7],256),
				      ScanPixelValue(argv[8],256),image);

	if (image==NULL) Error("image memory allocation error.");

	if ((tiff=TIFFOpen(path,"w"))==NULL) BAD_FILE("file not created.");

	if ((code=WriteTIFF(tiff,image,cm))>0) BAD_FILE(WriteTIFFError(code));

	TIFFClose(tiff); return 0;
}
