
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define STRLEN	256
#define DELIM	" \t,:;@#"

static int	ParseLine(gray,xSize,ySize,image,cm,x1,y1,x2,y2,x0,y0)
int		gray,xSize,ySize;
Image		**image;
CM		**cm;
int		*x1,*y1,*x2,*y2,*x0,*y0;
{
	char	line[STRLEN],*eol,backup[STRLEN],*name,*ptr[6];
	TIFF	*tiff;
	int	code,num;

	*image=NULL; *cm=NULL;
loop:;
	if (*image!=NULL) {
	    FreeImage(*image); *image=NULL;
	}
	if (*cm!=NULL) {
	    FreeCM(*cm); *cm=NULL;
	}
	do {
	    if (fgets(line,STRLEN,stdin)==NULL) return 0;

	    if ((eol=strchr(line,'\n'))==NULL) Error("too long line.");

	    *eol='\0'; strcpy(backup,line);
	} while ((name=strtok(line,DELIM))==NULL);

	if ((tiff=TIFFOpen(name,"r"))==NULL) {
	    fprintf(stderr,"TIFF file not found in: %s\n",backup);
	    goto loop;
	}
	code=ReadTIFF(tiff,image,cm); TIFFClose(tiff);

	if (code>0) {
	    fprintf(stderr,"TIFF %s in: %s\n",ReadTIFFError(code),backup);
	    goto loop;
	}
	if (gray && *cm==NULL) {
	    fprintf(stderr,"RGB color image appeared in: %s\n",backup);
	    goto loop;
	}
	for (num=0; num<6; num++)
	    if ((ptr[num]=strtok(NULL,DELIM))==NULL) break;

	*x0=0; *x1=0; *x2=(*image)->xSize-1;
	*y0=0; *y1=0; *y2=(*image)->ySize-1;
	switch(num) {
	    case 0 :
		break;
	    case 2 :
		*x0=atoi(ptr[0]); *y0=atoi(ptr[1]); break;
	    case 4 :
		*x1=atoi(ptr[0]); *y1=atoi(ptr[1]);
		*x2=atoi(ptr[2]); *y2=atoi(ptr[3]); break;
	    case 6 :
		*x1=atoi(ptr[0]); *y1=atoi(ptr[1]);
		*x2=atoi(ptr[2]); *y2=atoi(ptr[3]);
		*x0=atoi(ptr[4]); *y0=atoi(ptr[5]); break;
	    default :
		fprintf(stderr,"bad number of parameters in: %s\n",backup);
		goto loop;
	}
	if (*x1<0 || *x2>=(*image)->xSize || *x1>*x2 ||
	    *y1<0 || *y2>=(*image)->ySize || *y1>*y2 ||
	    *x0<0 || *x0+(*x2-*x1)>=xSize ||
	    *y0<0 || *y0+(*y2-*y1)>=ySize) {
	    fprintf(stderr,"bad parameter value(s) in: %s\n",backup);
	    goto loop;
	}
	return 1;
}

#define TSF	TIFFSetField
#define TWS	TIFFWriteScanline

static void	openTIFF(name,iw,il,bps,spp,pi,tiff,tsl)
char		*name;
int		iw,il,bps,spp,pi;
TIFF		**tiff;
uint8		**tsl;
{
	int	tss;

	if ((*tiff=TIFFOpen(name,"w"))==NULL) Error("TIFF file not created.");

	TSF(*tiff,TIFFTAG_IMAGEWIDTH,(uint32)iw);
	TSF(*tiff,TIFFTAG_IMAGELENGTH,(uint32)il);
	TSF(*tiff,TIFFTAG_BITSPERSAMPLE,(uint16)bps);
	TSF(*tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)spp);
	TSF(*tiff,TIFFTAG_PHOTOMETRIC,(uint16)pi);
	TSF(*tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);

	if ((*tsl=(uint8 *)malloc(tss=TIFFScanlineSize(*tiff)))==NULL)
	    Error("TIFF scanline memory allocation error.");

	TSF(*tiff,TIFFTAG_ROWSPERSTRIP,(tss>8192)?(uint32)1:(uint32)(8192/tss));
#ifdef LZW
	TSF(*tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
}

static void	PasteGray(base,bg,name)
Image		*base;
int		bg;
char		*name;
{
	int	x,y,pixels,x1,y1,x2,y2,x0,y0,h,v,pixel,bps,msb,mask;
	Image	*paste;
	CM	*cm;
	TIFF	*tiff;
	uint8	*tsl,*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;

	for (x=0; x<base->xSize; x++)
	for (y=0; y<base->ySize; y++) PUTPIXEL(base,x,y,bg);

	pixels=bg+1;
	while (ParseLine(1,base->xSize,base->ySize,
			 &paste,&cm,&x1,&y1,&x2,&y2,&x0,&y0)) {
	    for (x=x0, h=x1; h<=x2; h++, x++)
	    for (y=y0, v=y1; v<=y2; v++, y++) { 
		pixel=GETPIXEL(paste,h,v); PUTPIXEL(base,x,y,pixel);
	    }
	    FreeImage(paste);

	    if (cm->pixels>pixels) pixels=cm->pixels; FreeCM(cm);
	}
	for (bps=1; pixels>1<<bps; bps++) ; msb=1<<(bps-1);

	openTIFF(name,base->xSize,base->ySize,bps,1,
		(bps==1)?PHOTOMETRIC_MINISWHITE:
			 PHOTOMETRIC_MINISBLACK,&tiff,&tsl);

	for (y=0; y<base->ySize; y++) {
	    if (bps==8)
		for (ptr=tsl, x=0; x<base->xSize; x++)
		    *(ptr++)=(uint8)GETPIXEL(base,x,y);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=0; x<base->xSize; x++)
		    *(ptr16++)=(uint16)GETPIXEL(base,x,y);
	    else if (bps==24)
		for (ptr=tsl, x=0; x<base->xSize; x++) {
		    *(ptr++)=base->b[x][y];
		    *(ptr++)=base->g[x][y];
		    *(ptr++)=base->r[x][y];
		}
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=0; x<base->xSize; x++)
		    *(ptr32++)=(uint32)GETPIXEL(base,x,y);
	    else
		for (ptr=tsl, bit=0x80, x=0; x<base->xSize; x++) {
		    pixel=GETPIXEL(base,x,y);
		    for (mask=msb; mask!=0; mask>>=1) {
			if (pixel&mask) *ptr|=bit; else *ptr&=(~bit);

			if ((bit>>=1)==0) {
			    ++ptr; bit=0x80;
			}
		    }
		}

	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline gray data error.");
	}
	TIFFClose(tiff);
}

static void	PasteRGB(base,bgR,bgG,bgB,name)
Image		*base;
int		bgR,bgG,bgB;
char		*name;
{
	int	x,y,x1,y1,x2,y2,x0,y0,h,v,pixel;
	Image	*paste;
	CM	*cm;
	TIFF	*tiff;
	uint8	*tsl,*ptr;

	for (x=0; x<base->xSize; x++)
	for (y=0; y<base->ySize; y++) {
	    base->r[x][y]=(byte)bgR;
	    base->g[x][y]=(byte)bgG;
	    base->b[x][y]=(byte)bgB;
	}
	while (ParseLine(0,base->xSize,base->ySize,
			 &paste,&cm,&x1,&y1,&x2,&y2,&x0,&y0)) {
	    if (cm!=NULL) {
		for (pixel=0; pixel<cm->pixels; pixel++) {
		    cm->r[pixel]>>=8; cm->g[pixel]>>=8; cm->b[pixel]>>=8;
		}
		for (x=x0, h=x1; h<=x2; h++, x++)
		for (y=y0, v=y1; v<=y2; v++, y++) { 
		    pixel=GETPIXEL(paste,h,v);
		    base->r[x][y]=(byte)cm->r[pixel];
		    base->g[x][y]=(byte)cm->g[pixel];
		    base->b[x][y]=(byte)cm->b[pixel];
		}
		FreeCM(cm);
	    }
	    else
		for (x=x0, h=x1; h<=x2; h++, x++)
		for (y=y0, v=y1; v<=y2; v++, y++) { 
		    base->r[x][y]=paste->r[h][v];
		    base->g[x][y]=paste->g[h][v];
		    base->b[x][y]=paste->b[h][v];
		}

	    FreeImage(paste);
	}
	openTIFF(name,base->xSize,base->ySize,8,3,PHOTOMETRIC_RGB,&tiff,&tsl);

	for (y=0; y<base->ySize; y++) {
	    for (ptr=tsl, x=0; x<base->xSize; x++) {
		*(ptr++)=(uint8)base->r[x][y];
		*(ptr++)=(uint8)base->g[x][y];
		*(ptr++)=(uint8)base->b[x][y];
	    }
	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline RGB data error.");
	}
	TIFFClose(tiff);
}

static int	readBackGround(arg,lim)
char		*arg;
int		lim;
{
	int	bg=atoi(arg);

	if (bg<0 || bg>lim) Error("bad background color.");

	return bg;
}

#define BG	0
#define BGR	255
#define BGG	255
#define BGB	255

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	xSize,ySize,spp;
	Image	*image;

	if (argc!=5 && argc!=6 && argc!=8)
	Error("usage: tiffpaste xSize ySize [1 {bg}]|[3 {bgR bgG bgB}] TIFF");

	if ((xSize=atoi(argv[1]))<=0 ||
	    (ySize=atoi(argv[2]))<=0) Error("bad image size.");

	if ((image=AllocImage(xSize,ySize))==NULL)
	    Error("image memory allocation error.");

	if ((spp=atoi(argv[3]))!=1 && spp!=3) Error("bad samples per pixel.");

	if ((spp==1 && argc==8) ||
	    (spp==3 && argc==6)) Error("bad backgound color assignment.");

	if (spp==1)
	    if (argc==6)
		PasteGray(image,readBackGround(argv[4],0xffffff),argv[5]);
	    else
		PasteGray(image,BG,argv[4]);
	else
	    if (argc==8)
		PasteRGB(image,readBackGround(argv[4],0xff),
			       readBackGround(argv[5],0xff),
			       readBackGround(argv[6],0xff),argv[7]);
	    else
		PasteRGB(image,BGR,BGG,BGB,argv[4]);

	return 0;
}
