
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

extern int	clipline();
extern void	drawline();

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static Image	*image;
static CM	*cm;

static void	PrintPixelValue(x,y)
int		x,y;
{
	if (cm!=NULL)
	    printf("%d\t%d\t%d\n",x,y,GETPIXEL(image,x,y));
	else
	    printf("%d\t%d\t%d\t%d\t%d\n",x,y,(int)image->r[x][y],
					      (int)image->g[x][y],
					      (int)image->b[x][y]);
}

static void	ScanPixels(x1,y1,x2,y2)
int		x1,y1,x2,y2;
{
	if (clipline(&x1,&y1,&x2,&y2,0,0,image->xSize-1,image->ySize-1))
	    drawline(x1,y1,x2,y2,PrintPixelValue);
}

#define STRLEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*tiff;
	int	code,x1,y1,x2,y2;
	char	msg[STRLEN],str[STRLEN];

	if (argc!=2 && argc!=4 && argc!=6)
	    Error("usage : tiffppv TIFF {x y}|{x1 y1 x2 y2}");

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(msg,"TIFF %s",ReadTIFFError(code)); Error(msg);
	}
	TIFFClose(tiff);

	switch(argc) {
	    case 2 :
		while (fgets(str,STRLEN,stdin)!=NULL)
		    switch(sscanf(str,"%d %d %d %d",&x1,&y1,&x2,&y2)) {
			case 2 :
			    ScanPixels(x1,y1,x1,y1); break;
			case 4 :
			    ScanPixels(x1,y1,x2,y2); break;
		    }
		break;
	    case 4 :
		x1=atoi(argv[2]);
		y1=atoi(argv[3]); ScanPixels(x1,y1,x1,y1); break;
	    case 6 :
		x1=atoi(argv[2]);
		y1=atoi(argv[3]);
		x2=atoi(argv[4]);
		y2=atoi(argv[5]); ScanPixels(x1,y1,x2,y2); break;
	}
	return 0;
}
