
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"

extern void	CopyTAG();

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define TGF	TIFFGetField
#define TGFD	TIFFGetFieldDefaulted
#define TSF	TIFFSetField
#define TSS	TIFFScanlineSize
#define TRS	TIFFReadScanline
#define TWS	TIFFWriteScanline

static void	CutRectangularArea(oTIFF,nTIFF,oTsl,nTsl,ra,x1,y1,x2,y2,BPS,SPP)
TIFF		*oTIFF,*nTIFF;
uint8		*oTsl,*nTsl;
int		ra,x1,y1,x2,y2,BPS,SPP;
{
	tsample_t	sample;
	uint32		y,x,row,bytes=((x2-x1+1)*BPS+7)/8;
	uint8		*org,*new,*begin=(&(oTsl[(x1*BPS)/8]));
	int		shift1=(x1*BPS)%8,shift2=8-shift1;

	for (sample=0; sample<SPP; sample++) {
	    if (!ra) for (y=0; y<y1; y++) (void)TRS(oTIFF,oTsl,y,sample);

	    for (row=0, y=y1; y<=y2; y++, row++) {
		org=begin; new=nTsl;

		if (TRS(oTIFF,oTsl,y,sample)<0)
		    for (x=0; x<bytes; x++) *(new++)=0;
		else if (shift1==0)
		    for (x=0; x<bytes; x++) *(new++)=(*(org++));
		else
		    for (x=0; x<bytes; x++, org++)
			*(new++)=(uint8)(*org<<shift1)|(uint8)(org[1]>>shift2);

		(void)TWS(nTIFF,nTsl,row,sample);
	    }
	}
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	TIFF	*oTIFF,*nTIFF;
	uint32	iw,il;
	int	x1,y1,x2,y2,ra;
	uint16	pc,bps,spp;
	tsize_t	tss;
	uint8	*oTsl,*nTsl;

	if (argc!=7) Error("usage : tiffrac orgTIFF x1 y1 x2 y2 newTIFF");

	if ((oTIFF=TIFFOpen(argv[1],"r"))==NULL ||
	    !TGF(oTIFF,TIFFTAG_IMAGEWIDTH ,&iw) ||
	    !TGF(oTIFF,TIFFTAG_IMAGELENGTH,&il)) Error("file not open.");

	if ((x1=atoi(argv[2]))<0 || (x2=atoi(argv[4]))>=iw || x1>x2 ||
	    (y1=atoi(argv[3]))<0 || (y2=atoi(argv[5]))>=il || y1>y2)
	    Error("bad rectangular area.");

	if (!TGF(oTIFF,TIFFTAG_PLANARCONFIG,&pc))
	    TGFD(oTIFF,TIFFTAG_PLANARCONFIG,&pc);

	if (pc!=PLANARCONFIG_CONTIG &&
	    pc!=PLANARCONFIG_SEPARATE) Error("unknown planr configuration.");

	tss=TIFFScanlineSize(oTIFF);
	if ((oTsl=(uint8 *)malloc((size_t)tss))==NULL ||
	    (nTsl=(uint8 *)malloc((size_t)tss))==NULL) Error("memory error.");

	if ((nTIFF=TIFFOpen(argv[6],"w"))==NULL) Error("file not stored.");

	CopyTAG(oTIFF,nTIFF);
	TSF(nTIFF,TIFFTAG_IMAGEWIDTH ,(uint32)(x2-x1+1));
	TSF(nTIFF,TIFFTAG_IMAGELENGTH,(uint32)(y2-y1+1));
	tss=TIFFScanlineSize(nTIFF);
	TSF(nTIFF,TIFFTAG_ROWSPERSTRIP,(tss>8192)?1:8192/tss);

	if (!TGF(oTIFF,TIFFTAG_BITSPERSAMPLE,&bps))
	    TGFD(oTIFF,TIFFTAG_BITSPERSAMPLE,&bps);

	if (!TGF(oTIFF,TIFFTAG_SAMPLESPERPIXEL,&spp))
	    TGFD(oTIFF,TIFFTAG_SAMPLESPERPIXEL,&spp);
{
	uint16	compression;

	if (!TGF(oTIFF,TIFFTAG_COMPRESSION,&compression))
	    TGFD(oTIFF,TIFFTAG_COMPRESSION,&compression);

	if (compression==COMPRESSION_NONE)
	    ra=1;
	else
{
	uint32	rps;

	if (!TGF(oTIFF,TIFFTAG_ROWSPERSTRIP,&rps))
	    TGFD(oTIFF,TIFFTAG_ROWSPERSTRIP,&rps);

	ra=(rps==1)?1:0;
}}
	if (pc==PLANARCONFIG_CONTIG)
	    CutRectangularArea(oTIFF,nTIFF,oTsl,nTsl,
			       ra,x1,y1,x2,y2,(int)(bps*spp),1);
	else
	    CutRectangularArea(oTIFF,nTIFF,oTsl,nTsl,
			       ra,x1,y1,x2,y2,(int)bps,(int)spp);

	TIFFClose(oTIFF); TIFFClose(nTIFF); return 0;
}
