
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"
#include "writetiff.h"

extern void	CopyTAG();

#define MSGLEN	256

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static int	CheckAxis(str)
char		*str;
{
	int	axis;

	switch(str[0]) {
	    case '+' : axis=0; break;
	    case '-' : axis=1; break; default : return -1;
	}
	switch(str[1]) {
	    case 'x' : case 'X' : axis+=0; break;
	    case 'y' : case 'Y' : axis+=2; break; default : return -1;
	}
	return axis;
}

static Image	*CopyImage(org,hAxis,vAxis)
Image		*org;
int		hAxis,vAxis;
{
	int	hn,vn,x0,xh,xv,y0,yh,yv,h,v,x,y;
	Image	*new;

	switch(hAxis) {
	    case 0 : hn=org->xSize; x0=0; xh=1; xv=0; break;
	    case 1 : hn=org->xSize; x0=hn-1; xh=(-1); xv=0; break;
	    case 2 : hn=org->ySize; y0=0; yh=1; yv=0; break;
	    case 3 : hn=org->ySize; y0=hn-1; yh=(-1); yv=0; break;
	}
	switch(vAxis) {
	    case 0 : vn=org->xSize; x0=0; xh=0; xv=1; break;
	    case 1 : vn=org->xSize; x0=vn-1; xh=0; xv=(-1); break;
	    case 2 : vn=org->ySize; y0=0; yh=0; yv=1; break;
	    case 3 : vn=org->ySize; y0=vn-1; yh=0; yv=(-1); break;
	}
	if ((new=AllocImage(hn,vn))==NULL) return NULL;

	for (h=0; h<hn; h++) for (v=0; v<vn; v++) {
	    x=x0+xh*h+xv*v;
	    y=y0+yh*h+yv*v; new->r[h][v]=org->r[x][y];
			    new->g[h][v]=org->g[x][y];
			    new->b[h][v]=org->b[x][y];
	}
	return new;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	hAxis,vAxis,code;
	TIFF	*orgT,*newT;
	Image	*orgI,*newI;
	CM	*cm;
	char	msg[MSGLEN];

	if (argc!=5) Error("usage : tiffrar orgTIFF hAxis vAxis newTIFF");

	if ((hAxis=CheckAxis(argv[2]))<0 ||
	    (vAxis=CheckAxis(argv[3]))<0 || hAxis/2==vAxis/2)
	    Error("bad direction of axis.");

	if ((orgT=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(orgT,&orgI,&cm))>0) {
	    sprintf(msg,"TIFF %s",ReadTIFFError(code)); Error(msg);
	}

	if ((newI=CopyImage(orgI,hAxis,vAxis))==NULL)
	    Error("memory allocation error.");

	if ((newT=TIFFOpen(argv[4],"w"))==NULL) Error("TIFF file not stored.");

	CopyTAG(orgT,newT); TIFFClose(orgT);

	if ((code=WriteTIFF(newT,newI,cm))>0) {
	    sprintf(msg,"TIFF %s",WriteTIFFError(code)); Error(msg);
	}
	TIFFClose(newT); return 0;
}
