
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tiffio.h"
#include "cmimage.h"
#include "readtiff.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static Image	*image;
static int	dx,dy;
static double	xO,yO,i[2][2],m[2][2];
static TIFF	*tiff;
static uint8	*tsl;

#define XSIZE	image->xSize
#define YSIZE	image->ySize

#define H(x,y)	(i[0][0]*(x)+i[0][1]*(y))
#define V(x,y)	(i[1][0]*(x)+i[1][1]*(y))
#define X(h,v)	(m[0][0]*(h)+m[0][1]*(v))
#define Y(h,v)	(m[1][0]*(h)+m[1][1]*(v))

#define TSF	TIFFSetField
#define TWS	TIFFWriteScanline

#ifdef	AVERAGE

static byte	**R,**G,**B;
static int	**cnt;
static double	**sum;

static void	RGB(bgR,bgG,bgB)
int		bgR,bgG,bgB;
{
	int	y,x,spl,bg,h,v;
	byte	**old,**new;
	double	h0,v0,x0,y0;
	uint8	*ptr;

	for (spl=0; spl<3; spl++) {
	    switch(spl) {
		case 0 : old=image->r; bg=bgR; new=R; break;
		case 1 : old=image->g; bg=bgG; new=G; break;
		case 2 : old=image->b; bg=bgB; new=B; break;
	    }
	    for (y=0; y<dy; y++)
	    for (x=0; x<dx; x++) {
		cnt[y][x]=0; sum[y][x]=0.0; new[y][x]=(byte)bg;
	    }
	    for (v=0; v<YSIZE; v++) { v0=(double)v+0.5;
	    for (h=0; h<XSIZE; h++) { h0=(double)h+0.5;
		++cnt[y=(int)(Y(h0,v0)-yO)]
		     [x=(int)(X(h0,v0)-xO)]; sum[y][x]+=(double)old[h][v];
	    }}
	    for (y=0; y<dy; y++)
	    for (x=0; x<dx; x++)
		if (cnt[y][x])
		    new[y][x]=(byte)(sum[y][x]/(double)cnt[y][x]+0.5);
		else {
		    x0=(double)x+xO+0.5;
		    y0=(double)y+yO+0.5;
		    if ((h=(int)H(x0,y0))>=0 && h<XSIZE &&
			(v=(int)V(x0,y0))>=0 && v<YSIZE) new[y][x]=old[h][v];
		}
	}
	for (y=0; y<dy; y++) {
	    for (ptr=tsl, x=0; x<dx; x++) {
		*(ptr++)=(uint8)R[y][x];
		*(ptr++)=(uint8)G[y][x];
		*(ptr++)=(uint8)B[y][x];
	    }
	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline RGB data error.");
	}
}

static int	Z(x,y,z)
int		x,y,z;
{
	double	x0,y0;
	int	h,v;

	if (cnt[y][x]) return (int)(sum[y][x]/(double)cnt[y][x]+0.5);

	x0=(double)x+xO+0.5;
	y0=(double)y+yO+0.5;
	if ((h=(int)H(x0,y0))>=0 && h<XSIZE &&
	    (v=(int)V(x0,y0))>=0 && v<YSIZE) return GETPIXEL(image,h,v);

	return z;
}

static void	Gray(bg,bps)
int		bg,bps;
{
	int	x,y,h,v,z,msk;
	double	h0,v0;
	uint8	*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;

	for (y=0; y<dy; y++)
	for (x=0; x<dx; x++) {
	    cnt[y][x]=0; sum[y][x]=0.0;
	}
	for (v=0; v<YSIZE; v++) { v0=(double)v+0.5;
	for (h=0; h<XSIZE; h++) { h0=(double)h+0.5;
	    ++cnt[y=(int)(Y(h0,v0)-yO)]
		 [x=(int)(X(h0,v0)-xO)]; sum[y][x]+=(double)GETPIXEL(image,h,v);
	}}
	for (y=0; y<dy; y++) {
	    if (bps==8)
		for (ptr=tsl, x=0; x<dx; x++, ptr++) *ptr=(uint8)Z(x,y,bg);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=0; x<dx; x++, ptr16++)
		    *ptr16=(uint16)Z(x,y,bg);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=0; x<dx; x++, ptr32++)
		    *ptr32=(uint32)Z(x,y,bg);
	    else
		for (ptr=tsl, bit=0x80, x=0; x<dx; x++)
		    for (z=Z(x,y,bg), msk=1<<(bps-1); msk!=0; msk>>=1) {
			if (z&msk) *ptr|=bit; else *ptr&=(~bit);

			if ((bit>>=1)==0) {
			    ++ptr; bit=0x80;
			}
		    }

	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline gray data error.");
	}
}
#else	/* AVERAGE */

static void	RGB(bgR,bgG,bgB)
int		bgR,bgG,bgB;
{
	int	x,y,h,v;
	double	x0,y0;
	uint8	*ptr;

	for (y=0; y<dy; y++) {
	    y0=(double)y+yO+0.5;
	    for (ptr=tsl, x=0; x<dx; x++) {
		x0=(double)x+xO+0.5;
		if ((h=(int)H(x0,y0))>=0 && h<XSIZE &&
		    (v=(int)V(x0,y0))>=0 && v<YSIZE) {
		    *(ptr++)=(uint8)image->r[h][v];
		    *(ptr++)=(uint8)image->g[h][v];
		    *(ptr++)=(uint8)image->b[h][v];
		}
		else {
		    *(ptr++)=(uint8)bgR;
		    *(ptr++)=(uint8)bgG;
		    *(ptr++)=(uint8)bgB;
		}
	    }
	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline RGB data error.");
	}
}

static int	Z(x0,y0,z)
double		x0,y0;
int		z;
{
	int	h,v;

	if ((h=(int)H(x0,y0))>=0 && h<XSIZE &&
	    (v=(int)V(x0,y0))>=0 && v<YSIZE) return GETPIXEL(image,h,v);

	return z;
}

#define X0	(double)x+xO+0.5

static void	Gray(bg,bps)
int		bg,bps;
{
	int	x,y,z,msk;
	double	y0;
	uint8	*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;

	for (y=0; y<dy; y++) {
	    y0=(double)y+yO+0.5;

	    if (bps==8)
		for (ptr=tsl, x=0; x<dx; x++, ptr++) *ptr=(uint8)Z(X0,y0,bg);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=0; x<dx; x++, ptr16++)
		    *ptr16=(uint16)Z(X0,y0,bg);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=0; x<dx; x++, ptr32++)
		    *ptr32=(uint32)Z(X0,y0,bg);
	    else
		for (ptr=tsl, bit=0x80, x=0; x<dx; x++) {
		    for (z=Z(X0,y0,bg), msk=1<<(bps-1); msk!=0; msk>>=1) {
			if (z&msk) *ptr|=bit; else *ptr&=(~bit);

			if ((bit>>=1)==0) {
			    ++ptr; bit=0x80;
			}
		    }
		}

	    if (TWS(tiff,tsl,y,0)<0)
		Error("TIFF write scanline gray data error.");
	}
}
#endif	/* AVERAGE */

static int	readBG(arg,lim)
char		*arg;
int		lim;
{
	int	bg=atoi(arg);

	if (bg<0 || bg>=lim) Error("bad background color.");

	return bg;
}

static void	MinMax(z,z1,z2)
double		z,*z1,*z2;
{
	if (z<*z1) *z1=z; if (z>*z2) *z2=z;
}

#define STR_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	CM	*cm;
	int	code,spp,x,y,bps,tss,pi,z,bg=0,bgR=255,bgG=255,bgB=255;
	char	str[STR_LEN];
	double	xRatio,yRatio,angle,xy;
#ifdef	AVERAGE
	char	*wmae="work memory allocation error.";
#endif

	if (argc!=7 && argc!=8 && argc!=10)
	Error(
"usage : tiffsr orgTIFF xRatio yRatio angle [1 {bg}]|[3 {bgR bgG bgB}] newTIFF"
	);

	if ((tiff=TIFFOpen(argv[1],"r"))==NULL) Error("TIFF file not found.");

	if ((code=ReadTIFF(tiff,&image,&cm))>0) {
	    sprintf(str,"TIFF %s",ReadTIFFError(code)); Error(str);
	}
	TIFFClose(tiff);

	if ((xRatio=atof(argv[2]))==0.0 ||
	    (yRatio=atof(argv[3]))==0.0) Error("bad magnification ratio.");

	angle=atof(argv[4])*M_PI/180.0;
	spp=atoi(argv[5]);

	if ((spp!=1 && spp!=3) ||
	    (spp==1 && cm==NULL)) Error("bad samples per pixel.");

	if ((spp==1 && argc==10) ||
	    (spp==3 && argc==8)) Error("bad background color assignment.");

	switch(argc) {
	    case 8 :
		bg=readBG(argv[6],cm->pixels); break;
	    case 10 :
		bgR=readBG(argv[6],256);
		bgG=readBG(argv[7],256);
		bgB=readBG(argv[8],256); break;
	}

	m[0][0]=  cos(angle) *xRatio; m[0][1]=(-sin(angle))*yRatio;
	m[1][0]=  sin(angle) *xRatio; m[1][1]=  cos(angle) *yRatio;
	i[0][0]=  cos(angle) /xRatio; i[0][1]=  sin(angle) /xRatio;
	i[1][0]=(-sin(angle))/yRatio; i[1][1]=  cos(angle) /yRatio;

	xO=xy=X(0.0,0.0); MinMax(X((double)XSIZE,0.0	      ),&xO,&xy);
			  MinMax(X((double)XSIZE,(double)YSIZE),&xO,&xy);
			  MinMax(X(0.0,		 (double)YSIZE),&xO,&xy);
	dx=(int)ceil(xy-xO);
	yO=xy=Y(0.0,0.0); MinMax(Y((double)XSIZE,0.0	      ),&yO,&xy);
			  MinMax(Y((double)XSIZE,(double)YSIZE),&yO,&xy);
			  MinMax(Y(0.0,		 (double)YSIZE),&yO,&xy);
	dy=(int)ceil(xy-yO);

	if (spp==1) {
	    for (bps=1; cm->pixels>(1<<bps); bps++) ;
	    tss=(dx*bps+7)/8;
	    pi=(bps==1)?PHOTOMETRIC_MINISWHITE:PHOTOMETRIC_MINISBLACK;

	    sprintf(str,"%lg %lg %lg %s%d%s",
			xRatio,yRatio,angle*180.0/M_PI,
			(argc==8)?"":"{",bg,(argc==8)?"":"}");
	}
	else {
	    bps=8; tss=dx*3; pi=PHOTOMETRIC_RGB;

	    sprintf(str,"%lg %lg %lg %s%d %d %d%s",
			xRatio,yRatio,angle*180.0/M_PI,
			(argc==10)?"":"{",bgR,bgG,bgB,(argc==10)?"":"}");
#ifdef	AVERAGE
	    if ((R=(byte **)calloc(dy,sizeof(*R)))==NULL ||
		(G=(byte **)calloc(dy,sizeof(*G)))==NULL ||
		(B=(byte **)calloc(dy,sizeof(*B)))==NULL) Error(wmae);

	    for (y=0; y<dy; y++)
		if ((R[y]=(byte *)calloc(dx,sizeof(**R)))==NULL ||
		    (G[y]=(byte *)calloc(dx,sizeof(**G)))==NULL ||
		    (B[y]=(byte *)calloc(dx,sizeof(**B)))==NULL) Error(wmae);
#endif
	}
#ifdef	AVERAGE
	if ((cnt=(int    **)calloc(dy,sizeof(*cnt)))==NULL ||
	    (sum=(double **)calloc(dy,sizeof(*sum)))==NULL) Error(wmae);

	for (y=0; y<dy; y++)
	    if ((cnt[y]=(int    *)calloc(dx,sizeof(**cnt)))==NULL ||
		(sum[y]=(double *)calloc(dx,sizeof(**sum)))==NULL) Error(wmae);
#endif
	if ((tsl=(uint8 *)malloc(tss))==NULL)
	    Error("memory for TIFF scanline data allocation error.");

	if ((tiff=TIFFOpen(argv[argc-1],"w"))==NULL)
	    Error("TIFF file not created.");

	TSF(tiff,TIFFTAG_IMAGEWIDTH,(uint16)dx);
	TSF(tiff,TIFFTAG_IMAGELENGTH,(uint16)dy);
	TSF(tiff,TIFFTAG_BITSPERSAMPLE,(uint16)bps);
	TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)spp);
	TSF(tiff,TIFFTAG_PHOTOMETRIC,(uint16)pi);
	TSF(tiff,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);

	TSF(tiff,TIFFTAG_ROWSPERSTRIP,(tss>8192)?(int32)1:(int32)(8192/tss));
#ifdef LZW
	TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW);
#endif
	TSF(tiff,TIFFTAG_IMAGEDESCRIPTION,str);

	if (spp==1)
	    Gray(bg,bps);
	else {
	    if (cm!=NULL) {
		for (z=0; z<cm->pixels; z++) {
		    cm->r[z]>>=8; cm->g[z]>>=8; cm->b[z]>>=8;
		}
		for (x=0; x<XSIZE; x++)
		for (y=0; y<YSIZE; y++) {
		    z=GETPIXEL(image,x,y); image->r[x][y]=(byte)cm->r[z];
					   image->g[x][y]=(byte)cm->g[z];
					   image->b[x][y]=(byte)cm->b[z];
		}
	    }
	    RGB(bgR,bgG,bgB);
	}
	TIFFClose(tiff); return 0;
}
