
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"
#include "cmimage.h"
#include "writetiff.h"

#define ERROR_NO	11
#define ERROR_SL	1
#define ERROR_SIZE	2
#define ERROR_PI	3
#define ERROR_BPS	4
#define ERROR_SPP	5
#define ERROR_PC	6
#define ERROR_RPS	7
#define ERROR_COMPRESS	8
#define ERROR_CM	9
#define ERROR_WS	10

static char	*Error[ERROR_NO]={
		"write error.",
/* SL        */	"scanline data allocation error.",
/* SIZE      */	"image size description write error.",
/* PI        */	"photometric interpretation write error.",
/* BPS       */	"bits per sample write error.",
/* SPP       */	"samples per pixel write error.",
/* PC        */	"planar configuration write error.",
/* RPS       */	"rows per strip write error.",
/* COMPRESS  */	"compression scheme write error.",
/* CM        */	"color map write error.",
/* WS        */	"scanline data write error."
		};

char	*WriteTIFFError(code)
int	code;
{
	if (code<1 || code>=ERROR_NO) return(*Error); return(Error[code]);
}

#define TWS	TIFFWriteScanline
#define XSIZE	image->xSize
#define YSIZE	image->ySize
#define R	image->r
#define G	image->g
#define B	image->b

static int	RGBImage(tiff,image,tsl)
TIFF		*tiff;
Image		*image;
uint8		*tsl;
{
	int	y,x;
	uint8	*ptr;

	for (y=0; y<YSIZE; y++) {
	    ptr=tsl;
	    for (x=0; x<XSIZE; x++) {
		*(ptr++)=R[x][y]; *(ptr++)=G[x][y]; *(ptr++)=B[x][y];
	    }
	    if (TWS(tiff,tsl,y,(uint16)0)<0) return(0);
	}
	return(1);
}

static int	CMImage(tiff,image,tsl,bps)
TIFF		*tiff;
Image		*image;
uint8		*tsl;
int		bps;
{
	int	y,x,pxl,msk;
	uint8	*ptr,bit;
	uint16	*ptr16;
	uint32	*ptr32;

	for (y=0; y<YSIZE; y++) {
	    if (bps==8)
		for (ptr=tsl, x=0; x<XSIZE; x++) *(ptr++)=GETPIXEL(image,x,y);
	    else if (bps==16)
		for (ptr16=(uint16 *)tsl, x=0; x<XSIZE; x++)
		    *(ptr16++)=GETPIXEL(image,x,y);
	    else if (bps==32)
		for (ptr32=(uint32 *)tsl, x=0; x<XSIZE; x++)
		    *(ptr32++)=GETPIXEL(image,x,y);
	    else
		for (ptr=tsl, bit=0x80, x=0; x<XSIZE; x++) {
		    pxl=GETPIXEL(image,x,y);
		    for (msk=1<<(bps-1); msk>0; msk>>=1) {
			if (pxl&msk) *ptr|=bit; else *ptr&=(~bit);
			if ((bit>>=1)==0) { ++ptr; bit=0x80; }
		    }
		}
	    if (TWS(tiff,tsl,y,(uint16)0)<0) return(0);
	}
	return(1);
}

#define TSF	TIFFSetField
#define PIXELS	cm->pixels
#define CMR	cm->r
#define CMG	cm->g
#define CMB	cm->b

#define PC	PLANARCONFIG_CONTIG

int	WriteTIFF(tiff,image,cm)
TIFF	*tiff;
Image	*image;
CM	*cm;
{
	int	spp,bps,pxl,idx,pi,tss;
	uint8	*tsl;

	if (cm==NULL) {
	    spp=3; bps=8; pi=PHOTOMETRIC_RGB;
	}
	else {
	    spp=1; for (bps=1; (1<<bps)<PIXELS; bps++);
	    pxl=(1<<bps)-1;
	    for (idx=0; idx<PIXELS; idx++)
		if (CMR[idx]!=(word)(0xffff-(0xffff*idx)/pxl) ||
		    CMR[idx]!=CMG[idx] || CMR[idx]!=CMB[idx]) break;
	    if (idx==PIXELS)
		pi=PHOTOMETRIC_MINISWHITE;
	    else {
		for (idx=0; idx<PIXELS; idx++)
		    if (CMR[idx]!=(word)((0xffff*idx)/pxl) ||
			CMR[idx]!=CMG[idx] || CMR[idx]!=CMB[idx]) break;
		if (idx==PIXELS)
		    pi=PHOTOMETRIC_MINISBLACK;
		else
		    pi=PHOTOMETRIC_PALETTE;
	    }
	}

	if (!TSF(tiff,TIFFTAG_IMAGEWIDTH, (uint32)XSIZE) ||
	    !TSF(tiff,TIFFTAG_IMAGELENGTH,(uint32)YSIZE)) return(ERROR_SIZE);

	if (!TSF(tiff,TIFFTAG_PHOTOMETRIC,(uint16)pi)) return(ERROR_PI);

	if (!TSF(tiff,TIFFTAG_BITSPERSAMPLE,(uint16)bps)) return(ERROR_BPS);

	if (!TSF(tiff,TIFFTAG_SAMPLESPERPIXEL,(uint16)spp)) return(ERROR_SPP);

	if (!TSF(tiff,TIFFTAG_PLANARCONFIG,(uint16)PC)) return(ERROR_PC);

	tss=TIFFScanlineSize(tiff);
	if (!TSF(tiff,TIFFTAG_ROWSPERSTRIP,
		(tss>8192)?(uint32)1:(uint32)(8192/tss))) return(ERROR_RPS);
#ifdef	LZW
	if (!TSF(tiff,TIFFTAG_COMPRESSION,COMPRESSION_LZW))
	    return(ERROR_COMPRESS);
#endif
	if ((tsl=(uint8 *)malloc(tss))==NULL) return(ERROR_SL);

	if (cm==NULL) {
	    if (!RGBImage(tiff,image,tsl)) return(ERROR_WS);
	}
	else {
	    if (pi==PHOTOMETRIC_PALETTE)
		if (!TSF(tiff,TIFFTAG_COLORMAP,CMR,CMG,CMB)) return(ERROR_CM);

	    if (!CMImage(tiff,image,tsl,bps)) return(ERROR_WS);
	}
	free(tsl); return(0);
}
